# -*- coding: latin-1 -*-
"""
   ADAM.py

   Classes for communicating with ADAM I/O modules:
      4060 relay module
      4050 digital I/O module
      4017/4117/4015 ADC-module
      4080 counter module

   First instantiate ONE ADAM class (per serial port), then
   the classes controlling the modules themselves - one/module.
   Can raise exceptions ADAM_IOError, ADAM_ValueError.

   Initialises modules to use no checksums in line .
   Line speed 9600 bps (FIXED!?). 
   Uses class cSerial (in cSerial.py) for serial communication.

   *****                                                 *****
   *****     ALL COMMANDS IN UPPERCASE ending with CR    *****
   *****                                                 *****
   
   Revision history:
      2006-09-13/14  MAJOR CHANGES -> incompatible with previous version,
                     +ADAM4117, +ADAM4015
      2006-09-19  Minor enhancements: try to write/get response thrice (SendCmd)
      2006-11-24  class ScanADAMs
      2007-06-15  MAJOR CHANGES in ADAM 40x7: conv_eqs from [] -> {}
      2007-10-25  ADAM-4017 (== 4017+), 4117 UNIPOLAR hex readings:
         RANGE SAME AS IN bipolar!!!!!!!!!!!!!!!!!!
      2008-01-18  4017 self.toV bug fix (to dict)
      2008-06-19  4016 address change, class ADAM4016
      2008-07-29  4117 range '07' added (4-20 mA)
      2009-03-25  +4021 (DAC)
      2009-05-09  4016 fixes
      2010-10-01  checksum handling added, tested ONLY for 4117
      2011-05-24  ADAM4069, ADAM4060: return status of 8 chs instead of 4
      2014-03-25  cSerial id->mid (for the new cSerial version)
      2014-04-04  bug fix (ADAM range: was OB -> 0B)
      2014-05-22  fixed (due to above mentioned change)
      2014-09-24  ADAM Init message change (normally no printout)
      2015-01-29  +4019+ (thermocouple/ADC), other minor changes (NOT TESTED)
"""
import re
from time import time,sleep
import cSerial

# 0.9 s too short a time...???
SERIAL_TIMEOUT=1.2      # max wait time for ADAM response

class ADAM_IOError(Exception):
   """
      I/O error (telnet, serial)
   """
   def __init__(self,value):
      self.value=value
   def __str__(self):
      return repr(self.value)

class ADAM_ValueError(Exception):
   """
      Illegal parameter etc.
   """
   def __init__(self,value):
      self.value=value
   def __str__(self):
      return repr(self.value)


#################################
# Serial communication with ADAMs
#
# This class must be instantiated first (as it is used for serial
# communication), after which it is passed
# as a parameter to ADAM4??? class init (ADAM_IO_ref)
#
class ADAM4000(cSerial.cSerial):
   """
      Serial communication with ADAM. This class (instance)
      reference is used by ADAM4xxx classes. Exceptions
      are raised for errors and also logged. Inherits the methods
      from class cSerial.

      Revision history
         2005-04-13 change-of-order in __init__ (ini,logger)
         2006-09-14 name change: ADAM -> ADAM4000, method SendCmd
   """

   def __init__(self,ini,logger=None,timeout=SERIAL_TIMEOUT,checksum=False):
      """
         Get serial port parameters and open the port
      """
      self.timeout=timeout
      self.checksum=checksum
      k="serial_port"
      if not ini.has_key(k): k="SerialPort"
      cSerial.cSerial.__init__(self,ini[k],
         timeout=self.timeout,logger=logger,mid='ADAM')
      self.open_port()


   def SendCmd(self,cmds,response=-1,raiseex=False,scanmode=False,nochecksum=False):
      '''
         Write command to ADAM-module, and if response!=0,
         wait for an answer, check its validity, and if response>0
         its length (without CR), and return it. By default
         (response==-1) returns response without checking its length.
         In case of invalid response and logger!=None, writes out
         the response as logger.error, in addition to writing it out
         to stdout (stderr?).

         If response is invalid, return None

         Revision history:
            2006-11-24  +scanmode, only print "sout" if last round
            2010-10-01  checksum handling

      '''
      if not cmds: return
      if self.checksum and not nochecksum:
         cmds=checksum(cmds)
      else:
         if not cmds[-1]=='\r': cmds+='\r'
      for i in xrange(3):
         try:
            s=self.write_port(cmds)    # this also empties buffer before writing
#            print "'%s'->'%s'"%(cmds[:-1],s)   # debug
            if not response: return None
            s=s.strip()
            sout=''
            if not s:
               sout="No response from ADAM4xxx, cmd:%s"%(cmds,)
            elif self.checksum and not nochecksum:
               ss=checkchecksum(s)
               if not ss:
                  sout="Invalid checksum ADAM4xxx: '%s' -> '%s'"%(cmds[:-1],s)
               s=ss
            if not sout and (s[0] not in ['>','!'] or response>0 and len(s)!=response):
               sout='Invalid cmd/wrong response length ADAM4xxx, cmd:"%s"->"%s"'%(
                  cmds.strip(),s)
            if sout and not scanmode:
               if i==2:
                  print sout
                  if self.logger: self.logger.error(sout)
                  if raiseex: 
                     raise ADAM_IOError,sout
                  sleep(.2)
            else:
               break
         except:
            if raiseex: raise
      return s



   def SendInit(self,cmds,modinfo=''):
      """
         Send init string to ADAM trying both with and without checksums
         (if checksum mode activated)
      """
      if self.checksum:
         try:
            self.SendCmd(cmds,raiseex=True,nochecksum=False)
            return
         except:
            print "  Error in ADAM %s init with checksum..."%modinfo
            print "Trying to initialise without checksum..."
      self.SendCmd(cmds,raiseex=True,nochecksum=True)
      

####################
## Helper functions
octhexdict = {0:'000',1:'001',2:'010',3:'011',4:'100',5:'101',6:'110',7:'111'}
def dec2Hex(decim):
   """
   Convert decimal to two character upper case hex for ADAM addressing
   """
   try:
      dec=int(decim)
      if not 0 <= dec <= 255:
         raise ADAM_ValueError,"Invalid ADAM address:%d"%dec
      hx=hex(dec).upper()[-2:]
      if hx[0]=='X': return "0"+hx[1]
      return hx
   except:
      return None

def binfromhex(hexstr):
   """
   Return 8 bit binary representation of value given in hex
   """
   try:
      x=int(hexstr,16)
      s=''.join([octhexdict[int(dig)] for dig in oct(x)])
      return ("00000000"+s)[-8:]
   except:
      return None

def dictfromhex(hexstr, length=8):
   """
   Return dict {0:0/1, 1:0/1, ...} of value given in hex.
   """
   try:
      bin=binfromhex(hexstr)     # hex character for status
      dct={}
      for i in range(0,length):
         dct[i]=int(bin[-i-1])
      return dct
   except:
      return None

def hex2int(hexs):
   '''
      Convert two's complement hex string (four chars, e.g. 10FF)
      to integer (range -32768 - 32767).
   '''
   try:
      i=int(hexs,16)
      if i>32767: i-=65536
      return i
   except:
      return None

def checksum(s,addCR=True):
   '''
      Return s with added 2-char (hex value) checksum
      (modulo 256 (100h) sum of all the ASCII values)
      
      If last character is CR, adds checksum before it,
      and by default appends CR as the last character.
   '''
   iscr=''
   ss=s[:]
   if ss[-1]=='\r': 
      ss=ss[:-1]
      iscr='\r'
   if addCR:
      iscr='\r'
   
   cs=sum([ord(x) for x in ss if x>=' '])%256
   csh=dec2Hex(cs)
   return "%s%s%s"%(ss,csh,iscr)
   

def checkchecksum(s):
   '''
      Check that the content and checksum match
      Return '' if not, otherwise s without checksum
      
      Strips white space (i.e. CRs) characters from end
      of the string
   '''
   ss=s.rstrip()
   csh=ss[-2:]
   ss=ss[:-2]
   if csh==checksum(ss,addCR=False)[-2:]: 
      return ss
   return ''
      
##################################################################
## ADAM 12 bit DAC module
class ADAM4021:
   """
      ADAM-4021 DAC module, ranges 0-10 VDC, 4-20mA, 0-20 mA
   """
   dac_ranges={'0-10V':'32','0-20mA':'30','4-20mA':'31'}
   
   def __init__(self,ADAM_IO_ref,addr,range='0-10V'):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address.

      Init: %aaaatt0600(cr)
         - aaaa: 2*addr
         - tt: 30: 0-20 mA, 31: 4-20 mA, 32: 0-10 V
         - 06: 9600 bps
         - 00: no checksum, immediate change, eng. units

      Module requires maximum of 20 ms seconds to perform auto calibration
      after reconfiguration.
      """
      self.addr=int(addr)
      self.haddr=dec2Hex(addr)
      self.ioref=ADAM_IO_ref
      self.checksum='00'
      if self.ioref.checksum: self.checksum='40'
      cmds="%%%s%s%s06%s\r"%(self.haddr,self.haddr,self.self.dac_ranges[range],
         self.checksum)
      self.ioref.SendInit(cmds)
      sleep(.5)
      self.SentOutput(0,True) # set to 0, and set 0 to startup-value

      
   def SetOutput(self,value,set2Startup=False):
      '''
         Set the module output value, return set value
      '''
      cmds="#%s%s"%(self.haddr,value)
      s=self.ioref.SendCmd(cmds)
      if set2Startup and s.find('>')>=0:
         cmds="$%s4"%(self.haddr,)
         self.ioref.SendCmd(cmds)
      return self.GetSetValue()
      
   def GetSetValue(self):
      '''
         Return set ("Last Value Readback") value
      '''
      cmds="$%s6"%(self.haddr,)
      s=self.ioref.SendCmd(cmds).strip()
      if s[0]!='!':  return None    # something wrong
      return float(s[3:])
      
      
##################################################################
## ADAM 2 ch counter/frequency input module
class ADAM4080:
   """
      ADAM-4080 and NUDAM-6080 two 32-bit ch isolated counter/frequency input module
      Max freq. 50 kHz, isolated input: <1 V == 0, 3.5-30 V == 1

      Selected commands:
         $aa5NS   start(S=1)/stop(S=0) counter 0(N=0)/1(N=1) -> !aa\r
         $aa6N    clear counter 0(N=0)/1(N=1) -> !aa\r [while counting: ok]
         $aa7N    read and clear overflow flag of counter 0(N=0)/1(N=1)
                  --> !aaV , flag was set, V=1, not set: V=0

   With 10 V signal it seems that ADAM can count up to 300 kHZ (tested with
   one module) and NUDAM up to 200 kHz (tested with one module)
   """
   def __init__(self,ADAM_IO_ref,addr):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address.

      Init: %aaaa080680(cr)
         - aaaa: 2*addr
         - 50: counter input mode (51h: frequency)
         - 06: 9600 bps
         - 00: no checksum (freq: gate time 0.1 s), 40 with checksum

      Module requires maximum of 7 seconds to perform auto calibration
      after reconfiguration. This method sleeps 4 seconds after the command.

      Also sets module to read isolated input (by calling
      self.SetInputModeToIsolated()), gate mode to disabled,
      digital filter to disabled and max. counter values to FFFFFFFF.

      HOW ABOUT "Set Maximum Counter Value", page 271 (4-179), CHECK!!!
      
      """
      self.addr=int(addr)
      self.haddr=dec2Hex(addr)
      self.ioref=ADAM_IO_ref
      self.prevCounts=[0,0]         # counts at last read (initialisation zeroes)
      self.readTime = self.prevReadTime=[0.,0.]
      self.checksum='00'
      if ADAM_IO_ref.checksum: self.checksum='40'
      strng="%%%s%s5006%s\r"%(self.haddr,self.haddr,self.checksum)
      self.ioref.SendInit(strng)
      sleep(4)
      self.SetInputModeToIsolated()
      strng="$%sA2\r"%self.haddr    # set gate "disabled"
      self.ioref.SendCmd(strng)
      strng="$%s40\r"%self.haddr    # disable filter
      self.ioref.SendCmd(strng)
      self.clear_limit=int('0x8FFFFFFF',16)  # clear the counter after this
      for cntr in range(2):
         strng="$%s3%dFFFFFFFF\r"%(self.haddr,cntr)   # set to max (in UPPERCASE!)
         self.ioref.SendCmd(strng)


   def SetInputModeToIsolated(self,isolated=True):
      '''
         By default (isolated==True) sets the input module to
         use photo-isolated inputs (==False: use non-isolated)
      '''
      strng="$%sB%d\r"%(self.haddr,int(isolated))
      self.ioref.SendCmd(strng)

   def StartStopCounter(self,cntr,start=1):
      """
         Start (default) or stop (start==0) counter cntr (default 0)
      """
      if start:
         self.ClearCounter(cntr)
         self.prevReadTime[cntr]=time()
      strng="$%s5%d%d\r"%(self.haddr,cntr,start)
      self.ioref.SendCmd(strng)

   def ClearCounter(self,cntr):
      '''
         Clear a counter

         Command: $aa6[1/0]     (clear counter)
      '''
      strng="$%s6%d\r"%(self.haddr,cntr)
      self.ioref.SendCmd(strng)
      self.prevCounts[cntr]=0


   def GetCounterResult(self,chno,asCPM=False):
      '''
         Return number of counts since last read,
         or optionally counts as CPM since last read.

         Revision history:
            2006-11-22  asCPM added
      '''
      clear=0
      chno=int(chno)
      if self.prevCounts[chno]>self.clear_limit:
         clear=1
      cnts=self._ReadCounter(chno,clear)
      if not clear:
         self.prevCounts[chno]=cnts
      dcnts=cnts-self.prevCounts[chno]
      if asCPM:
         try:
            ret=60.*dcnts/(self.readTime[chno]-self.prevReadTime[chno])
         except:
            ret=0.
      else:
         ret=dcnts
      self.prevReadTime=self.readTime
      return ret


   def _ReadCounter(self,cntr,clear=0):
      '''
         Read and return counter value, optionally clears counter after read
         (with some counts lost, i.e. ca 0.06 s worth of counts).
         Command: #aa[0/1]      (read -> return as hex)
            Response: >data(cr)
      '''
      v=None
      strng="#%s%d\r"%(self.haddr,cntr)
      self.readTime[cntr]=time()
      s=self.ioref.SendCmd(strng)
      n=s.find('>')
      if n>=0:
         hexs=s[n+1:n+9]
         v=int(hexs,16)
      if clear:
         self.ClearCounter(cntr)
      return v



##################################################################
## ADAM ADC module (16-bit, 8 chs (6 differential/2 SE)
class ADAM4017:
   """
   ADAM 8 ch 16-bit ADC module

   Revision history:
      2005-04-13  added: chs_to_enable,conv_eqs
      2008-07-29  +ch_ranges['07']
   """
   ch_ranges={
      '00':(-.015,.015),'01':(-.05,.05),'02':(-.1,.1),'03':(-.5,.5),
      '06':(-20,20),'07':(4.,20.),'08':(-10.,10.),'09':(-5.,5.),
      '0A':(-1.,1.),'0B':(-.5,.5),'0C':(-.15,.15),'0D':(-20.,20.),
      '0E':(0.,760.),'0F':(0.,1370.),'10':(-100.,400.),'11':(0.,1000.),
      '12':(500.,1750.),'13':(500.,1750.),'14':(500.,1800.),
      '15':(-15.,15.), '20':(-50,150),'21':(0,100),
      '22':(0,200),'23':(0,400),'24':(-200,200),'25':(-50,150),'26':(0,100),
      '27':(0,200),'28':(0,400),'29':(-200,200),'2A':(-40,160),'2B':(-30,120),
      '2C':(-80,100),'2D':(0,100),'48':(0.,10),'49':(0.,5.),
      '4A':(0.,1.),'4B':(0.,.5),'4C':(0.,.15),'4D':(0.,20.),'55':(0.,15.),
      }
   # 0E - 14: thermocouple ranges
   # 2? ranges are for PT-100: not used
   mV_ranges=['02','03','07','0B','0C','0D','4B','4C','4D'] # also mA

   def __init__(self,ADAM_IO_ref,addr,scale=None,
      chs_to_enable=[0,1,2,3,4,5,6,7],
      conv_eqs=None,sleeptime=5.,logger=None):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address.
      scale = DICTONARY, default 10: see self.input_range_dict
      chs_to_enable = list of channels to enable, by default all 8 chs (0..7)
      conv_eqs = DICTIONARY: volts -> units conversion equations, default "=v*1." for
         all the channels

      Init: %aaaa080600(cr)
         - 08: +-10V (09: +-5V, 0A: +-1V, 0B: +-0.5V) (00 for 4117)
         - 06: 9600 bps
         - 00: 50 Hz, no checksum, two's complement HEX
           (00: engineering units, 01: % from full scale, 02:twos complement)
           ONLY 4017+, 4117 support 01,02
           WITH cheksum: 40,41,42

      Module requires maximum of 7 seconds to perform auto calibration
      after reconfiguration. This method sleeps sleeptime seconds after the command.

      Revision history:
         2006-09-13 BUG FIX: is scale < 1V, ADAM gives values in mV, 4117 handling added
         2008-01-18 self.toV fix
         2010-10-01 checksum
         2015-01-29 4-20ma -> 4-20mA
      """
      self.input_range_dict={"0.15":"0C","0.5":"0B","1":"0A","5":"09","10":"08","4-20mA":"07"}
      self.type="4017" # could be read from the module (also)
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)



   def InitModule(self,ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=5,logger=None):
      '''
         Initialise module

         Revision history:
            2007-05-16 MAJOR CHANGE:  conv_eqs changed from [] to {}
            2007-07-03 Fix for 4017 scale, if it is given as dict
            2015-01-29 +4019p
      '''
      assert conv_eqs==None or type(conv_eqs)==type({})
      self.logger=logger
      self.addr=int(addr)
      self.haddr=dec2Hex(addr)
      self.ioref=ADAM_IO_ref
      self.checksum='0'
      if self.ioref.checksum: self.checksum='4'
      # reg exp to get data from Analog Data In (all channels)
      self.get_readings_re=re.compile(r'''
         [+|-]          # starts with + or -
         [\d.]*         # digit or decimal point
         [\b]*          # white space (should not occur)
      ''', re.VERBOSE)
      rng="00" ; fmat="2"
      self.inHex=True ; self.toV={}
      for i in range(8): self.toV[i]=1.
      if self.type in ["4017","4015","4017p","4016","4019p"]:
         if self.type in ("4015","4019p"):
            rng="00"    # must
         else:
            sc=10
            if self.type=='4016': sc=3
            if type(scale)==type({}):
               sc=0
               for i in scale:
                  try:
                     if int(scale[i])>sc: sc=int(scale[i])
                  except:
                     pass
            elif scale!=None:
               sc=scale
            rng=self.input_range_dict[str(sc)]     
            if rng in ["0B","0C"] or self.type=="4016":     # how about 4019+
               for i in range(8): self.toV[i]=1e-3  # convert mv to V
         fmat="0"
         self.inHex=False
      strng="%%%s%s%s06%s%s\r"%(self.haddr,self.haddr,rng,self.checksum,fmat)
      self.ioref.SendInit(strng)
      sleep(2.)      # some sleep required

      rng=None ; onerng=False
      if type(scale)==type("") or type(scale)==type(10) or self.type=="4016":
         onerng=True
         rng=self.input_range_dict[str(scale)]

      if self.type in ["4117","4015","4017p","4019p"]:  # set ch ranges individually
         for ch in xrange(8):
            if self.type=="4015" and ch>5: break
            if not ch in chs_to_enable: continue
            if not onerng:
               try:
                  rng=self.input_range_dict[scale[ch]]
               except:
                  if self.type=="4015":
                     rng=self.input_range_dict['Pt100IEC-50to150']
                  else:
                     rng=self.input_range_dict['10']  # default
                  print "ADAM.py: Range value (%s) for ch #%d, type %s not found, using default (%s)!"%(
                     scale[ch],ch,self.type,rng)
            strng="$%s7C%dR%s\r"%(self.haddr,ch,rng)
            self.ioref.SendCmd(strng)

      self.eqs={}
      for ch in chs_to_enable:
         try:
            self.eqs[ch]=compile(conv_eqs[ch],'adc_eq','eval')
         except:
            self.eqs[ch]=compile("v*1.",'adc_eq','eval')
      sleep(sleeptime)
      self.enabled=chs_to_enable[:]
      self.enabled.sort()     #added 2008-01-18
      self.SetMultiplexing(self.enabled)
      self.ReadChRanges()     # read from ADAM

   def SetMultiplexing(self,chs_to_enable):
      '''
         Disable/enable channels for multiplexing (faster read).
         By default all the channels will be enabled (i.e.
         chs_to_enable=[0,1,2,3,4,5,6,7])

         Command sent: $AA5VV (VV=chs to enable as hex)
         Seems to require some wait before next cmd
      '''
      # form a decimal value from channels...
      v=0
      for ch in chs_to_enable: v+=pow(2,ch)
      if self.type=="4016":return
      # and convert it to hex
      vhex=dec2Hex(v)
      strng="$%s5%s\r"%(self.haddr,vhex)
      self.ioref.SendCmd(strng)
      self.enabled=chs_to_enable
      sleep(1.5)

   def GetReadings(self):
      '''
         Return voltage (in volts) readings for all the enabled channels
         as a dict of floats, e.g. {0:None,1:value,...,7:None}.

         Data returned by ADAM in format (no spaces???):
         +1.1111+2.2222+...(cr)
      '''
      strng="#%s\r"%self.haddr
      for i in xrange(3):
         xx=strng[:]
         s=self.ioref.SendCmd(strng)
         strng=xx
         if not s: return None
         try:
            if not self.inHex:   # == 4017,4015,4017+
               res=[float(x) for x in self.get_readings_re.findall(s)]
               for i,ch in enumerate(self.enabled):
                  res[i]=res[i]*self.toV[ch]
            else:
               res=[]
               for i in xrange(8): res.append(self.fromHex(i,s[1+i*4:1+i*4+4]))
            break
         except:
            sout="Could not parse data:'%s'"%s
            if self.logger: self.logger.warning("ADAM40xx.GetReadings: %s",sout)
            print sout
            if i==2: return None

      ret={}
      for i in xrange(8):
         v=None
         if i in self.enabled and len(res):
            v=res.pop(0)
         elif self.inHex:  # always gives values for all the chans
            res.pop(0)
         ret[i]=v

      return ret

   def GetReadingsInUnits(self):
      """
         Return all the enabled channel readings converted to
         units (not as voltages) as a dict of floats, e.g. {x:value,}
      """
      volts=self.GetReadings()
      ret={}
      for ch in volts:
         v=volts[ch]
         if v==None: continue
         ret[ch]=eval(self.eqs[ch])
      return ret


   def GetAReading(self,ch=0):
      '''
         Return voltage reading of channel ch
      '''
      if not 0 <= ch <=7:
         raise ADAM_ValueError,"Illegal ADAM Analog Input ch:%d"%ch
      strng="#%s%d\r"%(self.haddr,ch)
      resp=self.ioref.SendCmd(strng,response=True)
      if not self.inHex:
         return self.toV[ch]*float(resp[1:])
      return self.fromHex(ch,resp[1:])

   def fromHex(self,chn,hexs):
      '''
         Return value in Volts based on channel number and A/D conversion
         results as two complement (in hex)

         Revision history:
            2007-10-25 also unipolar range gives same reading as bipolar!
      '''
      i=hex2int(hexs)
      rng=self.rng[chn]
      if i<0: return -i*rng[0]/32768.
      return i*rng[1]/32767.


   def ReadChRanges(self):
      '''
         Get channel ranges to be used in fromHex()

         Set var self.ranges[]

         Revision history:
            2007-10-29 bug fix (self.rng)
      '''
      self.rng=[]
      if self.type in ("4017","4016"): # only one common range
         strng="$%s2\r"%self.haddr
         s=self.ioref.SendCmd(strng,raiseex=True)
         for i in range(8): self.rng.append(self.ch_ranges[s[3:5]])
      else:                # each ch can have individual range
         for ch in range(8):
            if self.type=="4015" and ch>5: break
            strng="$%s8C%d"%(self.haddr,ch)
            s=self.ioref.SendCmd(strng,raiseex=True)
            rr=s[6:8]
            self.rng.append(self.ch_ranges[rr])
#            print "ReadCHRanges",ch,rr
            if rr in self.mV_ranges:      # only applicable if not in hex mode
               self.toV[ch]=1e-3
            else:
               self.toV[ch]=1.


class ADAM4117(ADAM4017):
   '''
      ADAM 8 ch 16-bt ADC module (better specs than 4017)
   '''

   def __init__(self,ADAM_IO_ref,addr,scale=None,
      chs_to_enable=[0,1,2,3,4,5,6,7], conv_eqs=None,sleeptime=5.,logger=None):
      '''

      '''
      self.input_range_dict={
         "0.15":"0C","0.5":"0B","1":"0A","5":"09","10":"08",
         "10u":"48","5u":"49","1u":"4A","0.5u":"4B","0.15u":"4C","0-20mA":"4D",
         "4-20mA":"07","15u":"55"
      }
      self.type="4117"
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)
      self.ReadChRanges()

class ADAM4017p(ADAM4017):
   '''
      ADAM 8 ch 16-bt ADC module 4017+
   '''

   def __init__(self,ADAM_IO_ref,addr,scale=None,
      chs_to_enable=[0,1,2,3,4,5,6,7], conv_eqs=None,sleeptime=5.,logger=None):
      '''

      '''
      self.input_range_dict={"0.15":"0C","0.5":"0B","1":"0A","5":"09","10":"08","20mA":"0D"}
      self.type="4017p"
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)
      self.ReadChRanges()


##################################################################
## ADAM analog in/out module (+ current in, 4 digital outs)
## "Strain gauge module"
class ADAM4016(ADAM4017):
   """
   Note: ranges bipolar and in mV!

   Revision history
      2008-06-19 devel start based on ADAM4117
   """

   def __init__(self,ADAM_IO_ref,addr,scale="03",
      chs_to_enable=[0],conv_eqs=None,sleeptime=5.,logger=None):
      '''

      '''
      self.input_range_dict={"15":"00","50":"01","100":"02","500":"03","20mA":"06"}
      self.type="4016"
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)

   def SetDAC(self,voltage,setAsStartup=False):
      '''
         Set excitation voltage (0-10 V)
      '''
      if not 0 <= voltage <=10:
         raise ADAM_ValueError,"Illegal ADAM-4016 analog output voltage: %s"%voltage
      cmds="$%s7+%06.3f\r"%(self.haddr,voltage)
      self.ioref.SendCmd(cmds)
      if setAsStartup:
         cmds="$%sS\r"%self.haddr
         self.ioref.SendCmd(cmds)
      return
      
   def GetAReading(self,ch=0):
      '''
      '''
      return self.GetReadings()[0]

##################################################################
## 6-ch RTD (Pt-100 etc) input module
class ADAM4015(ADAM4017):
   """
      ADAM 6 ch RTD input module
   """
   def __init__(self,ADAM_IO_ref,addr,scale=None,
      chs_to_enable=[0,1,2,3,4,5], conv_eqs=None,sleeptime=5.,logger=None):
      '''

      '''
      self.input_range_dict={"Pt100IEC-50to150":"20",
         "Pt100IEC0to100":"21","Pt100IEC0to200":"22","Pt100IEC0to400":"23",
         "Pt100IEC-200to200":"24","Pt100JIS-50to150":"25",
         "Pt100JIS0to100":"26","Pt100JIS0to200":"27","Pt100JIS0to400":"28",
         "Pt100JIS-200to200":"29","Pt1000-40to160":"2A",
         "BALCO500-30to120":"2B","Ni604-80to100":"2C",
         "Ni6040to100":"2D"}
      self.type="4015"
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)

##################################################################
## Analog in module 4019+, supports also thermocouples
class ADAM4019p(ADAM4017):
   """
   Revision history
      2015-01-29 initial version
   """

   def __init__(self,ADAM_IO_ref,addr,scale="08",
      chs_to_enable=[0],conv_eqs=None,sleeptime=5.,logger=None):
      '''
         default scale "08": +-10V
      '''
      self.input_range_dict={
         "0.1":"02","0.5":"03","1":"04","2.5":"05","4-20mA":"07",
         "10":"08","5":"09","20mA":"0D","J":"0E",
         "K":"0F","T":"10","E":"11","R":"12","S":"13","B":"14"
      }
      self.type="4019p"
      self.InitModule(ADAM_IO_ref,addr,scale,chs_to_enable,conv_eqs,sleeptime=sleeptime,logger=logger)


##################################################################
## ADAM Relay Module
class ADAM4060:
   """
      ADAM 4 ch relay output module
   """
   def __init__(self,ADAM_IO_ref,addr):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address
      """
      self.type="4060"
      self.addr=int(addr)
      self.haddr=dec2Hex(addr)
      self.ioref=ADAM_IO_ref
      self.checksum='00'
      if self.ioref.checksum:
         self.checksum='40'
      cmds="%%%s%s4006%s\r"%(self.haddr,self.haddr,self.checksum)
      self.ioref.SendInit(cmds)
      sleep(.5)
      self.relaystatus_dict=self.DigitalDataIn()

      
   def GetRelayStatus(self):
      """
      Return current relay status as dict
      """
      return self.relaystatus_dict

   def DigitalDataIn(self):
      """
      Return status (0/1) of relay/DIN contacts as dict: {0:0/1,1:...}
      cmd: $AA6\r -> !(dataOutput)0000\r
      """
      strng="$%s6\r"%dec2Hex(self.addr)
      s=self.ioref.SendCmd(strng)
      if not s: return None
      return dictfromhex(s[1:3])

   def AllOff(self):
      """
      Turn all relays to off

      Revision history:
         2007-06-12 bug fix (was self.ioref.DigitalDataOut)
      """
      self.DigitalDataOut(self.addr,-1,0)
      self.relaystatus_dict=self.DigitalDataIn()

   def SwitchRelay(self, relstat):
      """
      Turn on/off relays.
      Dict relstat: relay number:0/1, e.g. {0:0,1:1}.
      Keeps up the status of the relays that are on/off, so that
      switching can be done more than one at the time 'correctly'.
      Raises ValueError exception if something goes wrong.

      Revision history:
         2007-06-12 bug fix (was self.ioref.DigitalDataOut)
         2011-05-24 4->8 chs
      """
      assert 0<len(relstat)<9
      for r in self.relaystatus_dict.keys(): # add key/vals missing from relaystatus_dict
         if not relstat.has_key(r): relstat[r]=self.relaystatus_dict[r]
      valtoset=sum([relstat[j]*2**j for j in range(len(relstat))])
      self.DigitalDataOut(self.addr,-1,valtoset)
      self.relaystatus_dict=self.DigitalDataIn()


   def DigitalDataOut(self,addr,channel,status):
      """
      Set/reset digital data out channels or relays, i.e.
      modules 405x/4060.

      If channel<0: ALL channels are set/reset as indicated
      by status (the 8 least significant bits).

      Command(s):
         #AA00hh     write to all channels: hh==binary comb. of chs
         #AA0n0[0/1] write to a ch n value 0,1
      """
      assert channel<8
      if channel>=0:
         s="1%d0%d\r"%(channel,(status!=0))
      else:
         s="00%s"%dec2Hex(status % 256)
      strng="#%s%s\r"%(dec2Hex(addr),s)
      s=self.ioref.SendCmd(strng)


class ADAM4069(ADAM4060):
   """
      ADAM 8 ch relay output module
   """
   def __init__(self,ADAM_IO_ref,addr):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address
      """
      ADAM4060.__init__(self,ADAM_IO_ref,addr)
      self.type="4069"
   


##########################
## ADAM Digital I/O Module
class ADAM4050(ADAM4060):
   """
   ADAM-4050 digital I/O module (7 DI + 8 DO)
      and
   ADAM-4055 digital I/O module isolated 8 DI + 8 DO

   """
   def __init__(self,ADAM_IO_ref,addr):
      """
      ADAM_IO_ref = reference to a instance of class ADAM_IO
      addr = Module address
      """
      ADAM4060.__init__(self,ADAM_IO_ref,addr)
      self.type="4050"
      self.dout_status_dict={}
      self.din_status_dict={}
      self.__GetDigitalStatus()

   def GetDigitalOutStatus(self):
      """
         Return the status of digital data out channels
      """
      return self.dout_status_dict

   def GetDigitalInStatus(self):
      """
         Return the status of digital in channels
      """
      self.__GetDigitalStatus()
      return self.din_status_dict


   def SetDigitalOut(self, status):
      """
      Turn on/off digital out. Dict status: dout number:0/1, e.g. {0:0,1:1}
      Raises ValueError exception if something goes wrong.
      """
      assert 0<len(status)<9
      for r in self.dout_status_dict.keys(): # add key/vals missing from relaystatus_dict
         if not status.has_key(r): status[r]=self.dout_status_dict[r]
      valtoset=status[0]+2*status[1]+4*status[2]+8*status[3]+16*status[4]+ \
         32*status[5]+64*status[6]+128*status[7]
      self.DigitalDataOut(self.addr,-1,valtoset)
      self.__GetDigitalStatus()

   def __GetDigitalStatus(self):
      """
      Read digital in/out status from 405x, and set the
      self.[din/dout]_status_dict {0:0/1,1:...}

      Sends command
         $AA6\r
      """
      strng="$%s6\r"%dec2Hex(self.addr)
      s=self.ioref.SendCmd(strng)
      if not s: return
      self.dout_status_dict=dictfromhex(s[1:3])
      self.din_status_dict=dictfromhex(s[3:5])




class ScanADAMs(ADAM4000):
   '''
      Scan for, and optionally change address of ADAM-4xxx module(s)
   '''
   def __init__(self,serportstr="/dev/ttyS0,9600,N,8,1",timeout=SERIAL_TIMEOUT,checksum=False):
      '''
      '''
      ini={"SerialPort":serportstr}
      self.checksum=checksum
      self.serportstr=serportstr
      ADAM4000.__init__(self,ini,timeout=.5,checksum=checksum)

   def Scan(self,startaddr="0",endaddr="99"):
      '''
         Scan for ADAM's using command
         $aaM     (module name)
         and print out the result together with firmware version
         ($aaF)
      '''
      for a in xrange(int(startaddr),int(endaddr)+1):
         cmd="$%sM\r"%dec2Hex(a)
         ret=self.SendCmd(cmd,scanmode=True)
         if ret:
            ret2=self.SendCmd(cmd[:-2]+"F\r",scanmode=True)
            print "address: %s (0x%s), module: %s  (%s)"%(
               a,dec2Hex(a),ret[3:],ret2[3:])

   def ChangeAddress(self,froma,toa):
      '''
         Change module address froma to toa

         Return True in case of success

         Revision history:
            2007-10-24  +4060,4017P
            2009-03-25  +4021
      '''
      print "\nNote! if ADAM-module has INIT/NORMAL switch (e.g. 4117,4069):"
      print "  set to INIT -> change address -> power off -> NORMAL -> should be OK (scan)"
      print
      CCs={1200:'03',2400:'04',4800:'05',9600:'06',19200:'07',38400:'08'}
      serpars=self.GetSerialParameters(self.serportstr)
      CC=CCs[serpars[1]]
      hexa=dec2Hex(int(froma)) ; hexto=dec2Hex(int(toa))
      cmd="$%sM\r"%hexa
      ret=self.SendCmd(cmd,scanmode=True)
      cs='0'
      if self.checksum: cs='4'
      if not ret:
         print "No ADAM-module found at address %s (%s)"%(froma,hexa)
      else:
         atype=ret.strip()[3:7]
         print "Changing ADAM's address from %s (%s) to %s (%s)"%(
            froma,hexa,toa,hexto)
         if atype=="4017":
            cmd="%%%s%s08%s%s0\r"%(hexa,hexto,CC,cs)
         elif atype=="4055":
            cmd="%%%s%s40%s%s0\r"%(hexa,hexto,CC,cs)
         elif atype=="4080":
            cmd="%%%s%s50%s%s0\r"%(hexa,hexto,CC,cs)
         elif atype in ("4060","4069"):
            cmd="%%%s%s40%s%s0\r"%(hexa,hexto,CC,cs)
         elif atype=="4015":
            cmd="%%%s%s00%s%s0\r"%(hexa,hexto,CC,cs)
         elif atype=="4117":
            cmd="%%%s%s00%s%s0\r"%(hexa,hexto,CC,cs)   # note: range==00
         elif atype=="4016":
            cmd="%%%s%s03%s%s0\r"%(hexa,hexto,CC,cs)
            print "  note: 4016's range set to +-500 mV (max range)"
         elif atype=="4021":     # 0-10 V DAC
            cmd="%%%s%s32%s%s0\r"%(hexa,hexto,CC,cs)
            print "  note: 4021 DAC range set to 0-10VDC"
         else:
            print "ADAM-module type %s currently not supported (add it!)"%atype
            return False
         print "  sending command: %s"%cmd
         ret=self.SendCmd(cmd)
         for i in range(3):
            sleep(2)
            cmd="$%sM\r"%hexto
            ret=self.SendCmd(cmd)
            if ret:
               print "  module %s at address %s (%s)"%(ret[3:],toa,hexto)
               return True
            else:
               print "  NO ANSWER from address %s (%s)!?"%(toa,hexto)
         return False
      return True


if __name__=='__main__':
   """
      tests...
   """
   print "\n\tADAM.py lib: ADAM.py [options] cmd"
   print "scan startaddr endaddr"
   print "change oldaddr newaddr"
   print
   DEFAULT_PORT='/dev/ttyS0,9600,N,8,1'
   from optparse import OptionParser
   parser=OptionParser()
   parser.add_option('--port',default=DEFAULT_PORT,
      help='Serial port, default: %s'%DEFAULT_PORT)
   parser.add_option('--checksum',action='store_true',default=False,
      help="Use checksum, defaul: No")
   (opts,args)=parser.parse_args()
   
   try:
      cmd=args[0]
      assert cmd in ['scan','change']
   except:
      print
      print parser.print_help()
      print
      raise

   print "serial port :",opts.port
   print "command     :",args
   scan=ScanADAMs(opts.port,checksum=opts.checksum)
   if cmd=='scan':
      scan.Scan(args[1],args[2])
   elif cmd=='change':
      scan.ChangeAddress(args[1],args[2])
   import sys
   sys.exit(0)

   import logging
   LOGFILE="./adam.log"
   print "Tests, log file:",LOGFILE
   logger=logging.getLogger("ADAM")
   hdlr=logging.FileHandler(LOGFILE)
   hdlr.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
   logger.addHandler(hdlr)
   logger.setLevel(logging.DEBUG)
   logger.info('ADAM.py logging started')

   addr=16 ; cntr=1
   aio=ADAM4000({'serial_port':'/dev/ttyS0,9600,N,8,1'},logger)
   chs=[0]
   eqs={0:"v*1."}
   rng={0:"0.15"}
   ca=ADAM4117(aio,addr,scale=rng,chs_to_enable=chs,conv_eqs=eqs)
   chs=[0] ; eqs={0:"v*1"}
   cb=ADAM4015(aio,32,chs_to_enable=chs,conv_eqs=eqs)
   import mx.DateTime as DateTime
   while 1:
      while time()%1>.1: sleep(.001)
      v=ca.GetReadings()
      t=DateTime.now()
      te=cb.GetReadings()
      print "%s %.0f %.2f"%(t,v[0]*1e6,te[0])
      sleep(9.0)



