"""
   CPC.py

   TSI CPC (3010) control/data acquisition class.

   (c)2005 Finnish Meteorological Institute, byjah

   Revision history
      2005-04-12   devel start
      2007-06-11   +CPC3772
      2007-10-02   fixes (flush=False)
      2009-05-13   3772: no more SSTOP,1/0, as it seems that
                   CPC's command interpreter hangs from time to time
      2015-08-06  + class AirmodusCPC, import mx.DateTime
"""
import sys,os
from time import time,sleep
import mx.DateTime as DateTime
try:
   import cSerial
except:
   sys.path.append(os.path.join('..','epylib'))
   import cSerial
   
NOW=DateTime.now

class CPC:
   """
      Control/read data from TSI CPC (3010)
   """
   def __init__(self,ini):
      '''
         serportstr: serial parameters
         only 3010 supported (currently)
      '''
      try:
         hw=ini['HW']
      except:
         hw=ini         # added 2006-02-02
      self.cpctype=int(hw['CPCType'])
      assert self.cpctype in [3010,3772]
      serportstr=hw['CPCSerialPort']
      self.td=None

      try:
         if self.cpctype==3010:
            self.td=float(hw['CPCTD'])*10.
            if self.td>1000: self.td/=10.
         else: #3772, sat,cond,optics temp
            self.td=[float(x) for x in hw['CPCTD'].split(',')]
      except:
         pass

      try:
         self.hvmin=float(hw['HVZero_V'])
         self.hvmax=float(hw['HVSpan_V'])
         self.hvspan=self.hvmax-self.hvmin
      except:
         self.hvmin=self.hvmax=self.hvspan=0
      self.cpcflow=1.
      try:
         self.cpcflow=float(hw['CPCFlow'])
      except:
         pass
      try:
         self.wait4ready=int(hw['CPCMaxWaitAtStart_s'])
      except:
         self.wait4ready=0
      self.ser=cSerial.cSerial(serportstr)
      self.ser.open_port()
      self.sstart=False

   def __del__(self):
      '''
      '''
      self.ser.close_port()
      self.td=None

   def initialise(self,nowait=False):
      '''
         initialises CPC by setting DAC to 0, writing temp difference
         and waiting for CPC to reach 'READY' status
         for 'CPCMaxWaitAtStart_s' seconds.

         Return 1 for 'READY', 0 for 'NOTREADY' CPC status.

         Note td=False par removed 2007-06-11
      '''
      waits=self.wait4ready
      if nowait:waits=0
      print "Initialising CPC (waiting for 'READY' max for %d s)..."%waits
      self.SetHV(0)
      if self.td:
         if self.cpctype==3010:
            tds="%s"%int(self.td+.1)
            r=self.ser.write_port('TD%s\r'%tds)  # write and get the response
            print "  setting TD%s (answer: '%s')"%(tds,r.replace('\r','\\r'))
         else:
               # saturator,condenser,optics
            for cmd,val in zip(['STS','STC','STO'],self.td):
               c="%s,%.1f"%(cmd,val)
               r=self.ser.write_port(c+'\r')
               print "  %s: %s"%(c,r)
      sta=self.ReadStatus()
      tend=time()+waits          # go th
      while time()<tend and not sta[0]: # wait for 'READY'
         sleep(2.)
         sta=self.ReadStatus()
      print "  CPC status, ready?:%d, delta_T:%.1f, vac:%d, liq:%d"%sta
      return sta[0]


   def ReadStatus(self):
      '''
         Read "Environment","Delta T", "Liquid Status", "Vacuum Status"
         from CPC and return 'em as (env_1/0,delta_t,vac_1/0,liq_1/0)
      '''
      dt=None
      while dt==None:
         if self.cpctype==3010:
            r=self.ser.write_port('RT\r')
            try:
               dt=float(r)
               break
            except KeyboardInterrupt:
               sys.exit(9)
            except:
               print "ReadStatus:'%s'"%r #debug
               sleep(.2)
               pass
         else: #3772
            sat=self.ser.write_port('RTS\r')
            con=self.ser.write_port('RTC\r')
            try:
               dt=float(sat)-float(con)
            except:
               dt=0.
      r=self.ser.write_port('R5\r')
      rdy=0
      if r.find('READY')>=0 and r.find('NOTREADY')<0: # found READY
         rdy=1

      # NOTREADY -> read liq,vac
      liq=self.ser.write_port('R0\r')
      if liq.find('FULL')>=0 and liq.find('NOTFULL')<0:
         liq=1
      else:
         liq=0

      if self.cpctype==3010:
         vac=self.ser.write_port('RV\r')
         if vac.find('VAC')>=0 and vac.find('LOVAC')<0:
            vac=1
         else:
            vac=0
      else: #3772
         errs=self.ser.write_port('RIE\r')
         try:
            ei=int(errs)
            if ei&8 or ei&16: # inlet flow rate, aerosol flow rate
               vac=0
            else:
               vac=1
         except:
               vac=0

      return (rdy,dt,vac,liq)


   def ReadCounts(self):
      '''
         Read and return cumulative counts as
         (time,counts), (-1,-1) in case of some error.

         Note: time returned by the CPC in in full seconds
         (i.e. not very usable for short time periods)!
      '''
      r=self.ser.write_port('DC\r')
      try:
         vals=[float(x) for x in r.split(',')]
      except:
         return (-1,-1)
      if len(vals)<2:
         print "CPC DC return:",r
         return (-1,-1)
      return vals

   def SetHV(self,voltage):
      '''
         Set the CPC DA output so, that the HV supply voltage is "voltage".
         HV supply control voltage range is assumed to be 0-10 V corresponding
         to whole HV supply range (for example 0 - 12500 V).

         Returns (setmV,voltage)

         CPC 3010 DA range: 0-11 V (V0,V11000), send as millivolts to CPC
            V0...V11000
         CPC 3772: 0.000-10.000 V (SVO,v.vvv

         Doesn't allow HV voltage to be set over 12000V.
      '''
      if voltage>12000: voltage=12000
      try:
         v=int((voltage-self.hvmin)/self.hvspan*10000)
      except:
         v=0
      if v<0: v=0
      if self.cpctype==3010:
         self.ser.write_port('V%d\r'%v)
      else:
         vf="%.3f"%(float(v)/1000.,)
         self.ser.write_port('SVO,%s\r'%vf)
      return (v,voltage)

   def StartMeas(self):
      '''
         Start CPC measurements - empty buffer by reading, store the time
      '''
      self.startt=time()
      if self.cpctype==3010:
         self.ReadCounts()
      else:
         if not self.sstart:
            self.ser.write_port('SSTART,0\r')
            sleep(.02)
            self.ser.read_port()    # empty buffer
            # starts to output in 1 second increments a lot of data
            for i in range(4):
               ret=self.ser.write_port('SSTART,1\r')
               if ret.find("OK")>=0: break
               sleep(i/100.+.01)
            self.sstart=True
         else:
            s=self.ser.read_port()    # empty buffer
            if s and s[-1]!='\r':
               while s[-1]!='\r':
                  if self.ser.inWaiting():
                     s=self.ser.read_port()
                  else:
                     sleep(.001)
                  
         self.startt=time()

   def GetFastConc(self):
      '''
         In "fast mode": StartMeas -> GetFastConc
         Wait until data, time = timestamp-1
         Return:
            [time,c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,caverage]
         or None if error

         Revision history:
            2007-09-21 first version
      '''
      flowfactor=self.cpcflow*.1*1000./60.
      if self.cpctype!=3772:
         raise "Only TSI CPC 3772 supported"

      while not self.ser.inWaiting(): sleep(.01)
      t=time()-1.01
      sleep(.01)
      data=self.ser.read_port()
      s=data.split(',')
      if len(s)<23:
         print "Too short data line from CPC:'%s'"%data
      c=[]
      try:
         for ind in xrange(1,11): c.append(int(s[ind])/flowfactor)
      except:  # partial data line or something
         print "Illegal data (ser.port?):'%s'"%data
         return None
      return [t]+c+[sum(c)/len(c)]

   def QuitMeas(self):
      '''
         Stop measuring (3772 in 'fast' mode)
      '''
      if self.cpctype!=3772: return
      self.ser.write_port('SSTART,0\r')
      self.sstart=False
      return


   def EndMeas(self,meastime_s,CPCflow=-1,counts=False,reallyEnd=False):
      '''
         Wait till meastime_s has elapsed, read the CPC and return
         particle number concentrations as #/cm3, or as counts/s,
         if counts==True

        !!!!!!!!!   NOTE   !!!!!!!!!!
        Under WIN32 buffer is 4k, linux:???
        -> max ca 20 s worth of data (if more -> change how data is read)
        !!!!!!!!!!  NOTE   !!!!!!!!!!
      '''
      cpcflow=CPCflow
      if not cpcflow>0: cpcflow=self.cpcflow
      tend=self.startt+meastime_s
      
      if self.cpctype==3010:
         while time()<tend: sleep(.01)
         enddt=time()
         t,c=self.ReadCounts()
         if counts:
            return c/(enddt-self.startt)
         conc=c/(enddt-self.startt)/(1000./60*cpcflow)
         
      else: #3772: data is coming once/sec
         bu=[]
         while True:
            enddt,da=self.ReadLines()
            bu.append(da)
            if enddt+.1>=tend: break
         dalines=''.join(bu).split('\r')
         if reallyEnd: self.QuitMeas()
         
         # go through all the data lines
         dalst=[]
         for d in dalines:
            if not d: continue
            s=d.split(',')
            if len(s)<23: 
               if d.find('OK')<0:
                  print "Strange data (# fields: %s, should be %s):'%s' - SKIPPED"%(
                     len(s),23,d)
               continue
            try:
               dalst.extend([int(s[x]) for x in xrange(1,11)])
            except:  # partial data line or something
               continue
               
         lenda=len(dalst)
         ein=lenda-int((enddt-tend)*10)
         sin=ein-int(10*meastime_s)
         if sin<0: sin=0
         eda=dalst[sin:ein]
         totc=sum(eda)
         tots=len(eda)/10.
#         print "cnts,time,len",totc,tots,len(eda),
         if counts: return totc/tots
         conc=totc/tots/(1000./60*cpcflow)
      return conc

   def ReadLines(self):
      '''
         Read a data line(s) from 3772 (ending at \r ONLY!)
         Return datatime,data
      '''
      ret=[]
      if self.ser.inWaiting():
         ret=[self.ser.read_port()]
         while ret[-1][-1]!='\r':
            sleep(.002)
            if self.ser.inWaiting():
               ret.append(self.ser.read_port())
      da=''
      while da=='':
         da,t=self.ser.wait_for_data()
      ret.append(da)
      while ret[-1][-1]!='\r':
         sleep(.002)
         if self.ser.inWaiting():
            ret.append(self.ser.read_port())
      return t,''.join(ret) 


class AirmodusCPC:
   """ 
      Control/read data from AirMocus CPC
   """
   
   def __init__(self,ini,opts):
      '''
         serportstr: serial parameters
      '''
      try:
         self.hw=ini['HW']
      except:
         self.hw=ini
      self.opts=opts
      self.td=None
#      self.ser=cSerial.cSerial(self.hw)
#      self.ser.open_port()
      self.ser=None
      self.initialise()
      # status read [dt,value]
      self.stat=[None,None]
      
   def initialise(self,nowait=False):
      '''
         initialises CPC
      '''
      print "Initialising CPC..."
      for k in self.hw:
         if k[:5]!='.SET.': continue
         kk=k.replace('.',':')
         cmd="%s %s"%(kk,self.hw[k])
         print "  %s"%(cmd,)
      n=NOW()
      for k,v in zip(('YY','MON','DD','HH','MM','SS'),
         (n.year%100,n.month,n.day,n.hour,n.minute,int(n.second))):
         cmd=":SET:TIME:%s %s"%(k,v)
         print "  %s"%(cmd,)
         
      
         
#      sta=self.ReadStatus()
#      tend=time()+waits          # go th
#      while time()<tend and not sta[0]: # wait for 'READY'
#         sleep(2.)
#         sta=self.ReadStatus()
#      print "  CPC status, ready?:%d, delta_T:%.1f, vac:%d, liq:%d"%sta
#      return sta[0]
      
   def ReadStatus(self):
      '''
         Return result from command :STAT:SYST, also update
         
         bit coded (bits 0-8)
      '''
      dt=NOW()
      r=self.ser.write_port(':STAT:SYST\r')
      try:
         v=int(r)
      except:
         v=None
      self.stat=[dt,v]
      return v
      
   def ReadConc(self):
      '''
         Return results from command :MEAS:OPC
      '''
      dt=NOW()
      r=self.ser.write_port(':MEAS:OPC\r')
      return r



def main():
   '''
      Tests
   '''
   print "\n\tepylib/CPC.py tests\n"
   ini={'CPCType':3772,'CPCSerialPort':'COM3,115200,N,8,1',
      'CPCTD':'40,21,41'}
   print ini
   cpc=CPC(ini)
   cpc.initialise()
   st=time()
   cpc.StartMeas()
   sleep(3)
   print cpc.EndMeas(3)
   print

if __name__ == '__main__': main()
