# -*- coding: utf-8 -*-
"""
   cMFC.py
   
   Control/read MKS MFC GE50A
   
   Revision history
      2017-07-10  Initial version
"""
from time import sleep
from cSerial_v2 import cSerial
from epylib import FormatExceptionMessage

############################################################
##
##    Main thread class
##
############################################################
class cMFC(object):
   """
      MFC control/read
   """
   def __init__(self,
   ini,               # ini-data for MFC in dictionary
   logger=None
   ):
      """
         Open serial port, initialise instrument(s)
         ini dct:
            'type':'GE50A'
            'serial_port':'port',speed,n/e/o,8,1
            'addresses':['addr']
            'addr':initialsetpoint
      """
      self.ini=ini
      self.logger=logger
      
      self.mfctype=ini['type']
      assert self.mfctype in ("GE50A")
      self.addrs=ini['addresses'].split(',')
      self.mfcs={}
      for a in self.addrs:
         if a not in ini:
            msg="MFC with address %s has no parameters!?"
            continue
         self.mfcs[a]=ini[a]
      self.cserial=None
      self._openPort()
      
   def __repr__(self):
      ret="MFC type: %s\n  addresses:%s\n"%(self.mfctype,self.addrs)
      for a in self.addrs:
         ret+="  %s: %s\n"%(a,self.mfcs[a])
      return ret
   
   def _openPort(self,reopen=False):
      '''
      '''
      if reopen and self.cserial:
         self.cserial.close_port()
      elif self.cserial:   # already open
         return
         
      try:
         self.cserial=cSerial(self.ini['serial_port'],timeout=2.,serid='MFC')
         self.cserial.open_port()              # open the serial port
      except:
         msg=FormatExceptionMessage("cMFC.__init__:")
         print msg
         if self.logger:
            self.logger.error(msg)


   def ReadData(self):
      '''
         Read data for all the MFCs and return them as a dict:
         {'addr':
            {
               'flow':flow,'intT':temp,
               'vdriver':valve_driver_level
            }
         }
      '''
      retdct={}
      for a in self.mfcs:
         retdct[a]={}
         for cm,key in zip(('FX?','TA?','VD?'),('flow','intT','vdriver')):
            for n in range(2):      # try  twice
               resp=self._sendCmd(a,cm)
               if not resp:
                  sleep(.1)
                  continue    # try again
               try:
                  retdct[a][key]=float(resp)
               except:
                  retdct[a][key]=resp
               break
      return retdct
      
   def GetStatus(self):
      '''
         Get status code for all the MFCs and return them as a dict:
         {'addr':'status'}
      '''
      retdct={}
      for a in self.mfcs:
         for n in range(2):      # try  twice
            resp=self._sendCmd(a,'T?')
            if not resp:
               sleep(.1)
               continue    # try again
            retdct[a]=resp
      return retdct


   def InitMFCs(self):
      '''
         For each:
            set softstartrate to 50 (50*0.032 s), SS
            set initial setpoint
      '''
      for a in self.mfcs:
         self._sendCmd(a,"SS!50")
         sleep(.1)
         self._sendCmd(a,"SX!%s"%(self.mfcs[a],))
      
   def _sendCmd(self,addr,cmd):
      '''
         Send the command, get and return response data
      '''
      if not self.cserial:
         self._openPort()
         if not self.cserial: return
         
      cmds="@%s%s;"%(addr,cmd)
      chks=self._chksum(cmds)
      cmdstr='@@%s%s'%(cmds,chks)
      resp=self.cserial.write_port(cmdstr,waitForResponse=.05) # 0.02 seems to enough
      ackok=resp.find('ACK')>0
      msgs="cMFC._sendCmd(): %s -> %s"%(cmdstr,resp)
      if self.logger:
         if ackok:
            self.logger.debug(msgs)
         else:
            self.logger.error(msgs)
      if not ackok:
         print "ERROR/NO response:",msgs
         return None
      rchks=self._chksum(resp[:-2])
      if rchks!=resp[-2:]:
         msgs="cMFC._sendCmd(): %s -> %s - CHKSUM error"%(cmdstr,resp)
         if self.logger:
            self.logger.warning(msgs)
         print msgs
         return None
      ack=resp.find('ACK')
      return resp[ack+3:-3]
      
      
   def _chksum(self,str):
      '''
         calc and return checksum
         (two last characters of hex value for the sum(ord(chars))
      '''
      sum=0
      for x in str: sum+=ord(x)
      return hex(sum).upper()[-2:]
      
      


if __name__ == '__main__':
   ini={'type':'GE50A','addresses':'021','021':'222','serial_port':'/dev/ttyS0,19200,n,8,1'}
   mfc=cMFC(ini)
   print mfc
   print "--- init ---"
   mfc.InitMFCs()
   print "--- read ---"
   print mfc.ReadData()
   print "-- status ---"
   print mfc.GetStatus()
   

