# -*- coding: utf-8 -*-
"""
   Serial port handling routines (class cSerial).
   Writing/reading is lock protected.

   Requires package pyserial

   Revision history:
      2006-08-29  check for 8,n
      2007-10-02  flush=True to write_port
      2008-09-04  +wait_for_data()
      2013-01-03  latin-1 -> utf-8
      2014-02-11  major update (+telnetlib): copy from PyBARFluxInc.py
      2015-03-26  update to write_port response handling (eol checking), add readline()
      2016-05-11  "telnet" (port): better handling of dropped connections
"""
import traceback,sys
import socket,select
import serial
from time import sleep,time
import thread


def FormatExceptionMessage(string):
   """
      Format exception to a one-liner with optional prefix (string)
   """
   try:
      s=string+''.join(traceback.format_exception(
         *sys.exc_info())[-2:]).strip().replace('\n',': ')
   except:
      s="ERROR in FormatExceptionMessage!!!!"
   return s
   
class cSerialException(Exception):
   def __init__(self,value):
      self.value=value
   def __str__(self):
      return repr(self.value)

class cSerial:
   '''
      Serial/socket port handling (RW)
   '''

   def __init__(self,portstr,timeout=.2,logger=None,mid='cSerial'):
      '''
         porstr: ser_port,speed,parity ('N'/'E'/'O'),databits(5-8),stopbits(1,2)
            where ser_port is passed on directly to serial (i.e. can be
            a number of dev name)
         if logger instance is given,
            logs the error/warning/debug messages
      '''
      self.timeout=timeout
      self.serial=None
      self.logger=logger
      self.mid=mid
      self.istelnet=False
      self.buff=''
      self.portstr=portstr
      self.portpars=self.GetSerialParameters(portstr)

   def SetLogger(self,logger):
      '''
         Set/remove logger/logging
         2010-10-27
      '''
      self.logger=logger

   def GetSerialParameters(self,strng):
      """
         Return (portname,bps,parity,databits,stopbits) from string
         ser_port,speed,parity ('N'/'E'/'O'),databits (5-8),stopbits (1,2)

         Revision history:
            2005-03-05  added parity,databits,stopbits and asserts
      """
      if strng.find(':')>0:      # IP port
         s=strng.split(':')
         if len(s)!=2:
            raise cSerialException("Wrong format IP:port, should be ipaddr:port")
         self.istelnet=True
         return (s[0],int(s[1]),None,None,None)
      s=strng.split(',')
      if len(s)!=5:
         ss="Wrong format serial param. string '%s', should be:"\
            "'dev,speed,parity,databits,stopbits'"%strng
         raise cSerialException(ss)
      bps=int(s[1])
      assert bps in [300,600,1200,1800,2400,4800,9600,19200,38400,57600,115200,
                    230400,460800,500000,576000,921600]
      parity=s[2].upper()
      assert parity in ['N','E','O']
      dbits=int(s[3])
      assert dbits in [5,6,7,8]
      sbits=int(s[4])
      assert sbits in [1,2]
      return (s[0],bps,parity,dbits,sbits)

   def __del__(self):
      'Delete (close) the port'
      try:
         self.close_port()
      except:
         pass
         
   def __repr__(self):
      ''
      if self.serial:
         return self.serial.__repr__()
      return ''

   def close_port(self):
      'Close the serial port or socket if open'
      if self.serial:
         if self.logger:
            self.logger.info('%s.close_port()',self.mid)
         self.serial.close()
         self.serial=None

   def open_port(self):
      '''
         Open the serial port. If port allready open close it first
         ("reopen"). In case of error write log message (if enabled)
         and reraise exception.
      '''
      if self.logger:
         self.logger.info('%s cSerial.open_port(%s)',self.mid,self.portstr)
      if self.serial:
         self.serial.close()
      if self.istelnet: #IP-port
         try:
            self.serial=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            self.serial.connect((self.portpars[0],self.portpars[1]))
            if self.logger:
               self.logger.info('%s cSerial.open_port(%s) OK',self.mid,self.portstr)
         except:
            sleep(5)
            try:
               self.serial.connect((self.portpars[0],self.portpars[1]))
            except:
               if self.logger:
                  self.logger.error("%s.open_port: no luck opening port",self.mid)
               self.serial=None
         return
      try:
         # port,baudrate=9600,bytesize=8,parity='N' ('E','O'),stopbits=1 (2),
         # timeout=None,xonxoff=0,rtscts=0,writeTimeout=None
         self.serial=serial.Serial(self.portpars[0],baudrate=self.portpars[1],
            parity=self.portpars[2],bytesize=self.portpars[3],
            stopbits=self.portpars[4],timeout=self.timeout,
            writeTimeout=self.timeout)
      except:
         if self.logger:
            self.logger.error(FormatExceptionMessage("cSerial.open (%s), timeout:%s, id=%s:"%(
               self.portpars,self.timeout,self.mid)))
         print "\n\nERROR in cSerial.open: %s, id=%s\n"\
               "Check log message for the error (if enabled)"%(
            self.portpars,self.mid)
         raise cSerialException("%s.open_port: ERROR"%self.mid)

   def inWaiting(self,nobuff=False):
      'Read data to buffer, return number of characters waiting in buffer'
      if not self.serial:
         self.open_port()
         if not self.serial: return 0
      da=''
      try:
         if self.istelnet:
            sss=select.select([self.serial],[],[self.serial],0)
            if sss[2]:
               self.close_port()
               return 0
            if sss[0]:
               da=self.serial.recv(1)
               if not da:  # connection dropped
                  self.close_port()
                  return len(self.buff)
               self.buff+=da
         else:
            n=self.serial.inWaiting()
            if n>0:
               da=self.serial.read(n)
               self.buff+=da
      except:
         self.close_port()
      if nobuff:
         return len(da)
      return len(self.buff)

   def sendBreak(self):
      '''
         added 2006-12-28 for LI-7500
      '''
      if self.istelnet: return      # no break with telnet
      if not self.serial:  self.open_port()
      return self.serial.sendBreak()


   def read_port(self,line=False,timeout=2.):
      '''
         Read and return everything or a line (\n, or \r). If reading
         a line, it is waited for timeout time.
      '''
      s=''
      try:
         x=self.inWaiting()
         if not self.serial: return 0
         if line:
            t=time()+timeout
            while time()<t:
               if self.istelnet:
                  self._read_socket(timeout=timeout,eol='\r\n')
                  s=self._buffline(wholeline=True)
                  if s: break
               else:
                  s=self._buffline(wholeline=True)
                  if s: break
                  sleep(.02)
                  x=self.inWaiting()
         else:
            if self.istelnet:
               self._read_socket()      # get all data in buffer
            s=self.buff
            self.buff=''      # !!!
         if self.logger:
            self.logger.debug("cSerial.R,id=%s:'%s'"%(self.mid,
               s.replace('\n','\\n').replace('\r','\\r').replace('\t','\\t')))
      except:
         raise #debug"
         pass

      return s

   def _buffline(self,wholeline=False):
      '''
         Return first data line (ending with '\n', and/or '\n')
         in self.buff and remove it from self.buff

         If whole line not found, and wholeline==True, don't adjust the buffer
         and return ''
      '''
      s=None
      i=self.buff.find('\n')
      if i<0:
         i=self.buff.find('\r')
      if i>=0:
         s=self.buff[:i+1]
         self.buff=self.buff[i+1:]
      if s: return s
      if wholeline: return ''
      s=self.buff
      self.buff=''
      return s


   def _read_socket(self,timeout=0.,eol=None):
      '''
         Read data from socket until buffer empty (eol==None)
         OR
         eol encountered
         Store read data to self.buff
      '''
      if not self.serial: return
      otout=self.serial.gettimeout()
      toutchanged=False
      if otout!=timeout:
         self.serial.settimeout(timeout)
         toutchanged=True
      chs=[]
      while True:
         try:
            s=self.serial.recv(1)
         except:
            break
         if not s:      # connection down (dropped)
            self.close_port()
            break
         chs.append(s)
         if eol and s in eol:
            break
      self.buff+=''.join(chs)
      if toutchanged:
         self.serial.settimeout(otout)


   def write_port(self,strn,response=True,eol='\r'):
      '''
         Write str and wait and return response (if response==True)
         if it comes within timeout (set in __init__). EOL handling???

         Before writing empty the serial port buffer and write it to log
         (if enabled) with level WARNING.
         
         Revision history:
            2014-02-12  do not log empty string (when emptying buffer)
            2015-03-26  update to response eol checking
      '''
      if not strn: return ''
      s=[]
      try:
         if not self.serial: self.open_port()
         if not self.serial: return
         ss=self.read_port()
         if self.logger:
            if ss:
               self.logger.warning('cSerial.W,id=%s,buffer:"%s"'%(
                  self.mid,self._convertStr(ss)))
               self.logger.debug('cSerial.W,id=%s:"%s" (l==%d)'%(self.mid,
               self._convertStr(strn),len(strn)))

         if not self.istelnet:
            self.serial.write(strn)
         else:
            self.serial.send(strn)

         if response:
            tend=time()+self.timeout
            while time()<=tend:
               sleep(.01)
               if self.inWaiting()>0:
                  sleep(.01)
                  s.append(self.read_port())
                  if s[-1].find(eol)>0: break
      except:
         pass
      return ''.join(s)


   def _convertStr(self,strn):
      '''
         Convert "\n","\r" to "\\n","\\r" in strn
      '''
      return strn.replace('\n','\\n').replace('\r','\\r')

   def FindDataStart(self,firstval=(None,' '),difft=0.):
      '''
         Find the time data arrives to port and return it.
         Uses readline to empty the buffer, and waits for the NEXT
         data line. If difft>0., wait reading until time between
         two consecutive lines are >= than difft.

         Return (datastart,dataline)

         Revision history:
            2006-02-02  parameter difft
      '''
      if not self.serial:
         self.open_port()
         if not self.serial:
            return (None,None)
      self.read_port(line=True)  # read and discard first line
      prevtm=None
      # now wait for the first character
      while True:
         if self.inWaiting()>0:
            tm=time()
            ss=self.read_port(line=True)
            if firstval[0]:
               s=ss.split(firstval[1],2)
               if s[0]!=firstval[0]: continue
            if difft:
               if prevtm:
                  if tm-prevtm>=difft: break
               prevtm=tm
            else:
               break
         sleep(.001)
      return tm,ss

   def wait_for_data(self,timeout=-1):
      '''
         Wait for 1st character to arrive (within timeout).
         If there is data allready available, read it and
         return next char time together with data - ONE LINE MAX

         Return (characters,time_in_ticks)

         In case of timeout, exception from serial is passed through

         Revision history:
            2008-09-24  First version
            2009-05-09  Also work with telnet connection
            2009-05-29  If allready data available, return next char t
            2010-11-15  bug fix (only one data line)
      '''
      tu=(None,None)
      if not self.serial:
         self.open_port()
         if not self.serial: return tu
      if not self.istelnet:
         if timeout>=0:
            self.serial.timeout=timeout
         try:
            ch=self.serial.read(1)
            t=time()
            if ch:
               self.buff+=ch
            tu=(self._buffline(),t)
         finally:
            if timeout!=self.timeout:
               self.serial.timeout=self.timeout
      else:
         if timeout>=0:
            tout=timeout
         else:
            tout=self.timeout
         # on timeout returns three empty lists
         sss=select.select([self.serial],[],[self.serial],tout)

         t=time()
         if not sss[0] and not sss[2]:
            if self.logger:
               self.logger.warning("%s.wait_for_data: socket timeout",self.mid)
            self.close_port()
         elif sss[2]:
            if self.logger:
               self.logger.warning("%s.wait_for_data: socket error",self.mid)
            self.close_port()
            return tu
         elif sss[0]:
            tu=(self.serial.recv(1),t)
            if not tu[0]:     # connection has been closed
               self.close_port()
               if self.logger:
                  self.logger.warning("%s.wait_for_data: no data from socket",self.mid)

      return tu



   def readline(self):
      '''
         pass on to pySerial
      '''
      if not self.serial: self.open_port()
      return self.serial.readline()
