# -*- coding: utf-8 -*-
"""
   Serial/TCP/zmq port handling routines (class cSerial).
   Also can handle some (LI-7000) USB communication.

   Requires packages pyserial, pyzmq, pyusb

   Config, i.e. "port string":
      io_port_name,speed,parity ('N'/'E'/'O'),databits(5-8),stopbits(1,2)
      IP:port     - open "telnet" connection (two way)
      IP:port,str - open zmq STREAM port (R/W), can be used to get data from "normal" socket
      IP:port,sub - open zmq subscribe port (read only)
      USB,idVendor,idProduct,config#,write_endpoint,control_read_endpoint,data_read_endpoint


   Revision history:
      2016-11-30  first version based on cSerial.py and PyBARFlux/PyBARFluxInc.py
      2017-06-20  minor enhancements
      2017-12-04  zmq PUB type added

   TODO:
      - two way ZMQ communication with Router(master)/Dealer (?)
        (should work now with zmq.STREAM type sockets)
      - a buffer for each USB read endpoint? - currently not needed
      - thread'ify this - maybe not: if needed make separate module using cSerial
"""
import traceback,sys
import socket
import time
import serial
import zmq
import usb


def FormatExceptionMessage(msgstr,printmsg=False):
   """
      Format exception to a one-liner with optional prefix (string)
   """
   try:
      s=msgstr+''.join(traceback.format_exception(
         *sys.exc_info())[-2:]).strip().replace('\n',': ')
   except:
      s="ERROR in FormatExceptionMessage!!!!"
   if printmsg:
      print "!!!\n%s\n!!!"%(s,)
   return s


class cSerial(object):
   '''
      Serial/TCP/zmq port handling
   '''

   def __init__(self,portstr,timeout=.2,logger=None,serid='cSerial'):
      '''
         Where portstr can be
            ser_port_name,speed,parity ('N'/'E'/'O'),databits(5-8),stopbits(1,2)
            IP:port     - open "telnet" connection (two way)
            IP:port,str - open zmq STREAM port (R/W), can be used to get data from "normal" socket
            IP:port,sub - open zmq subscribe port/socket (read only)
            IP:port,pub - open zmq publish port/socket (write only)
            or     = USB,idVendor,idProduct,bConfigurationValue,bEndpointAddress_IN,
                     bEndpointAddress_OUT
         Note USB,... only tested with LI-7000. Values can be found e.g. with lsusb -v

         self.istelnet == False, self.isusb == False if serial port
         self.isusb == True, if usb
         self.istelnet == True, and
            self.zmqtype==0  normal socket port (does not use zmq)
            self.zmqtype==1  STReam socket (R/W, for "normal" socket data)
            self.zmqtype==2  SUBsribe socket (read only)
            self.zmqtype==3  PUBlish socket (write only)
      '''
      # self.serial: "pointer" to serial or socket
      self.timeout,self.serial,self.logger=timeout,None,logger
      self.portstr,self.serid,self.zmqtype=portstr,serid,0
      self.istelnet,self.context=False,None
      self.isusb,self.usbpacketsizedct,self.usbreadendpoint=False,{},None
      # also updates self.istelnet,self.zmqtype,self.isusb,self.usbreadendpoint
      self.portpars=self._parse_serial_parameters(portstr)
      # characters in buffer(s), list of buffers
      self.bufflen=0
      self.buff=[]

   def set_usb_endpoint_to_data(self,todata=True):
      '''
         By default (todata==True) set USB read endpoint to data (idx -1),
         otherwise set it to control (idx -2)
      '''
      if self.isusb:
         self.usbreadendpoint=self.portpars[-1] if todata else self.portpars[-2]
   
   def find_usb(self,configno=None,current=False):
      '''
         Find and return print USB device data based on idVendor,idProduct
         
         Note: does not work with identical devices
      '''
      if not self.isusb:
         print "\ncSerial.find_usb(%s,%s): Not configured (%s) as USB device!?"%(configno,current,self.portstr)
         return
      if current and not self.serial:
         print "\nUSB device not defined, i.e. there is no current configuration"
         return
      if current:
         dev=self.serial
      else:
         dev=usb.core.find(idVendor=self.portpars[0],idProduct=self.portpars[1])
         if dev is None: return None
      print "\nUSB device, idVendor:%s (%s), idProduct:%s (%s)"%(dev.idVendor,
         hex(dev.idVendor),dev.idProduct,hex(dev.idProduct))
      hexlst=['0x%04x'%(x,) for x in self.portpars]
      print "cSerial() port definition: %s\n\t\t   in hex: %s"%(
         self.portpars,hexlst)
      print "Manufacturer:",dev.manufacturer
      print "Product:",dev.product
      print "Serial: ",dev.serial_number
      print "Kernel driver active: ",dev.is_kernel_driver_active(0)
      if configno is not None:
         devs=[dev[configno-1]]
      else:
         devs=dev
      try:
         print "Active configuration:\n",dev.get_active_configuration()
         if current: return
      except:
         print "No active configuration (or other error)"
         
      for cfg in devs:
         print "\nbConfigurationValue: %s, %s"%(cfg.bConfigurationValue,
            usb.util.get_string(dev,cfg.iConfiguration,None))
         for intf in cfg:
            print "\tbInterfaceNumber:%s, bAlternateSetting:%s"%(intf.bInterfaceNumber,intf.bAlternateSetting)
            for ep in intf:
               epa=ep.bEndpointAddress
               dirs="OUT" if epa&128 else "IN"
               eps=epa&15
               v=ep.bmAttributes&3
               if v==0:
                  tys="Control"
               elif v==1:
                  tys="Isochronous"
               elif v==2:
                  tys="Bulk"
               else:
                  tys="Interrupt"
               print "\t\tbEndpointAddress: %3d (0x%02x),  EP %s %3s, %s, max packet size:%s"%(
                  epa,epa,eps,dirs,tys,ep.wMaxPacketSize)

   def set_logger(self,logger):
      '''
         Set/remove "pointer" to logger (logging)
      '''
      self.logger=logger

   def get_logger(self):
      '''
         Return self.logger
      '''
      return self.logger

   def _parse_serial_parameters(self,portstr):
      """
         Parse port string,
         set self.istelnet, self.zmqtype
         return self.portpats[]
      """
      if portstr[:4].upper()=='USB,':
         self.isusb=True
         si=[int(x,0) for x in portstr.split(',')[1:]]   # ,0 means "guess radix"
         assert len(si)==6
         self.usbreadendpoint=si[-1]            # by default data ep
         return si
      if portstr.find(':')>0:      # IP port
         s=portstr.split(':')
         if len(s)!=2:
            raise AttributeError("cSerial.__init__(%s): Wrong format IP:port, should be ipaddr:port[,sub/str]",self.serid)
         ss=s[1].split(',')
         self.istelnet=True
         sport=s[1]
         if len(ss)==2:
            sport=ss[0]
            zmqt=ss[1].upper()
            if zmqt not in ('SUB','STR','PUB'):
               raise ValueError("ZMQ socket type must in (SUB,STR,PUB) (not %s) in %s",ss[1],self.serid)
            if zmqt=='STR': 
               self.zmqtype=1
            elif zmqt=='SUB':
               self.zmqtype=2
            else:
               self.zmqtype=3
         return (s[0],int(sport),None,None,None)
      s=portstr.split(',')
      if len(s)!=5:
         ss="Wrong format serial param. string '%s', should be:"\
            "'dev,speed,parity,databits,stopbits' in %s (for serial port)"%(portstr,self.serid)
         print ss
         if self.logger: self.logger.error(ss)
         raise ValueError(ss)
      bps=int(s[1])
      assert bps in [300,600,1200,1800,2400,4800,9600,19200,38400,57600,115200,
      230400,460800,500000,576000,921600]
      parity=s[2].upper()
      assert parity in ['N','E','O']
      dbits=int(s[3])
      assert dbits in [5,6,7,8]
      sbits=int(s[4])
      assert sbits in [1,2]
      return (s[0],bps,parity,dbits,sbits)

   def __del__(self):
      'Delete (close) the port'
      try:
         self.close_port()
      except:
         pass

   def is_port_open(self):
      'Return True if open'
      return self.serial is not None

   def is_usb(self):
      'Return True in case of USB communication'
      return self.isusb

   def __repr__(self):
      'for print'
      if self.serial:
         if not self.istelnet and not self.isusb: 
            return self.serial.__repr__()
         opens="OPEN" if self.is_port_open() else "CLOSED"
         return self.portstr+" : "+opens
      return self.portstr+" : NOT OPEN(ED)"

   def open_port(self):
      '''
         Open the port. If port allready open close it first
         ("reopen"). In case of error write log message (if enabled)
      '''
      if self.logger:
         self.logger.info('%s cSerial.open_port(),%s',self.serid,self.portstr)
      if self.serial: 
         self.close_port()

      if self.isusb:
         try:
            self.serial=usb.core.find(idVendor=self.portpars[0],idProduct=self.portpars[1])
            if not self.serial:
               raise IOError("Could not find USB instrument")
            self.serial.set_configuration(self.portpars[2]) 
            cfg=self.serial[self.portpars[2]-1]    # current configuration
            for intf in cfg:                       
               for ep in intf:
                  self.usbpacketsizedct[ep.bEndpointAddress]=ep.wMaxPacketSize
               break                               # only the first interface
         except:
            emsg=FormatExceptionMessage("%s.open_port: USB device (idVendor:%s, idProduct:%s), config# %s!?"%(
               self.serid,self.portpars[0],self.portpars[1],self.portpars[2]),printmsg=True)
            if self.logger: self.logger.error(emsg)
            if self.serial: 
               self.close_port()
         return


      if self.istelnet: #IP-port
         try:
            if self.zmqtype==0:     # normal "telnet"
               self.serial=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
               self.serial.connect((self.portpars[0],self.portpars[1]))
               self.serial.setblocking(0)    # do not block
            else:
               if not self.context:
                  self.context=zmq.Context()
               soctype=zmq.PUB
               if self.zmqtype!=3: 
                  soctype=zmq.SUB if self.zmqtype==2 else zmq.STREAM
               self.serial=self.context.socket(soctype)
               if self.zmqtype==3:
                  self.serial.bind("tcp://%s:%s"%(self.portpars[0],self.portpars[1]))
               else:
                  if self.zmqtype==2: 
                     self.serial.setsockopt(zmq.SUBSCRIBE,"")
                  self.serial.set_hwm(50)   # pylint: disable=E1101
                  self.serial.connect('tcp://%s:%s'%(self.portpars[0],self.portpars[1]))
            if self.logger:
               self.logger.info('%s cSerial.open_port sock, %s',self.serid,self.portstr)
         except:
            if self.logger:
               self.logger.error("%s.open_port: no luck opening port",self.serid)
            self.close_port()
         return
      try:
         # port,baudrate=9600,bytesize=8,parity='N' ('E','O'),stopbits=1 (2),
         # timeout=None,xonxoff=0,rtscts=0,writeTimeout=None
         self.serial=serial.Serial(
         self.portpars[0],baudrate=self.portpars[1],
         parity=self.portpars[2],bytesize=self.portpars[3],
         stopbits=self.portpars[4],timeout=self.timeout,writeTimeout=self.timeout
         )
      except:
         if self.logger:
            self.logger.error(FormatExceptionMessage("cSerial.open (%s), timeout:%s, id=%s:"%(
            self.portpars,self.timeout,self.serid),printmsg=True))
         self.close_port()

   def close_port(self):
      'Close the serial port or socket if open'
      if self.serial:
         if self.logger:
            self.logger.info('%s.close_port()',self.serid)
         if self.isusb:
            try:
               usb.util.dispose_resources(self.serial)
            except:
               pass
         else:
            self.serial.close()
         self.serial=None

   def inWaiting(self):
      '''
         Read all available data to buffer, return number of characters waiting
         in buffer
         self.zmqtype==0  normal socket port (does not use zmq)
         self.zmqtype==1  STReam socket (R/W, for receiving "normal" socket data)
         self.zmqtype==2  SUBsribe socket (read only)
         self.zmqtype==3  PUBlish socket (write only): return 0
      '''
      if self.zmqtype==3: return 0
      if not self.serial:
         self.open_port()
         if not self.serial: return self.bufflen
      da=''
      try:
         if self.isusb:
            dalst=[]
            try:
               psz=self.usbpacketsizedct[self.usbreadendpoint]
               while True:       # loop untime timeout
                  da=self.serial.read(self.usbreadendpoint,psz,1)    # 2 ms timeout(?)
                  # da is array with type code 'B' (unsigned char)
                  dalst.append(da.tostring())      # convert to string
            except usb.core.USBError as e:
               if e.errno<>110:     # not timeout
                  emsg="cSerial.inWaiting() in %s: USBErr:%s"%(self.serid,e)
                  print emsg
                  if self.logger: self.logger.error(emsg)
                  raise IOError(emsg)
            except:
               emsg=FormatExceptionMessage("cSerial.inWaiting() in %s"%(self.serid,),printmsg=True)
               if self.logger: self.logger.error(emsg)
               raise IOError(emsg)
            da=''.join(dalst)
         elif self.istelnet:
            try:
               if self.zmqtype==0:              # telnet
                  da=self.serial.recv(4096)     # 4096 should be large enough
               else:
                  try:
                     if self.zmqtype==1:        # get the frame and discard
                        da=self.serial.recv(zmq.NOBLOCK,0)
                     da=self.serial.recv(zmq.NOBLOCK,True)  # get the string copy of msg
                  except zmq.error.Again:
                     return self.bufflen
                  except:
                     emsg=FormatExceptionMessage("cSerial.inWaiting() in %s"%(self.serid,),printmsg=True)
                     if self.logger: self.logger.error(emsg)
                     raise IOError(emsg)
            except:  # check for timeout
               da=''
         else:
            n=self.serial.inWaiting()
            if n>0:
               da=self.serial.read(n)
         if da:
            self.buff.append(da)
            self.bufflen+=len(da)
      except:
         print FormatExceptionMessage("inWaiting:")
         self.close_port()
         raise    #2017-06-21
      return self.bufflen

   def sendBreak(self):
      '''
         added 2006-12-28 for LI-7500
      '''
      if self.istelnet or self.isusb: return      # no break
      if not self.serial: self.open_port()
      return self.serial.sendBreak()


   def read_port(self,line=False,timeout=-1):
      '''
         Read and return everything or a line (\n, or \r).
         If reading a line and no line is available, wait for it max timeout seconds
      '''
      retv=''
      tout=self.timeout if timeout<0 else timeout
      try:
         self.inWaiting()
         if not self.serial or (not line and self.bufflen==0): return retv
         if not line:
            retv=''.join(self.buff)
            self.bufflen,self.buff=0,[]
            return retv
         else:
            t=time.time()+tout
            while time.time()<t:
               retv=self._buffline(wholeline=True)
               if retv: break
               time.sleep(.02)
               self.inWaiting()      # read data to buffer
         if self.logger:
            self.logger.debug("cSerial.R,id=%s:'%s'"%(self.serid,
               retv.replace('\n','\\n').replace('\r','\\r').replace('\t','\\t')))
      except:
         pass
      return retv

   def _buffline(self,wholeline=False):
      '''
         Return first data line (ending with '\n', and/or '\n')
         in self.buff and remove it from self.buff

         If whole line not found, and wholeline==True, don't adjust the buffer
         and return ''
      '''
      if self.bufflen==0: return ''
      if len(self.buff)>1:    # more than one -> combine
         self.buff=[''.join(self.buff)]
      retv=''
      i=self.buff[0].find('\n')
      if i<0:
         i=self.buff[0].find('\r')
      if i>=0:
         retv=self.buff[0][:i+1]
         self.buff=[self.buff[0][i+1:]]
         self.bufflen=len(self.buff[0])
      if retv or wholeline: return retv
      self.bufflen=0
      retv=self.buff[0]
      self.buff=[]
      return retv

   def write_port(self,strn,response=True,waitForResponse=.5):
      '''
         Write str and wait and return response (if response==True) for
         waitForResponse seconds: write - wait - read everything and return it.
         
         NOTE: waits for single line, but returns everything incoming
         without timeout

         Before writing empty the serial port buffer and write it to log
         (if enabled) with level WARNING.
      '''
      retstr=''
      if not strn: return retstr
      if self.istelnet and self.zmqtype==2:      # no writing to SUB
         raise NotImplementedError("write_port, no writing to ZMQ SUBscribe socket in %s!",self.serid)
      try:
         if not self.serial: self.open_port()
         if not self.serial: return retstr
         ss=self.read_port()
         if self.logger and ss:
            self.logger.warning('cSerial.W,id=%s,buffer:"%s"'%(self.serid,
               ss.replace('\n','\\n').replace('\r','\\r')))

         if self.isusb:
            try:
               self.serial.write(self.portpars[-3],strn)
            except: # Exception as e:
               msg=FormatExceptionMessage('%s.cSerial.write_port(%s), USB error'%(
                  self.serid,strn.strip()))
               if self.logger: self.logger.error(msg)
               print ; print msg
               self.close_port()
               #print e.errno,e.strerror  # 16, "Resource busy"
               raise
         elif not self.istelnet:
            self.serial.write(strn)
         else:
            self.serial.send(strn)
         if response and self.serial and not self.zmqtype==3:
            time.sleep(waitForResponse)
            if self.inWaiting()>0: retstr=self.read_port()
      except:
         pass
      return retstr

def main():
   print "\ncSerial - reading test..."
   #portstr="localhost:9099,str"
   portstr="USB,0x1509,0x0a02,1,2,0x81,0x86"
   print "Port: %s"%(portstr,)
   
   import logging
   logger=logging.getLogger(__name__)
   logger.setLevel(logging.DEBUG)
   #logger.setLevel(logging.INFO)
   hndlr=logging.StreamHandler()
   fmat=logging.Formatter('%(asctime)s %(levelname)s %(message)s')
   hndlr.setFormatter(fmat)
   logger.addHandler(hndlr)
   logger.info("Logging start")
   
   s=cSerial(portstr,logger=logger)
   print "Port parameters:",s.portpars
   s.open_port()
   s.find_usb(current=True)
   da=''
   while True:
      try:
         n=s.inWaiting()
         if n>0:
            da=s.read_port(line=True,timeout=.1).strip()
            if da:
               print time.time(),da
         else:
            time.sleep(.01)
      except:
         s.close_port()
         break



if __name__=="__main__": main()
