# -*- coding: utf-8 -*-
"""
   Routines for accessing db tables cylinders and
   cylinder_mixing_ratios in db lasers
   

   Revision history:
      2015-05-12  start of devel
      2015-05-27..28  valid_start/end_utc removed from calibration_cylinders, 
         +install_date, +removal_date, +conc_eq
         cCalibrationCylinders update, +WSCYLTABLE
      2015-07-01  enhancements (new methods)
      2015-08-20  update cWSCylinders + bug fix
      2015-08-27..28  devel & bug fix - REWRITE: 3 classes -> one
      2015-09-21  'T','WS' -> 't','ws', +'mob'
      2015-09-23  +GetAllCylConcData()
      2015-09-24  + GetCylPortFromID()
      2015-10-05..06  fixes, removal of sqlite3, bug fixes
      2015-10-13  bug fix (if alias used)
      2015-10-15  bug fix in _GetCylcConcData
      2015-11-03  bug fix in GetCylIDsFromPort: did not check port
      2016-02-22  check for "dd" in conc_eq
      2016-05-30  +debugprint's
      2016-10-21  +foreign keys, CHECKs to table definition
      2016-10-21  some more error printouts
      2016-12-21  require station and instrid for role 't','ws'
      2017-02-02  +eq_ref_date
      2017-02-03  CYLCONCTABLE foreign key removed
      2017-02-06  +GetCylIDFromPort()
      2017-02-07  bug fix
      2017-02-08..10  +GetWSCylConc(), bug fixes (_readCylinders), other additions
      2017-02-20  replace pic.GetPicarrosAtStation() with instruments.GetInstrumentsAtStation()
      2017-02-21  +GetCylAge()
      2017-07-20  If conc_eq -> require "stdev" separated by comma in the same field
      2017-09-19  'pal' added (ICOS)
      2018-01-10  warning message if no cyl. data found
      2018-01-18  bug fix(?) in GetWSCylIDs()
      2018-01-23  return first picarro ID if dt not found in meta_gases_hmeans table
"""
from jahdblib import cDB
try:
   import instruments
except:
   import sys
   sys.path.append('/home/src/eric/epylib')
import instruments

CYLDBNAME='lasers'
CYLTABLE='cylinders'
CYLCONCTABLE='cylinder_mixing_ratios'
STATIONS=['mar','sam','sod','kum','mob','uto','pal']

## enable/dissable some printing for debugging purposes
DEBUG=0

TABLEDEFS={
   CYLCONCTABLE:
   """
      cylid                TEXT,
      component            TEXT,
      scale                TEXT,
      install_utc          TIMESTAMP WITHOUT TIME ZONE,
      version              INT2,
      conc_at_install      FLOAT4,
      dev                  FLOAT4,
      noaa_conc            FLOAT4,
      noaa_dev             FLOAT4,
      conc_eq              TEXT,
      eq_ref_date          TIMESTAMP WITHOUT TIME ZONE,
      comment              TEXT,
      CHECK (component IN ('co','co2','ch4','n2o','sf6')),
      CHECK ((component='co' AND scale IN ('X2004','X2014','X2014A')) OR
         (component='co2' AND scale IN ('X2007')) OR
         (component='ch4' AND scale IN ('X2004','X2004A')) OR
         (component='n2o' AND scale IN ('X2006A')) OR
         (component='sf6' AND scale IN ('X2006','X2014'))),
      PRIMARY KEY(cylid,component,scale,install_utc,version)
   """,
   CYLTABLE:
   """
      cylid                TEXT,
      install_utc          TIMESTAMP WITHOUT TIME ZONE,
      removal_utc          TIMESTAMP WITHOUT TIME ZONE,
      station              TEXT,
      instrid              TEXT,
      port                 INT2,
      alias                TEXT,
      role                 TEXT,
      comment              TEXT,
      CHECK (removal_utc IS NULL OR removal_utc>install_utc),
      CHECK (station IS NULL OR station IN (%s)),
      PRIMARY KEY(cylid,install_utc)
   """%(','.join(["'%s'"%x for x in STATIONS]),)
}

################################################################################      

      
class cCylinders:
   '''data access for table cylinders & cylinder_mixing_ratios'''
   def __init__(self,station,scales,calstdversions,instrid=''):
      '''
         scales is a dct {'compo':'scalename'}
         calstdversions is a dct {"compo":version}
      '''
      if DEBUG:
         print "cCylinders('%s',%s,%s,'%s')"%(station,scales,calstdversions,instrid)
      ## fetch all the station cylinders, and all the cylinders with no station
      assert station in STATIONS
      self.station=station
      self.instrid=instrid
      self.scales=scales
      self.calstdversions=calstdversions 
         
      ## self.cyls={'cylid':[[install_utc,removal_utc,'station','instrid',port,'alias','role','comment']]}
      ## self.role={'role':[[cylid,install_utc,removal_utc,'station','instrid',port,'role']]}
      ## self.alias={'alias':[[cylid,install_utc,removal_utc,'station','instrid',port,'role']]}
      ## self.ports={port:[cylid]}
      self.cyls={}
      self.role={}
      self.alias={}
      self.ports={}
      self._readCylinders()
      
      ## self.cylconcs={'cylid':{'compo':[installdt,conc_at,stdev,noaa_conc,dev,conc_eq,eq_ref_date]}}
      self.cylconcs=self._readCylinderConcs()
      ## [[iid,instaldt,remdt,instrtype,station,comment]]
      self.instratstation=instruments.GetInstrumentsAtStation(self.station,instrid=self.instrid)
      if instrid:
         iats=[]
         for i in self.instratstation:
            if i[0]==instrid: iats.append(i)
         self.instratstation=iats
      
   def __repr__(self):
      '''
      '''
      return "cCylinders(%s,%s,%s,%s)"%(
         self.station,self.scales,self.calstdversions,self.instrid)
         
   def _readCylinders(self):
      '''
      '''
      db=cDB(CYLDBNAME)
      cu=db.cursor()
      if self.instrid:
         iids="AND (instrid IS NULL OR instrid='%s')"%(self.instrid,)
      else:
         iids=""
      sq="SELECT DISTINCT * FROM \"%s\" WHERE (station='%s' OR station IS NULL) "\
         " %s ORDER BY install_utc;"%(
         CYLTABLE,self.station,iids)
      cu.execute(sq)
      if cu.rowcount==0: 
         raise ValueError("cCylinders(%s) - no data?:\n%s"%(self.station,sq))
      da=cu.fetchall()

      for d in da:
         cylid=d[0]
         if cylid not in self.cyls: self.cyls[cylid]=[]
         role=d[-2] if not d[-2] else d[-2].lower()
         alias=d[-3]
         port=d[-4]
         self.cyls[cylid].append([d[1],d[2],d[3],d[4],port,alias,role,d[-1]])
         if role:
            if role not in ('cal','ws','t','comp','limed','liref','lihigh','lilow'):
               raise ValueError("Table %s, cylinder %s, role '%s' NOT in accepted list - fix!"%(CYLTABLE,cylid,role))
            if role not in self.role: self.role[role]=[]
            if role in ('ws','t'):
               if self.station!=d[3] or d[4] is None:
                  raise ValueError("Table %s, cyl %s, role '%s': MUST HAVE both station and instrid defined!"%(
                     CYLTABLE,cylid,role))
            #                      cylid,idt,rdt,sta,instrid,port,alias,role
            self.role[role].append([d[0],d[1],d[2],d[3],d[4],port,alias,role])
            #if role!='cal': print "role[%s]:%s"%(role,self.role[role][-1]) #debug
         if alias:
            if alias not in self.alias: self.alias[alias]=[]
            self.alias[alias].append([d[0],d[1],d[2],d[3],d[4],port,alias,role])
         if port is not None and port>=0: # port
            if port not in self.ports: self.ports[port]=[]
            self.ports[port].append(d[0])
      #print self.role['ws']

   def _readCylinderConcs(self):
      '''
         {'cylid':{'compo':[installdt,conc_at,stdev,noaa_conc,dev,conc_eq,eq_ref_date]}}
      
         "stdev": in case of conc_eq, get if from there
      '''
      ret={}
      db=cDB(CYLDBNAME)
      cu=db.cursor()
      whlst=[]
      for compo in self.scales:
         whlst.append("(scale='%s' AND (component='%s' OR component='%s') AND version=%s)"%(
            self.scales[compo],compo.upper(),compo.lower(),self.calstdversions[compo]))
      wh='(%s)'%(' OR '.join(whlst),)
      sq="SELECT cylid,component,install_utc,conc_at_install,dev,noaa_conc,noaa_dev,conc_eq,eq_ref_date "\
         "FROM \"%s\" WHERE cylid IN (%s) AND %s ORDER BY cylid,install_utc;"%(
          CYLCONCTABLE,str(self.cyls.keys())[1:-1],wh)
      cu.execute(sq)
      if cu.rowcount==0:
         print "\nWARNING: cCylinders._readCylinderConcs(): NO cylinder concentration data found:\n%s\n"%(
            sq,)
      da=cu.fetchall()
      cu.close()
      db.close()
      if len(da)==0: return ret
      for d in da:
         ceq,stdev=d[7],d[4]
         if ceq:
            emsg="cylinders._readCylinderConcs: missing %%s (%s,%s,%s)"%(d[0],d[1],d[2])
            if 'dd' not in ceq:  raise ValueError(emsg%("dd in conc_eq",))
            if not d[8]: raise ValueError(emsg%("eq_ref_date",))
            if ceq.find(',')<4: raise ValueError(emsg%(",stdev in conc_eq",))
            ceq,stdev=ceq.split(',')
            stdev=float(stdev)
         if d[0] not in ret: ret[d[0]]={}
         if d[1] not in ret[d[0]]: ret[d[0]][d[1]]=[]
         ret[d[0]][d[1]].append([d[2],d[3],stdev,d[5],d[6],ceq,d[8]])
      return ret
      
   def GetCylID(self,cylid_alias_role,dtutc=None):
      '''
         If not found in either alias or role, 
         return as entered
      '''
      #print "GetCylID(%s,%s)"%(cylid_alias_role,dtutc) #debug
      car=cylid_alias_role
      if cylid_alias_role in self.cyls:   # proper cylid, return as is
         return cylid_alias_role
      if car in ('WS','T'):  car=car.lower()
      if car not in self.role and car not in self.alias:  # not found - return as is
         return cylid_alias_role
      
      dct=self.role if car in self.role else self.alias
      if not dtutc: 
         if car not in ('ws','t'):
            return dct[car][-1][0]
         ## otherwise must match station
         for x in dct[car]:
            if x[-5]==self.station: return x[0]
         return None
         
      for cylid,idt,rdt,sta,instrid,port,alias,role in dct[car]:
         #print cylid,idt,rdt,sta,instrid,port,alias,role     #debug
         if idt<=dtutc:
            if not rdt or dtutc<rdt: 
               if car not in ('ws','t'):
                  return cylid
               # must match station
               if sta==self.station: return cylid
      return  cylid_alias_role
      
   def DaysSinceInstall(self,cylid,dt):
      '''
      '''
      return (dt-self.GetCylInstallDt(cylid,dt)).days
      
   def GetCylIDsFromPort(self,port,dtstart,dtend=None):
      '''
         Return [[dtinstall,cylid]] for port at dtstart, or cylids between
         dtstart,dtend.  [] if not found
      '''
      #print "\nGetCylIDsFromPort(%s,%s,%s)"%(port,dtstart,dtend)
      ret=[]
      if port not in self.ports: return ret
      cylids=self.ports[port]    # list of cylids
      for cylid in cylids:
         for cylda in self.cyls[cylid]:
            ## [install_utc,removal_utc,'station','instrid',port,'alias','role','comment']
            if not dtend:
               if dtstart>=cylda[0] and (cylda[1] is None or startdt<=cylda[1]):
                  ret=[[cylda[0],cylid]]
                  break
            else:
               if cylda[1] and dtstart>cylda[1]: continue
               if dtend<cylda[0]: break
               if cylda[4]==port:
                  v=[cylda[0],cylid]
                  if v not in ret:
                     ret.append(v)
      return ret
      
   def GetCylIDFromPort(self,port,dt):
      '''
         Return cylid based on port,dt
      '''
      cylids=self.GetCylIDsFromPort(port,dt,dt)
      if len(cylids)!=1:
         print "\nGetCylIDFromPort(%s,%s): no/too many cylinders found:\n%s\n"%(port,dt,cylids)
         raise ValueError("Check dbtable lasers/cylinders")
      return cylids[0][1]
         
      
   def GetCylConcFromPort(self,port,compo,dt):
      '''
         Return [conc,stdev] based on port,compo,dt
      '''
      if DEBUG: print "GetCylConcFromPort(%s,%s,%s)"%(port,compo,dt)  #debug
      dt_id=self.GetCylIDsFromPort(port,dt,dt)
      if not dt_id:
         print "\ncCylinders.GetCylConcFromPort(%s,%s,%s) - NO DATA, check that port is defined in cylinders table"%(port,compo,dt)
         return None
      if len(dt_id)!=1:
         print "\ncCylinders.GetCylConcFromPort(%s,%s,%s) - too much data:\n%s"%(
            port,compo,dt,dt_id)
      idt,cylid=dt_id[0]
      return self.GetCylConc(cylid,compo,dt)
         
      
   def GetCylPortFromID(self,cylid,dt):
      '''
         Return port based on cylinder id, None if not found
      '''
      cyl=cylid
      #print "GetCylPortFromID(%s,%s)"%(cylid,dt)   #debug
      try:
         if cylid not in self.cyls: 
            cyl=self.GetCylID(cyl,dt)
            #print " -->",cyl     #debug
         for cylda in self.cyls[cyl]:
            if dt>=cylda[0] and (cylda[1] is None or dt<=cylda[1]):
               return cylda[-4]
      except:
         print "\ncCylinders.GetCylPortFromID(%s,%s) -> cyl:%s\n"%(cylid,dt,cyl)
         ks=self.cyls.keys()
         ks.sort()
         print "self.cyls.keys():\n",ks
         raise
      return None
      
   def IsWS(self,dt,cylid):
      '''
         return True if cylinder with cylid is a WS at datetime dt
      '''
      if cylid.lower()=='ws': return True
      try:
         for cyl in self.role['ws']:
            if cylid==cyl[0] and cyl[1]<=dt<=cyl[2]: return True
      except:
         return False
      return False
   
   def GetWSCylinders(self,instrid=None):
      '''
         return [[cylid,installdt,enddt,station,instrid,port,role]]
      '''
      if 'ws' not in self.role: return []
      if instrid is None: return self.role['ws']
      ret=[]
      for cylda in self.role['ws']:
         if cylda[4]!=instrid: continue
         ret.append(cylda)
      return ret
        
   def GetWSCylIDs(self,dtstart,dtend=None,instrid=None):
      '''
         return [[dtinstall,cylid]] for WS cylinder(s)
      '''
      if DEBUG: print "GetWSCylIDs(%s,%s,%s)"%(dtstart,dtend,instrid)
      ret=[]
      ## self.role={'role':[[cylid,install_utc,removal_utc,'station','instrid',port,'role']]}
      for cylda in self.role['ws']:
         if instrid and cylda[4]!=instrid: continue
         if not dtend:
            if dtstart>=cylda[1] and (cylda[2] is None or dtstart<cylda[2]):
               return [[cylda[1],cylda[0]]]
         else:
            if cylda[2] and dtstart>cylda[2]: continue
            if cylda[1]>=dtend: 
               if not ret: ret.append([cylda[1],cylda[0]])
               break
            ret.append([cylda[1],cylda[0]])
      if not ret:
         print
         for cylda in self.role['ws']: 
            print "%s: %16.16s - %16.16s %s"%(cylda[0],cylda[1],cylda[2],cylda[3:])
         print "cCylinders.GetWSCylIDs(%s,%s,%s) - no matching cylinder found - skipping !?"%(dtstart,dtend,instrid)
         return None
      return ret
      
   def GetWSCylPort(self,dt):
      '''
         Return port# of the WS cylinder in use at dt
      '''
      if DEBUG: print "GetWSCylPort(%16.16s)"%(dt,)
      for cylda in self.role['ws']:
         if cylda[1]<=dt and (not cylda[2] or dt<cylda[2]): return cylda[5]
      return None
      
      
   def GetWSCylConc(self,dt,compo):
      '''
         Return WS cyl. conc at datetime dt as a dict or a value if compo is not None
         
         2018-01-23: if earlier than 1st instrument, return 1st
      '''
      if DEBUG: print "GetWSCylConc(%s,%s)"%(dt,compo)
      iid=None
      for n,i in enumerate(self.instratstation):
         if n==0 and dt<i[1]:
            iid=i[0]
            break
         if i[1]<=dt<i[2]: 
            iid=i[0]
            break
      if iid is None: 
         print "\nGetWSCylConc(%s): no instrument found!"%(dt,)
         raise ValueError("FIX (add to) table lasers/meta_instruments)")
      wscyl=self.GetWSCylIDs(dt,dt,instrid=iid)
      if not wscyl or len(wscyl)>1:
         print "\nGetWSCylConc(%s): No/too many WS cylinders found, instr:%s\n%s"%(dt,iid,wscyl)
         raise ValueError("FIX dbtables")
      cylid=wscyl[0][1]
      return self.GetCylConc(cylid,compo,dt)[0] # return ONLY concentration
      
            
         
   def GetCylInstallDt(self,cylid,dt):
      '''
         Return install datetime for cylinder cylid at dt
      '''
      cyl=cylid
      if cyl not in self.cyls: cyl=self.GetCylID(cyl,dt)
      if cyl not in self.cyls:
         #for c in self.cyls: print c,self.cyls[c]
         raise ValueError('cCylinders.GetCylInstallDt(%s (->%s),%s) not found - check cylinders table!'%(cylid,cyl,dt))
      for d in self.cyls[cyl]:
#         print cyl,dt,d,dt>=d[0],not d[1]
         if dt>=d[0] and (not d[1] or dt<=d[1]): 
            return d[0]
      raise ValueError('cCylinders.GetCylInstallDt(%s,%s) install_utc ? (%s)'%(cylid,dt,self.cyls[cyl]))
      
   def _GetCylConcData(self,cylid,compo,dt=None):
      '''
         If dt==None, return latest
         If no data for the cylinder, return None
      '''
      ret=[]
      cyl=cylid
      if cyl not in self.cyls: cyl=self.GetCylID(cyl,dt)
      if cyl not in self.cylconcs: return ret
      try:
         da=self.cylconcs[cyl][compo]
      except KeyError:
         return None
      if not dt or len(da)==1: 
         return da[-1]   # return latest/only
      idx=0
      while idx<len(da)-1:
         if da[idx+1][0]>=dt: 
            return da[idx]
         idx+=1
      if da[-1][0]<=dt: return da[-1]
      raise ValueError("cCylinders._GetCylConcData(%s (-> %s),%s,%s)\n  NO DATA FOUND, check install/removal times!"%(
         cylid,cyl,compo,dt))
      
   def GetAllCylConcData(self,cylids):
      '''
         cylids: [[installdt,cylid]]
         Return [[cylid,compo,scale,installdt,version,concat,stdev,noaac,stdev,conceq,eq_ref_date]]
      '''
      ret=[]
      for idt,cid in cylids:
         try:
            cyldata=False
               #{'cylid':{'compo':[installdt,conc_at,stdev,noaa_conc,dev,conc_eq,eq_ref_date]}}            
            cylc=self.cylconcs[cid]
            for compo in cylc:
               for i,d in enumerate(cylc[compo]):
                  if d[0]<=idt:
                     try:
                        if cylc[compo][i+1][0]<=idt: continue
                     except:
                        pass
                     ret.append([cid,compo,self.scales[compo],idt,self.calstdversions[compo],d[1],d[2],d[3],d[4],d[5],d[6]])
                     cyldata=True
            if not cyldata: raise
         except:
            raise ValueError('cCylinders.GetAllCylConcData: no conc. data for cyl %s install at %s'%(cid,idt))
      return ret
   
      
   def GetCylConc(self,cylid,compo,dt):
      '''
         Return [conc,stdev]
      '''
      if DEBUG: print "GetCylConc(%s,%s,%s)"%(cylid,compo,dt)
      cyldata=self._GetCylConcData(cylid,compo,dt)
      if not cyldata: 
         print "cCylinders.GetCylcConc(%s,%s,%s) - NO DATA - fix cylinder_mixing_ratios table"%(cylid,compo,dt)
         return None
      if DEBUG:
         print "CylcConcData:"
         for cylda in cyldata: print cylda
      if not cyldata[-2]: return [cyldata[1],cyldata[2]]
      dd=dt.absdays-cyldata[-1].absdays
      return [eval(cyldata[-2]),cyldata[2]]
      
   def GetRole(self,cylid,dts):
      '''
      '''
      if cylid not in self.cyls: return ''
      dt=dts
      if type(dt)==type(''): dt=PARSEDT(dt)
      for cyl in self.cyls[cylid]:
         if dt>=cyl[0] and (cyl[1] is None or dt<=cyl[1]):
            if cyl[5]: return cyl[5]
            return ''
      return ''
      
   def GetCylAge(self,cylid,compo,dt):
      '''
         Return number of days (eq_ref_date - dt).days
      '''
      cyldata=self._GetCylConcData(cylid,compo,dt)
      if not cyldata: 
         print "\ncCylinders.GetCylcAge('%s','%s',%s)\nNO DATA from cCylinders._GetCylConcData():\nFix cylinder_mixing_ratios table(?)\n"%(cylid,compo,dt)
         return 0
         #raise ValueError("Add cylinder!")
      return (dt-cyldata[-1]).days

################################################################################      

def main():
   #return
   tbl=CYLCONCTABLE
   #tbl=CYLTABLE
   print "\nCalled directly only creates table %s, without dropping old one(s)"%(tbl,)
   print "\nEnable by editing main()"
   
   db=cDB(CYLDBNAME)
   if True:
      return
   if False:
      print "Copying data..."
      cu=db.cursor()
      cu.execute("SELECT * FROM %s_ ORDER BY cylid,component,install_utc;"%(tbl,))
      da=cu.fetchall()
      for d in da:
         sq="INSERT INTO %s VALUES('%s','%s','%s','%s',%s,%s,%s,%s,%s,'%s','%s','%s');"%(
            tbl,d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7],d[8],d[9],d[10],d[11])
         sq=sq.replace('None','NULL')
         sq=sq.replace("'NULL'","NULL")
         cu.execute(sq)
      db.commit()
      return
   if False:
      db.CreateTable(tbl,TABLEDEFS[tbl],drop_existing=True)
      cu=db.cursor()
      if False:
         cu.execute("SELECT * from cylinders_0;")
         da=cu.fetchall()
         sqo="INSERT INTO cylinders VALUES('%s','%s','%s','%s','%s',%s,'%s','%s','%s');"
         for d in da:
            sq=sqo%tuple(d)
            sq=sq.replace("'None'","NULL").replace("None","NULL")
            sq=sq.replace("n't","n''t")
            try:
#               print sq
               cu.execute(sq)
            except:
               print sq
               raise
            db.commit()
            
      if False:
         cu.execute("SELECT * from cylinder_mixing_ratios_0 ORDER BY cylid,install_utc,component;")
         da=cu.fetchall()
         sqo="INSERT INTO cylinder_mixing_ratios VALUES('%s','%s','%s','%s',%s,%s,%s,%s,%s,'%s','%s');"
         for d in da:
            sq=sqo%tuple(d)
            sq=sq.replace("'None'","NULL").replace("None","NULL")
            try:
#               print sq
               cu.execute(sq)
            except:
               print sq
               raise
            db.commit()
   else:
      print "Creating table %s ..."%(tbl,)
      try:
         db.CreateTable(tbl,TABLEDEFS[tbl])
      except Exception as e:
         db.rollback()
         print "Exception - table already exists...(?):"
         print e
            


if __name__ == "__main__":
   main()
