# -*- coding: latin-1 -*-
"""
c name:      inverse
c version:   200208.19
c author:    stephen j. frakes
c purpose:   to compute a geodetic inverse
c            and then display output information
c
c input parameters:
c -----------------
c
c output parameters:
c ------------------
c
c local variables and constants:
c ------------------------------
c answer           user prompt response
c b                semiminor axis polar (in meters)
c baz              azimuth back (in radians)
c buff             input char buffer array
c dd,dm,ds         temporary values for degrees, minutes, seconds
c dlon             temporary value for difference in longitude (radians)
c dmt,d_mt         char constants for meter units
c edist            ellipsoid distance (in meters)
c elips            ellipsoid choice
c esq              eccentricity squared for reference ellipsoid
c faz              azimuth forward (in radians)
c filout           output file name
c finv             reciprocal flattening
c hem              hemisphere flag for lat & lon entry
c ierror           error condition flag with d,m,s conversion
c lgh              length of buff() array
c option           user prompt response
c r1,r2            temporary variables
c ss               temporary variable
c tol              tolerance for conversion of seconds
c
c name1            name of station one
c ld1,lm1,sl1      latitude  sta one - degrees,minutes,seconds
c ald1,alm1,sl1    latitude  sta one - degrees,minutes,seconds
c lat1sn           latitude  sta one - sign (+/- 1)
c d_ns1            latitude  sta one - char ('N','S')
c lod1,lom1,slo1   longitude sta one - degrees,minutes,seconds
c alod1,alom1,slo1 longitude sta one - degrees,minutes,seconds
c lon1sn           longitude sta one - sign (+/- 1)
c d_ew1            longitude sta one - char ('E','W')
c iaz1,maz1,saz1   forward azimuth   - degrees,minutes,seconds
c isign1           forward azimuth   - flag  (+/- 1)
c glat1,glon1      station one       - (lat & lon in radians )
c p1,e1            standpoint one    - (lat & lon in radians )
c
c name2            name of station two
c ld2,lm2,sl2      latitude  sta two - degrees,minutes,seconds
c ald2,alm2,sl2    latitude  sta two - degrees,minutes,seconds
c lat2sn           latitude  sta two - sign (+/- 1)
c d_ns2            latitude  sta two - char ('N','S')
c lod2,lom2,slo2   longitude sta two - degrees,minutes,seconds
c alod2,alom2,slo2 longitude sta one - degrees,minutes,seconds
c lon2sn           longitude sta two - sign (+/- 1)
c d_ew2            longitude sta two - char ('E','W')
c iaz2,maz2,saz2   back azimuth      - degrees,minutes,seconds
c isign2           back azimuth      - flag  (+/- 1)
c glat2,glon2      station two       - (lat & lon in radians )
c p2,e2            forepoint two     - (lat & lon in radians )
c
c global variables and constants:
c -------------------------------
c a                semimajor axis equatorial (in meters)
c f                flattening
c pi               constant 3.14159....
c rad              constant 180.0/pi
c
c    this module called by:  n/a
c
c    this module calls:      elipss, getrad, inver1, todmsp
c    gethem, trim,   bufdms, gvalr8, gvali4, fixdms, gpnhri
c    datan,  write,  read,   dabs,   open,   stop
c
c    include files used:     n/a
c
c    common blocks used:     const, elipsoid
c
c    references:             see comments within subroutines
c
c    comments:
c
c********1*********2*********3*********4*********5*********6*********7**
c::modification history
c::1990mm.dd, sjf, creation of program
c::199412.15, bmt, creation of program on viper
c::200203.08, crs, modified by c.schwarz to correct spelling of Clarke
c::200207.15, rws, modified i/o & standardized program documentation
c::                added subs trim, bufdms, gethem, gvali4, gvalr8
c::200207.23, rws, replaced sub inver1 with gpnarc, gpnloa, gpnhri
c::200208.15, rws, fixed an error in bufdms
c::              - renamed ellips to elipss "common error" with dirct2
c::              - added FAZ & BAZ to printed output
c::200208.19, rws, added more error flags for web interface code
c::              - added logical nowebb
c::200208.xx, sjf, program version number 2.0
c********1*********2*********3*********4*********5*********6*********7**
"""
'''
   Revision history:
      2011-01-14  BoundingCoordinates()
'''
import mx.DateTime as DateTime
import math

rad=180./math.pi


def DecDeg2DegMinSec(decdec):
   """
   Decimal degree to degminsec.x (0640524.2)
   """
   deg=int(decdec)
   z=(decdec-deg)*3600.
   min=int(z/60)
   z=(z-min*60)
   return "%03d%02d%02d.%05d"%(deg,min,int(z),100000*(z-int(z)))

def DecMinSec2DecDeg(dms):
   """
   degminsec.x to decimal degree
   """
   z=float(dms)
   deg=int(z/10000)
   z-=10000*deg
   min=int(z/100)
   z-=100*min
   sec=int(z/100)
   z-=100*sec
#   print deg,min,sec,z
   return deg+min/60.+(sec+z)/3600.



a=6378137.
f=1./298.25722210088
esq=f*(2.-f)
elips="GRS80 / WGS84 (NAD83)"
def GeodeticInverse(lalo1,lalo2):
   """
   Calculate geodetic inversion.
   Return (fwe azimuth, backward azimuth, ellipsoidal distance) as
          (decimal degree, decimal degree, km)
   """
   faz,baz,edist=_gpnhri(lalo1,lalo2)
#   z=_gpnhri(lalo1,lalo2)
#   print z,type(z)
   if 1000.*edist<.00005:
      faz=0. ; baz=0.
#   print faz,baz,edist
#   print faz*rad,baz*rad
   return (faz*rad,baz*rad,edist)

def _gpnhri(lalo1,lalo2):
   """
   Return (az1, az2, geod_dist)

   c  name:        gpnhri
   c version:     200208.09
   c written by:  robert (sid) safford
   c purpose:     subroutine to compute helmert rainsford inverse problem
   c
   c     solution of the geodetic inverse problem after t. vincenty
   c     modified rainsford's method with helmert's elliptical terms
   c     effective in any azimuth and at any distance short of antipocal
   c     from/to stations must not be the geographic pole.
   c     parameter a is the semi-major axis of the reference ellipsoid
   c     finv=1/f is the inverse flattening of the reference ellipsoid
   c     latitudes and longitudes in radians positive north and west
   c     forward and back azimuths returned in radians clockwise from south
   c     geodesic distance s returned in units of semi-major axis a
   c     programmed for ibm 360-195   09/23/75
   c
   c     note - note - note -
   c     1. do not use for meridional arcs and be careful on the equator.
   c     2. azimuths are from north(+) clockwise and
   c     3. longitudes are positive east(+)
   c
   c input parameters:
   c -----------------
   c a            semi-major axis of reference ellipsoid      meters
   c f            flattening (0.0033528...)
   c esq          eccentricity squared
   c p1           lat station 1                               radians
   c e1           lon station 1                               radians
   c p2           lat station 2                               radians
   c e2           lon station 2                               radians
   c
   c output parameters:
   c ------------------
   c az1          azi at sta 1 -> sta 2                       radians
   c az2          azi at sta 2 -> sta 1                       radians
   c s            geodetic dist between sta(s) 1 & 2          meters
   """
   tol0=5.e-15 ; tol1=5.e-14 ; tol2=7.e-3
   twopi=math.pi*2

   p1=lalo1[0]/rad ; e1=lalo1[1]/rad
   p2=lalo2[0]/rad ; e2=lalo2[1]/rad
   if e1<0: e1=e1+2.*math.pi
   if e2<0: e2=e2+2.*math.pi

#     test the longitude difference with tol1
#     tol1 is approximately 0.000000001 arc seconds
   ss = e2-e1
   if abs(ss)<tol1:
      e2=e2+tol1
#      print ' longitudal difference is near zero '
      r2 = p2
      r1 = p1
      arc=_gpnarc(r1,r2)
      s=abs(arc)
      if p2>p1:
         az1 = 0.
         az2 = math.pi
      else:
         az1 = math.pi
         az2 = 0.
      return (az1,az2,arc/1000)

#     test for longitude over 180 degrees
   dlon=e2-e1
   if dlon>=0.:
      if math.pi<=dlon and dlon<twopi:
         dlon = dlon-twopi
   else:
      ss = abs(dlon)
      if math.pi<=ss and ss<twopi:
         dlon = dlon+twopi

   ss = abs(dlon)
   if ss>math.pi:
#c::     write(*,*) '  '
#c::     write(*,*) ' Longitude difference over 180 degrees  '
#c::     write(*,*) ' Turn it around '
      ss = twopi-ss

#     compute the limit in longitude (alimit), it is equal
#     to twice the distance from the equator to the pole,
#     as measured along the equator (east/ewst)
   alimit = math.pi*(1.-f)
#     test for anti-nodal difference
   if ss>=alimit:
      r1 = abs(p1) ; r2 = abs(p2)
#           latitudes r1 & r2 are not near the equator
      if not(r1>tol2 and r2>tol2): #goto60
#       longitude difference is greater than lift-off point
#       now check to see if  "both"  r1 & r2 are on equator

         if not (r1<tol1 and r2>tol2): #goto60
            if not(r2<tol1 and r1>tol2):  #goto60
   #       check for either r1 or r2 just off the equator but < tol2
               if r1>tol1 or r2>tol1:
                  return (0.,0.,0.)

   #       compute the azimuth to anti-nodal point
   #c::     write(*,*) '  '
   #c::     write(*,*) ' Longitude difference beyond lift-off point '
   #c::     write(*,*) '  '
               az2,aa,bb,sms=_gpnloa(dlon,az1)
   #       compute the equatorial distance & geodetic
               equ=a*abs(dlon)
               s=equ-sms
               return(az1,az2,s)

   f0   = (1.-f)
   b    = a*f0
   epsq = esq/(1.-esq)
   f2   = f*f
   f3   = f*f2
   f4   = f*f3
#     the longitude difference
   dlon  = e2-e1
   ab    = dlon
   kount = 0
#     the reduced latitudes
   u1    = f0*math.sin(p1)/math.cos(p1)
   u2    = f0*math.sin(p2)/math.cos(p2)
   u1    = math.atan(u1)
   u2    = math.atan(u2)
   su1   = math.sin(u1)
   cu1   = math.cos(u1)
   su2   = math.sin(u2)
   cu2   = math.cos(u2)

#c     counter for the iteration operation
   while 1:
      kount = kount+1
      clon=math.cos(ab)
      slon=math.sin(ab)
      csig=su1*su2+cu1*cu2*clon
      ssig=math.sqrt((slon*cu2)**2+(su2*cu1-su1*cu2*clon)**2)
      sig=math.atan2(ssig,csig)
      sinalf=cu1*cu2*slon/ssig
      w=(1.-sinalf*sinalf)
      t4=w*w
      t6=w*t4
#     the coefficients of type a
      ao=f-f2*(1.+f+f2)*w/4.+3*f3*(1.+
         9.*f/4.)*t4/16.-25.*f4*t6/128.
      a2=f2*(1.+f+f2)*w/4.-f3*(1.+9.*f/4.)*t4/4.+ \
         75.*f4*t6/256.
      a4=f3*(1.+9.*f/4.)*t4/32.-15.*f4*t6/256.
      a6=5.*f4*t6/768.
#     the multiple angle functions
      qo=0.
      if w>tol0: qo=-2.*su1*su2/w
      q2=csig+qo
      q4=2.*q2*q2-1.
      q6=q2*(4.*q2*q2-3.)
      r2=2.*ssig*csig
      r3=ssig*(3.-4.*ssig*ssig)
#     the longitude difference
      s=sinalf*(ao*sig+a2*ssig*q2+a4*r2*q4+a6*r3*q6)
      xz=dlon+s
      xy=abs(xz-ab)
      ab=dlon+s
      if xy<0.5e-13: break
      if not (kount<7): break

#     the coefficients of type b
   z=epsq*w
   bo=1.+z*(1./4.+z*(-3./64.+z*(5./256.-
      z*175./16384.)))
   b2=z*(-1./4.+z*(1./16.+z*(-15./512.+
      z*35./2048.)))
   b4=z*z*(-1./128.+z*(3./512.-z*35./8192.))
   b6=z*z*z*(-1./1536.+z*5./6144.)
#     the distance in meters
   s=b*(bo*sig+b2*ssig*q2+b4*r2*q4+b6*r3*q6)
#c     first compute the az1 & az2 for along the equator
   if dlon>math.pi: dlon = (dlon-2.*math.pi)
   if abs(dlon)>math.pi: dlon = (dlon+2.*math.pi)
   az1=math.pi/2.
   if dlon<0.: az1 = 3.*az1
   az2=az1+math.pi
   if az2>2.*math.pi: az2 = az2-2.*math.pi
#     now compute the az1 & az2 for latitudes not on the equator
   if not(abs(su1)<tol0 and abs(su2)<tol0):
      tana1 =  slon*cu2/(su2*cu1-clon*su1*cu2)
      tana2 =  slon*cu1/(su1*cu2-clon*su2*cu1)
      sina1 =  sinalf/cu1
      sina2 = -sinalf/cu2
#       azimuths from north,longitudes positive east
      az1= math.atan2(sina1,sina1/tana1)
      az2= math.pi-math.atan2(sina2,sina2/tana2)

   if az1<0.: az1 = az1+2.*math.pi
   if az2<0.: az2 = az2+2.*math.pi
   return (az1,az2,s/1000.)

def _gpnarc(p1,p2):
   """
   Return geodetic distance in km

   C NAME:        GPNARC
   C VERSION:     200005.26
   C WRITTEN BY:  ROBERT (Sid) SAFF
   C PURPOSE:     SUBROUTINE TO COM
   C              BETWEEN TWO LATIT
   C
   C INPUT PARAMETERS:
   C -----------------
   C AMAX         SEMI-MAJOR AXIS O
   C FLAT         FLATTENING (0.003
   C ESQ          ECCENTRICITY SQUA
   C PI           3.14159...
   C P1           LAT STATION 1
   C P2           LAT STATION 2
   C
   C OUTPUT PARAMETERS:
   C ------------------
   C ARC          GEODETIC DISTANCE
   """
   tt=5.e-15
#     CHECK FOR A 90 DEGREE LOOKUP
   FLAG=0
   S1=abs(p1) ; S2=abs(p2)
   if (math.pi/2.-tt)<S2 and S2<(math.pi/2.+tt): FLAG=1
   if S1>tt: FLAG=0
   DA = (p2-p1) ; S1 = 0. ; S2 = 0.

#     COMPUTE THE LENGTH OF A MERIDIONAL ARC BETWEEN TWO LATITUDES
   E2 = esq
   E4 = E2*E2
   E6 = E4*E2
   E8 = E6*E2
   EX = E8*E2

   T1 = E2*(3./4.)
   T2 = E4*(15./64.)
   T3 = E6*(35./512.)
   T4 = E8*(315./16384.)
   T5 = EX*(693./131072.)

   A=1.+T1+3.*T2+10.*T3+35.*T4+126.*T5

   if not FLAG:
      B = T1+4.*T2+15.*T3+56.*T4+210.*T5
      C = T2+06.*T3+28.*T4+120.*T5
      D = T3+08.*T4+045.*T5
      E = T4+010.*T5
      F = T5
      DB = math.sin(p2*2.)-math.sin(p1*2.)
      DC = math.sin(p2*4.)-math.sin(p1*4.)
      DD = math.sin(p2*6.)-math.sin(p1*6.)
      DE = math.sin(p2*8.)-math.sin(p1*8.)
      DF = math.sin(p2*10.)-math.sin(p1*10.)
#     COMPUTE THE S2 PART OF THE SERIES EXPANSION
      S2 = -DB*B/2.+DC*C/4.-DD*D/6.+DE*E/8.-DF*F/10.

#     COMPUTE THE S1 PART OF THE SERIES EXPANSION
   S1=DA*A
#     COMPUTE THE ARC LENGTH
   return a*(1.-esq)*(S1+S2)


def _gpnloa(dl,az1):
   """
   Return (az2,a0,b0,sms)

   C NAME:        GPNLOA
   C VERSION:     200005.26
   C WRITTEN BY:  ROBERT (Sid) SAFFORD
   C PURPOSE:     SUBROUTINE TO COMPUTE THE LIFF-OFF-AZIMUTH CONSTANTS
   C
   C INPUT PARAMETERS:
   C -----------------
   C AMAX         SEMI-MAJOR AXIS OF REFERENCE ELLIPSOID
   C FLAT         FLATTENING (0.0033528 ... )
   C ESQ          ECCENTRICITY SQUARED FOR REFERENCE ELLIPSOID
   C PI           3.14159...
   C DL           LON DIFFERENCE
   C AZ1          AZI AT STA 1 -> STA 2
   C
   C OUTPUT PARAMETERS:
   C ------------------
   C AZ2          AZ2 AT STA 2 -> STA 1
   C AO           CONST
   C BO           CONST
   C SMS          DISTANCE ... EQUATORIAL - GEODESIC  (S - s)   "SMS"
   """
   TT=5.e-13
   DLON=abs(dl)
   CONS=(math.pi-DLON)/(math.pi*f)
   F=f
#     COMPUTE AN APPROXIMATE AZ
   AZ=math.asin(CONS)
   T1=1.
   T2=(-1./4.)*F*(1.+F+F*F)
   T4=3./16.*F*F*(1.+(9./4.)*F)
   T6=(-25./128.)*F*F*F

   for ITER in range(1,7):
      S=math(AZ)
      C2= S*S
#     COMPUTE NEW AO
      AO=T1+ T2*C2 + T4*C2*C2 + T6*C2*C2*C2
      CS=CONS/AO
      S=math.asin(CS)
      if abs(S-AZ)<TT: break
      AZ=S

   AZ1= S
   if dl<0.: AZ1 = 2.*math.pi-AZ1
   AZ2=2.*math.pi-AZ1
#     EQUATORIAL - GEODESIC  (S - s)   "SMS"
   ESQP = esq/(1.-esq)
   S= math.cos(AZ1)

   U2   = ESQP*S*S
   U4   = U2*U2
   U6   = U4*U2
   U8   = U6*U2

   T1   =     1.
   T2   =    (1./4.)*U2
   T4   =   (-3./64.)*U4
   T6   =    (5./256.)*U6
   T8   = (-175./16384.)*U8

   BO   = T1 + T2 + T4 + T6 + T8
   S    = math.sin(AZ1)
   SMS  = a*math.pi*(1. - f*abs(S)*AO - BO*(1.-f))
   return (AZ2,A0,B0,SMS)

#################################################################

radius=6371.01 # earth radio in km
MIN_LAT,MAX_LAT=-90/rad,90/rad
MIN_LON,MAX_LON=-180/rad,180/rad
def BoundingCoordinates(lalo,distance):
   '''
      Compute bounding coordinates of bounding box with distance km
      radius, assuming earth to be spherical. lalo=[latitude,longitude]

      Return  ((minlat,maxlat),)(minlon,maxlon))

      lat,lon in decimal degrees!

      Adapted from:
         http://janmatuschek.de/LatitudeLongitudeBoundingCoordinates
   '''
   radDist=distance/radius

   ll=[lalo[0]/rad,lalo[1]/rad]
#   if ll[1]<0: ll[1]+=2*math.pi    # not?!
   minLat=ll[0]-radDist
   maxLat=ll[0]+radDist

   if minLat>MIN_LAT and maxLat<MAX_LAT:
      deltaLon=math.asin(math.sin(radDist)/math.cos(ll[0]))
      minLon=ll[1]-deltaLon
      if minLon<MIN_LON: minLon+=2*math.pi
      maxLon=ll[1]+deltaLon
      if maxLon>MAX_LON: maxLon-=2*math.pi
   else: # a pole is within the distance
      minLat=max(minLat,MIN_LAT)
      maxLat=min(maxLat,MAX_LAT)
      minLon=MIN_LON
      maxLon=MAX_LON
   return ((minLat*rad,maxLat*rad),(minLon*rad,maxLon*rad))



def DayLength(dt,coord):
   """
   Return length of day in hours

   coord=(latitude (N), longitude (E))

   Doesn't seem to give very accurate results!

   """

   solardecl=.409*math.sin((2*math.pi/365.)*dt.day_of_year-1.405)
   print "DL..." ; print solardecl
   sunset_hour_angle=math.acos(-math.tan(coord[0]/rad
      )*math.tan(solardecl))
   print sunset_hour_angle
   dl=24*(sunset_hour_angle/math.pi)
   print dl
   return dl




#################################################################

if __name__ == "__main__":
   c1=(68.0,24.24)   #Matorova
   c2=(67.987222,24.243333)   #Kenttrova
   c3=(68.063333,24.033889)   #Laukukero
   c4=(67.973333,24.116111)   #Sammaltunturi
   c5=(67.973528,24.115111)   #Sammaltunturi old
   c6=(67.92,23.92)           # Srkijrvi
   oul=(66.32,29.4)           # Oulanka
   x1=(c1[0],-c1[1])
   x3=(c3[0],-c3[1])
   z3=(-c3[0],-c3[1])
   print "Matorova-Kenttrova:",GeodeticInverse(c1,c2)
   print "Matorova-Sammal:",GeodeticInverse(c1,c4)
   print "Sammal-Laukukero:",GeodeticInverse(c4,c3)
   print "Matorova-Srkijrvi:",GeodeticInverse(c1,c6)
   print "Sammal-Srkijrvi:",GeodeticInverse(c4,c6)
   print "Matorova-Oulanka:",GeodeticInverse(c1,oul)
   print "Matorova-Laukukero:",GeodeticInverse(c1,c3)
   print "Kenttrova-Laukukero:",GeodeticInverse(c2,c3)
   GeodeticInverse(c1,c3)
   GeodeticInverse(x1,x3)
   GeodeticInverse(c3,z3)
   z=DecDeg2DegMinSec(c3[0])
   zz=DecMinSec2DecDeg(z)
   print c3[0],z,zz

   blder=(40.2,-50)
   dt=DateTime.DateTime(2002,4,1)
   dl=DayLength(dt,c4)



