"""
   Common/general file handling routines.
   
   Revision history
      2015-08-24  
      2016-02-19  bug fix in ReadFileContent when reading zip file
      2016-05-18  bug fix: import also bz2
"""
import os,glob,zipfile,gzip,bz2,time
import mx.DateTime as DateTime
from epylib import epylib_Error

DTDT=DateTime.DateTime
DTISO=DateTime.ISO.ParseDateTime

def ReadIniFile(inifile,printmsg=True,calledby=None):
   """
      Read inifile and return its content as dictionary
      of dictionaries: dict[section][option] (==value).

      It appears as if ALL ConfigParsers makes .lower() to
      option names, and (Safe)ConfigParses does wildcard(?) expansion
      --> RawConfigParser() used...
      
      if printmsg is True, print "Reading ini file", and if calledby
      is not empty, prints also that information.
      
      2014-11-28: replace \n -> ' ' in bodies (continuation lines)
   """
   if printmsg: 
      print "Reading ini file %s"%inifile,
      if calledby:
         print "(%s)"%calledby,
   from ConfigParser import RawConfigParser

   # subclass RawConfigParser so that options can be made CASE SENSITIVE!
   class RCParser(RawConfigParser):
      def __init__(self,defaults=None):
         RawConfigParser.__init__(self,defaults)
      def optionxform(self,optionstr):
         return optionstr

   parser=RCParser()
   parser.read(inifile)
   dct={}
   for sect in parser.sections():
      dct[sect]={}
      for opt in parser.options(sect):
         dct[sect][opt]=parser.get(sect,opt)
         dct[sect][opt]=dct[sect][opt].replace('\n',' ')
   if printmsg:
      if not dct: print " - no luck"
      else: print " - OK"
   return dct


def CreateDir(mask,dt=None,printout=False):
   '''
      Create directory dt.strftime(mask), if it doesn't
      exist

      if dt==None, use current date

      Created 2006-11-17 byjah
   '''
   d=dt
   if not d: d=DateTime.now()
   odir=d.strftime(mask)
   if not os.path.exists(odir):
      os.makedirs(odir)
      if printout:
         print "  directory '%s' created"%odir


def FilenameListByDt(fnpat,startdt,enddt=None):
   '''
      Return list of filenames in time order with file
      names containing yyyymm or yyyy-mm between startdt
      and enddt (months inclusive).

      fnpat may be string, or list of strings (patterns)

      If enddt=='' (or None), use current month as the end time.

      Revision history:
         2006-07-25 %1 == year in one number (with current century/decade)
         2007-02-06 max 500 loops
         2007-09-17 wildcard support added
         2007-12-20 counter to 500 fixed
         2007-12-21 option %j (day of year) support added
   '''
   if not enddt: enddt=DateTime.now()
   fls=[]
   dts=DateTime.DateTime(startdt.year,startdt.month,startdt.day)
   edt=DateTime.DateTime(enddt.year,enddt.month,enddt.day)

   if type(fnpat)==type([]):
      fnpatterns=fnpat[:]
   else:
      fnpatterns=[fnpat]

   for fp in fnpatterns:
      n=0
      dt=dts
      if fp.find('%d')>=0 or fp.find('%j')>=0:
         step=1
      elif fp.find('%m')>=0:
         dt=DTDT(dt.year,dt.month,1)
         step=2
      else:
         step=3
         dt=DTDT(dt.year,1,1)

      fps=fp
      while dt<=edt and len(fls)<5000:
         if fp.find('%1'): fps=fp.replace('%1','%1d'%(dt.year%10,))
         fl=dt.strftime(fps)
         if step==1:
            dt+=1
         elif step==2:  # next month
            dt+=dt.days_in_month
         else:          # only year
            dt=DTDT(dt.year+1,1,1)
         if os.path.exists(fl):
            fls.append(fl)
         else:
            f=glob.glob(fl)
            if len(f)>0:
               fls.extend(f)
   return fls


def ReadRESEDAData(ifile,yearcol=2,startdt=None,enddt=None,replaceNANs=True):
   '''
      Read and return data (NOT headers) from RESEDA file as
      [[dt_AS_ISO_STRING,datacol1,datacol2,...]]

      Revision history:
         2006-07-25  two digit year handling, replace single ' # ' with ''
         2006-07-31  6 digit time handling (hhmmss)
   '''
   data=[]
   lines=open(ifile,'r').readlines()
   started=False
   for ln in lines:
      ln=ln.strip()
      if not ln: continue
      if started:
         if ln[0]=="-" and ln.find("/")>1:
            started=False
            continue
         s=[x.strip() for x in ln.split()]
         if len(s)<yearcol+4:     # not enough lines
            print "Too short a line:'%s' - skipped"%ln
            continue
         yr=int(s[yearcol])
         if yr<50:
            yr+=2000
         elif yr<100:
            yr+=1900
         lentcol=len(s[yearcol+3])
         tcol=int(s[yearcol+3])
         if lentcol>4:
            secs=tcol%100
            mins=(tcol%10000)/100
            hours=tcol/10000
         else:
            secs=0
            mins=tcol%100
            hours=tcol/100
         dt=DTDT(yr, int(s[yearcol+1]),int(s[yearcol+2]),hours,mins,secs)
         if startdt and dt<startdt: continue
         if enddt and dt>enddt: break
         da=["%19.19s"%dt]
         if replaceNANs:
            for d in s[yearcol+4:]:
               if d=='#': d=''
               da.append(d)
         else:
            da+=s[yearcol+4:]
         data.append(da)
      else:
         if ln[:2]=="--": started=True
   return data


def ReadCSVData(ifile,startdt=None,enddt=None,nCols=None,dtcol=0,asStr=True,
   inorder=None,no_duplicate_dtcols=False):
   '''
      Read and return data from .csv file (not headers) as
      [[dt_AS_ISO_STRING,datacol1,datacol2,...]] if asStr,
      otherwise returned as [dt,float(dcol),...].
      inorder is a list of columns to return in order of the list.

      datetime in ISO format

      if nCols ([n,n2]), data lines with differing amount of columns
      are discarded

      Revision history:
         2006-09-22  If first (non blank) character is ", discard line
         2006-11-22  par dtcol added
         2006-12-04  par asStr added
         2007-09-19  par inorder added
         2007-10-03  s2.append(None) in case of exception
         2007-12-20  .zip file support added (in case archive contains only one file)
         2010-02-23  no_duplicate_dtcols added
         2011-02-09  no_duplicate_dtcols handling enhanced (checks ALL the previous dts)
         2012-02-01  .gz file handling
   '''
   data=[]
   datadts=[]
   if ifile[-4:].lower()=='.zip':
      zip=zipfile.ZipFile(ifile,'r')
      lns=zip.read(zip.namelist()[0])
      zip.close()
      lines=lns.split('\n')
   elif ifile[-3:].lower()=='.gz':
      zip=gzip.GzipFile(ifile,'r')
      lns=zip.read()
      zip.close()
      lines=lns.split('\n')
   else:
      lines=open(ifile,'r').readlines()

   try:
      fline=lines[0].strip()
   except:
      return data
   if fline[0]=='-' and fline[-1]=='/':
      print "  %s in RESEDA(?) - skipped"%ifile
      return data
   for ln in lines:
      ln=ln.strip()
      if not ln or ln[0]=='"' or not ln[0].isdigit(): continue
      s=[x.strip() for x in ln.split(',')]
      if nCols and len(s) not in nCols: continue
      if inorder and type(inorder) in (type([]),type(())):
         s2=[]
         for ino in inorder:
            try:
               s2.append(s[ino])
            except:
               s2.append(None)
         s=s2
      if no_duplicate_dtcols and data:
         if s[dtcol] in datadts:
            print "ReadCSVData: duplicate datetime '%s' - later one discarded"%s[dtcol]
            continue
      if no_duplicate_dtcols:
         datadts.append(s[dtcol])
      if not startdt and not enddt and asStr:
         data.append(s)
      else:
         try:
            dt=DTISO(s[dtcol])
         except:
            print "epylib/ReadCSVData: no good ISO-datetime: %s"%s[dtcol]
            continue
            raise epylib_Error("DTISO(%s)"%s[dtcol])
         if startdt or enddt:
            if startdt and dt<startdt: continue
            if enddt and dt>enddt: break
         if asStr:
            data.append(s)
         else:
            da=[]
            lens=len(s)
            for i in xrange(lens):
               if i==dtcol:
                  da.append(dt)
               else:
                  try:
                     da.append(float(s[i]))
                  except:
                     da.append(None)
            data.append(da)
   return data


def ReadRadaktFile(ifile,startdt=None,enddt=None,nCols=None):
   '''
      Read "old style", i.e. date/time in format yymmdd hhmmss,
      space (or any whitespace) separated straight ASCII file

      Return [[dt,datacols]]
   '''
   data=[]
   lines=open(ifile,'r').readlines()
   for ln in lines:
      ln=ln.strip()
      if not ln or not ln[0].isdigit(): continue
      s=[x.strip() for x in ln.split()]
      if nCols and len(s) not in nCols: continue
      dai=int(s[0]) ; tii=int(s[1])
      yr=dai/10000
      yr+=1900
      if yr<1950: yr+=100
      dt=DateTime.DateTime(yr,(dai%10000)/100,dai%100,tii/10000,(tii%10000)/100,tii%100)
      if startdt and dt<startdt: continue
      if enddt and dt>enddt: break
      data.append(["%19.19s"%dt]+s[2:])
   return data


def ReadZipFile(ifile,flname=None):
   '''
      Read content of a (zip) file returning its content.
      If flname is given, reads content of that file in the archive,
      otherwise the archive must contain only one file.

      Raise exception if no data file found
   '''
   ifl=ifile
   if not os.path.exists(ifl):
      ifile2=ifile+'.zip'
      if not os.path.exists(ifile2):
         raise epylib_Error('ReadZipFile: file "%s" not found'%ifile)
      ifl=ifile2
   if ifl[-4:]!=".zip":
      return open(ifl,'r').read()
      
   import zipfile
   zipf=zipfile.ZipFile(ifl,'r')
   names=zipf.namelist()
   if len(names)>1:
      if flname==None:
         raise epylib_Error('ReadZipFile: more than one file in .zip, and not flname given')
      else:
         if flname not in names: # remove dirs
            for f in names:
               if f.basename==flname:
                  ifl=f
                  break
         else:
            ifl=flname
   elif len(names)==0:
      raise epylib_Error('ReadZipFile: no files in archive "%s"?'%ifl)
   else:
      ifl=names[0]
   return zipf.read(ifl)
   
   
def ReadFileContent(ifile,tolines=False):
   '''
      Return content of the file ifile. If the file is compressed
      (extension .zip,.gz,.bz2), decompresses it.
   '''
   fext=os.path.splitext(ifile)[-1]
   if fext=='.zip':
      con=ReadZipFile(ifile)
   elif fext=='.gz':
      con=gzip.open(ifile,'r').read()
   elif fext=='.bz2':
      con=bz2.BZ2File(ifile,'r').read()
   else:
      con=open(ifile,'r').read()
   if not tolines: return con
   return con.splitlines()


def FileAgeInHours(files):
   '''
      Return age of file(s) since their creation in hours as a list
      None for file not found
   '''
   ret=[]
   t=time.time()
   if not files:
      return ret
   if type(files)!=type([]):
      fls=[files]
   else:
      fls=files[:]
   for fl in fls:
      try:
         secs=t-os.path.getctime(fl)
         ret.append(secs/3600.)
      except:
         ret.append(None)
   return ret



####################################################################
####################################################################

def main():
   '''
      Tests
   '''
   zf="/home/data/public/pallas/krova/PyBARGrad/data/2006/KenGrad_2006-01-01.grad"
   lns=ReadZipFile(zf)
   print "len(data):",len(lns)


if __name__ == "__main__":
   main()


