# -*- coding: utf-8 -*-
"""
   Miscellaneous python helper routines.

   Revision history:
      2007-11-27  +CalcHTRH()
      2010-10-06  +CalcH(), vappres-> VaporPressure(T)
      2017-05-04  +ppmv2Dewpoint(), +TDp2RH()
      2017-07-27  fix to WindDirMean sqrt(neg)
"""
import math

def WindDirMean(dirlst,speedlst=None):
   """
   Calculate mean for wind direction based on wind direction and speed
   (i.e. wind vector). In case speedar is not given, uses wind speed 1.

   Input: wind speed and wind direction lists
   Output: (winddir,stdev)

   Revision history
      2002-10-01  None handling added
      2003-11-17  BUG fix: did not handle wind dir 0 at all!!!!
                  Wind speed limit from 0.4 to 0.2
      2006-09-19  speed,dir->dir,speed, speedar=None, return also
                  stdev based on method by Skibin (1984)
                  "A Simple Method for Determining the Stdev of Wind Dir"
      2010-01-20  angle int(a) -> int(a + 0.5)
      2017-07-27  sd=math.sqrt(): negative? -> abs()
   """
   xi=0.; yi=0.
   i=0 ; n=0
   speedar=speedlst
   if speedar==None: speedar=[1.]*len(dirlst)
   ar1=[] ; s1=0 ; ar2=[] ; s2=0
   for i,spd in enumerate(speedar):
      if spd>0.2 and dirlst[i]!=None:
         ang=dirlst[i]
         if ang>180 or ang==0:
            if ang==0: ang=360
            ar2.append(ang)
            s2+=ang*ang
         else:
            ar1.append(ang)
            s1+=ang*ang
         angle=ang/180.*math.pi
         xi+=spd*math.sin(angle)
         yi+=spd*math.cos(angle)
         n+=1
   if n==0: return None,None
   # note: atan2(y,x)
   angle=math.atan2(xi/n,yi/n)/math.pi*180.
   if angle<0: angle+=360
   n1=len(ar1) ; n2=len(ar2)
   if n2==0:
      d=sum(ar1)/n1
      try:
         sd=math.sqrt((s1-sum(ar1)*sum(ar1)/n1)/(n1-1))
      except:
         sd=0.
   elif n1==0:
      d=sum(ar2)/n2
      try:
         sd=math.sqrt((s2-sum(ar2)*sum(ar2)/n2)/(n2-1))
      except:
         sd=0.
   elif max(ar2)-min(ar1)<180:
      d=(sum(ar1)+sum(ar2))/(n1+n2)
      sd=math.sqrt((s1+s2-d*d*(n1+n2))/(n1+n2-1))
   else:
      try:
         d=(sum(ar1)+sum(ar2)-n2*360)/(n1+n2)
         v=(s1+s2+n2*360*(360-2*sum(ar2)/n2)-d*d*(n1+n2))/(n1+n2-1)
         sd=math.sqrt(abs(v))
      except:
         print "\nERROR in WindDirMean():"
         print "dir_in (len: %s):%s"%(len(dirlst),dirlst)
         print "spd_in (len:%s):%s"%(len(speedlst),speedlst)
         print "n1,n2:",n1,n2,", s1,s2:",s1,s2,",ar1,ar2:",ar1,ar2
         raise
      #print "n1,n2:",n1,n2,", s1,s2:",s1,s2,",ar1,ar2:",ar1,ar2
   #print angle,d,sd
   return int(angle+.5),sd


def DistanceBetweenTwoSites(site1coord,site2coord):
   """
   Return (distance,direction) as (km,deg) between to sites with
   coordinates given in decimal decgrees as (latitude,longitude).
   The direction is FROM site1 TO site2!
   d=R arccos[...
   Cos d = SinØ1.SinØ2+CosØ1.CosØ2.Cos (L1-L2)
   """
#   print site1coord,site2coord
   la0,lo0=site1coord
   la1,lo1=site2coord
   la0=la0/180.*math.pi ; lo0=lo0/180.*math.pi
   la1=la1/180.*math.pi ; lo1=lo1/180.*math.pi
   d=6371.*math.acos(math.sin(la0)*math.sin(la1)+
      math.cos(la0)*math.cos(la1)*math.cos(lo0-lo1))
   if d<0: d=-d      # ? is this possible?
#   print "(%f,%f),(%f,%f):"%(la0,lo0,la1,lo1),
   lax=la1-la0
   lox=lo1-lo0
   ang=math.atan2(lax,-lox)
#   print lax,lox,ang,ang/math.pi*180
   ang=ang/math.pi*180-90
   if ang<0:ang+=360

   return (d,ang)


def CalcHFromTRH(T,RH):
   '''
      Return water content as {'mass':g/kg, 'mol':mmol/mol} based
      on temperature (in deg C) and relative humidity (RH)
   '''
   es=611.2*math.exp(17.67*T/(T+273.15-29.65))
   H=RH/100.*es/100./(28.97/18.02)
   H2=RH/100.*es/100
   return {'mass':H,'mol':H2}

def CalcH(t,RH,water=True,iterateDp=True):
   '''
      Return water content as a dictionary based on
      temperature t (deg C) and relative humidity RH (%)
      (vapor pressure calculated over water)

      {'mass','mol','td'}
   '''
   es=VaporPressure(t,water)
   e=RH/100.*es
   H=e/100./(28.97/18.02)
   H2=RH/100.*es/100.
   DP=CalcDpFromP(e,water,iterateDp)
   return {'mass':H,'mol':H2,'dp':DP}

def CalcDpFromP(e,water=True,iterate=True):
   '''
      Return dewpoint/frostpoint temperature from water vapor pressure
      (calc. iteratively with more exact equations if iterate==True)

      e in Pascals, return dp in C
   '''
   try:
      if water:
         dp=243.12/(17.62/(math.log(e)-math.log(611.2))-1)
      else:
         dp=272.62/(22.46/(math.log(e)-math.log(611.2))-1)
      if iterate:
         # try to find a closer value...
         vpi=VaporPressure(dp,water)
         step=.01
         if vpi>e: step=-step
         diff=None
         for i in range(100):    # max adj. 100*step (1C)
            ddp=dp+step*i
            vp=VaporPressure(ddp,water)
            d=abs(vp-e)
            if i==0:
               diff=d
            else:
               if d>diff:
                  return dp+step*(i-1)
         # no proper results, return the original dp
   except:
      return -99.
   return dp


def VaporPressure(dp,water=True):
   '''
      Given dewpoint/frostpoint temperatures in C, by default over water,
      returns water vapor pressure in Pascals

      http://www.npl.co.uk/reference/faqs/
         how-do-i-convert-between-units-of-dew-point-and-relative-humidity-%28faq-thermal%29

      Equations from Sonntag 1990
   '''
   t=float(dp)+273.15
   if water:
      ewt=-6096.9385/t+21.2409642-2.711193e-2*t+1.673952e-5*t*t+2.433502*math.log(t)
   else:
      ewt=-6024.5282/t+29.32707+1.0613868e-2*t-1.3198825e-5*t*t-0.49382577*math.log(t)
   return math.exp(ewt)

def Dewpoint2VolumeMixingRatio(dp,press=1013.25,water=True):
   '''
      Return h2o volume mixing ratio in permill over water by default,
      of over ice if water==False
   '''
   e=VaporPressure(dp,water)
   return e/(press*100)*1e3

def ppmv2Dewpoint(ppmv,press=1013.25,water=True):
   '''
      Return dewpoint (or frostpoint if water==False) based
      on volume mixing ratio (pw/(ptot-pw)*1e6).
      Pressure in hPa
   '''
   pw=press*100/(1e6/ppmv+1)
   return CalcDpFromP(pw,water,iterate=True)
   
def TDp2RH(t,dp):
   '''
      Return RH based on temperature (in C) and dewpoint 
   '''
   return 100*VaporPressure(dp)/VaporPressure(t)
   