# -*- coding: utf-8 -*-
'''
jahplotlib
=============

   :Release: |__version__|

   A plotting library using `matplotlib <http://matplotlib.sourceforge.net>`_.
   Requires external libary mx.DateTime found in
   `eGenix mx Base <http://www.egenix.com/products/python/mxBase/>`_.

   Usage is based on instantiating  suitable class with figure size (in pixels)
   and dpi, and calling its plotting method(s) with a dictionary containing data
   and other required/optional information. See documentation of individual
   classes and methods.

   Revision history:
      * Oct 2009 first (0.1) version partially based on jahplotlib.py
      * 2009-12-08 (0.2) __del__
      * 2010-06-28 (0.3) bug fixes, par. tosamefig
      * 2010-09-17 (0.4) bug fix in 'xgrid,','ygrid'
      * 2010-09-20 (0.5) "textrel" + fixes
      * 2010-10-19 (0.6) "yfilter"
      * 2011-03-25 (1.0) Major update
         "y2" 2nd ("right") y-axis plots, remove "tosamefig"
      * 2011-03-29 (1.1) bug fixes + some changes
      * 2011-03-30 (1.2) xlabel change + bug fixes
      * 2011-04-04..06 (1.3) bug fix in x(auto)scaling
      * 2011-07-04 (1.4) fig size: also wxh (width x height) useable
      * 2011-07-13 (1.5) 'x2', 'x','x2' also [[[]]], by default x-ticks ONLY to last axis
      * 2011-07-14 (1.6) bug fix in 'x2','y2' handling, unicode_TYPE to _setlabel
      * 2011-07-15 (1.7) some additional data checks
      * 2011-07-18 (1.8) yminmax axno handling
      * 2011-09-13 (1.9) daylocator addition/adjustments, some error checkings added
      * 2011-09-15 (1.10) 'kwargs2' for 2nd y-axis, other kwargs changes
      * 2011-09-28 (1.11) non-date x-axis ticks + other updates
      * 2011-10-05 (1.12) minor date tick enhancements
      * 2011-12-15 (1.13) bug fixes ('ymaxnbins')
      * 2012-02-02 (1.14) options 'line','xmaxnbins','xformatter','xlog' added
      * 2012-02-06 (1.15) bug fix to prev
      * 2012-09-14 (1.16) minor fix to "yfilter"
      * 2012-12-07 (1.17) (partial) handling of arrays containing no data
      * 2013-01-11 (1.18) minor fixes
      * 2013-01-14 (1.19) +"ylabeloptions","xlabeloptions","yticks"
      * 2013-05-10 (1.20) +"titleoptions" (documentation)
      * 2013-06-06 (1.21) minor fix in _getTxtProp
      * 2013-06-07 (1.22) empty subplot plotting
      * 2013-12-17 (1.23) bug fix in _getTxtProp (in case of ('txt',))
      * 2016-07-09 (1.24) yfilter: "is_relative" added
      * 2016-09-06 (1.25) get rid of epylib.FormatExceptionmessage dependency, +"xdtformatter"
      * 2017-07-24 (1.26) label: check for ', replace with \\'
      * 2017-09-21 (1.27) minor text changes
      
.. |__version__| replace:: v 1.27 Sep 2017 byjah
'''
__version__="v 1.27 Sep 2017 byjah"
import warnings,traceback,sys
warnings.filterwarnings("ignore")
import matplotlib
try:
   matplotlib.use('Agg')
except:
   pass
import mx.DateTime as DateTime
import datetime
import matplotlib.pylab as plt
from mpl_toolkits.axes_grid import AxesGrid



def FormatExceptionMessage(strng='',shortform=True):
   """
      Format the last occurred exception as a one-line message
      with optional prefix (strng). Return the formatted string.
   """
   try:
      s=strng+''.join(traceback.format_exception(
         *sys.exc_info())[-2:]).strip().replace('\n',': ')      
   except:
      s="ERROR in FormatExceptionMessage!!!!"
   rf=0
   if shortform:
      rf=s.rfind('Exception:')
      if rf>0:
         while rf>=0 and s[rf]!=' ': rf-=1
         rf+=1
      else:
         rf=0
   return s[rf:]



class cPlotError(Exception):
   '''
   This exception is raised when some checked-by-the-lib error
   occurs. "Normal" (not caught) exceptions are passed upwards.
   '''
   def __init__(self,value,addexc=False):
      if addexc:
         self.value=FormatExceptionMessage(value)
      else:
         self.value=value
   def __str__(self):
      return repr(self.value)



class _cPlotBase:
   '''
   Common base class for plotting classes.
   '''
   unicode_TYPE=type(u's')
   str_TYPE=type('s')
   list_TYPE=type([])
   tuple_TYPE=type(())
   dict_TYPE=type({})
   datetime_TYPE=type(datetime.datetime.now())
   DateTime_TYPE=type(DateTime.now())


   def __init__(self,size="800,600",dpi=100):
      '''
         size      fig size in pixels, "width,height" OR widthxheight
         dpi       dots-per-inch
      '''
      if size.find(',')<0:
         self.size=size.replace('x',',')
      else:
         self.size=size
      self.dpi=dpi
      self.figure=None
      self.axes=None
      self.nfigs=0

   def __del__(self):
      '''
         Call to self.close()
      '''
      self.close()

   def Fig(self,nPlots=1,useAxesGrid=True,axesPad=.1):
      '''
         Create a figure with nplots subplots with ncols columns
         (number of rows automatically calculated). If figure
         allready exists, return next subfig.

         ===========   =======================================================
         nplots        total number of plots
         useAxesGrid   use AxesGrid instead of "normal" figure, default: True
         axesPad       padding (used with useAxesGrid)
         ===========   =======================================================

         Return list of all subplot axes (self.axes) if useAxesGrid==False,
         otherwise return ???
      '''
      self.close()      # close possibly existing figure
      w,h=[float(x) for x in self.size.split(',')]
      self.figure=plt.figure(num=1,figsize=(w/self.dpi,h/self.dpi),dpi=self.dpi)
      if useAxesGrid:
         self.axes=AxesGrid(self.figure,111,nrows_ncols=(nPlots,1),axes_pad=axesPad)
      else:
         self.axes=[plt.subplot(nPlots,1,i) for i in range(1,nPlots+1)]
      self.nfigs=nPlots
      return self.axes

   def close(self):
      '''
         Close current figure
      '''
      if self.figure:
         plt.close(self.figure)
         self.figure=None



   def GetListDim(self,lst):
      '''
         Return 0 if not a list, 1 ([]) if one-dim, 2 if two ([[]]),
         and 3 if three ([[[]]])
      '''
      n=0
      try:
         if type(lst)!=self.list_TYPE: return n
         n+=1
         if type(lst[0])!=self.list_TYPE: return n
         n+=1
         if type(lst[0][0])!=self.list_TYPE: return n
         n+=1
         if type(lst[0][0][0])!=self.list_type: return n
      except:
         pass
      return n

   def dtLocators(self,ddays,dtformatter=None):
      '''
         Return majorLocator, minorLocator, dateFormatter
         based on number of days (ddays)
      '''
      if not ddays or ddays<=1:
         if ddays and ddays<.1:
            majLocator=plt.MinuteLocator(interval=30)
            minLocator=None
         elif ddays and ddays<.5:
            majLocator=plt.HourLocator(interval=1)
            minLocator=None
         else:
            majLocator=plt.HourLocator(interval=4)
            minLocator=plt.HourLocator(interval=1)
         dateFormatter=plt.DateFormatter("%H:%M")
      elif ddays<=2:
         majLocator=plt.HourLocator(interval=8)
         minLocator=plt.HourLocator(interval=2)
         dateFormatter=plt.DateFormatter("%H:%M")
      elif ddays<8:
         majLocator=plt.DayLocator()
         minLocator=plt.HourLocator(12)
         dateFormatter=plt.DateFormatter('%m-%d')
      elif ddays<16:
         majLocator=plt.DayLocator(interval=3)
         minLocator=plt.DayLocator()
         dateFormatter=plt.DateFormatter('%m-%d')
      elif ddays<32:
         majLocator=plt.DayLocator(interval=7)
         minLocator=plt.DayLocator()
         dateFormatter=plt.DateFormatter('%m-%d')
      elif ddays<64:
         majLocator=plt.DayLocator(interval=14)
         minLocator=plt.DayLocator(interval=7)
         dateFormatter=plt.DateFormatter('%m-%d')
      elif ddays<182:
         majLocator=plt.MonthLocator()
         minLocator=plt.DayLocator(interval=7)
         dateFormatter=plt.DateFormatter('%m-%d')
      elif ddays<366:
         majLocator=plt.MonthLocator(interval=2)
         minLocator=plt.MonthLocator()
         dateFormatter=plt.DateFormatter("%Y-%m-%d")
      elif ddays<3*366:
         majLocator=plt.MonthLocator(interval=3)
         minLocator=plt.MonthLocator()
         dateFormatter=plt.DateFormatter("%Y-%m-%d")
      else:
         majLocator=plt.MonthLocator(interval=12)
         minLocator=plt.MonthLocator(interval=3)
         dateFormatter=plt.DateFormatter("%Y")
      if dtformatter:
         dateFormatter=plt.DateFormatter(dtformatter)
      return majLocator,minLocator,dateFormatter

   def _getTxtProp(self,labeltuple):
      '''
         Return "txt","properties_string" or None,None if
         not found/error
      '''
      try:
         if type(labeltuple)==self.tuple_TYPE:
            txt=labeltuple[0]
            if len(labeltuple)>1 and labeltuple[1]:
               optstr=','+','.join(labeltuple[1:])
            else:
               optstr=''
         else:
            txt=labeltuple
            optstr=''
         txt=self._checkformathtext(txt)
         return txt,optstr
      except:
         return None,None


   def _checkformathtext(self,txt,fonttoroman=True,space='\\hspace{0.3}'):
      '''
         Check whether txt is to be formatted as mathtext
         (start/end with $), and change the font to Roman.
         Also replace double spaces with space entered as option.
      '''
      if type(txt) not in (self.unicode_TYPE,self.str_TYPE): return txt
      if type(txt)!=self.unicode_TYPE:
         txt=unicode(txt,'utf-8')
      if not txt or len(txt)<3 or txt[0]!='$' or txt[-1]!='$':
         return txt
      txt=txt.replace('  ',space)
      if fonttoroman:
         txt=ur"$\mathrm{%s}$"%(txt[1:-1],)
      else:
         txt=ur"$%s$"%(txt[1:-1],)
      return txt

   def _setlabel(self,ax,dd,labname,nsubfig=None,axno=None):
      '''
         Set title, x/y -axis label

         =======  ====================================
         ax       axis instance
         dd       data dictionary
         labname  label name (xlabel,ylabel)
         nsubfig  subfig# (not applicable with xlabel)
         axno     None,0,1
         =======  ====================================

      '''
      #print "_setlabel:\nlabname:%s, nsubfig:%s, axno:%s, dd[labname]:%s (%s)"%(labname,nsubfig,axno,dd[labname],type(dd[labname][0]))
      try:
         lbl=dd[labname]
         if type(lbl) not in (self.tuple_TYPE,self.str_TYPE,self.unicode_TYPE):
            lbl=dd[labname][nsubfig]
            if type(lbl) not in (self.tuple_TYPE,self.str_TYPE,self.unicode_TYPE):
               lbl=dd[labname][nsubfig][axno]
      except (KeyError,IndexError):
         return
      except:
         raise

      txt,optstr=self._getTxtProp(lbl)   
      
      if False: #debug
         los="%soptions"%labname
         print los, los in dd, "optstr:'%s'"%optstr
         if los in dd: print "  ",dd[los]
      if not optstr and "%soptions"%labname in dd:
         optstr=",%s"%dd["%soptions"%labname]
      
#      print '  lbl "%s" -> txt "%s", optstr "%s"'%(lbl,txt,optstr) ; print #debug#

      if not txt: return
      if txt.find("'")>=0:
         txt=txt.replace("'","\\'")
      evalstr="ax.set_%s(u'%s'%s)"%(labname,txt,optstr)
      try:
         eval(evalstr)
      except:
         print "evalstr:'%s'"%evalstr
         raise


   def _setticklabelsize(self,ax,dd,labname,nsubfig=None,axno=None):
      '''
         Set x/y tick label size for the axis ax

         =======  ====================================
         ax       axis instance
         dd       data dictionary
         labname  label name (x/yticklabelsize)
         nsubfig  subfig# (not applicable with x)
         =======  ====================================

      '''
      try:
         lbl=dd[labname][nsubfig][axno]
      except:
         try:
            lbl=dd[labname][nsubfig]
         except:
            try:
               lbl=dd[labname]
            except:
               return
      txt,optstr=self._getTxtProp(lbl)
      if not txt: return
      if labname[0]=='y':
         for lbl in ax.get_yticklabels():
            lbl.set_size(float(txt))
      else:
         for lbl in ax.get_xticklabels():
            lbl.set_size(float(txt))

   def _setticklabelcolor(self,ax,dd,labname,nsubfig=None,axno=None):
      '''
         Set x/y tick label color for the axis ax

         =======  ====================================
         ax       axis instance
         dd       data dictionary
         labname  label name (x/yticklabelsize)
         nsubfig  subfig# (not applicable with x)
         =======  ====================================
      '''
      try:
         lbl=dd[labname][nsubfig][axno]
      except:
         try:
            lbl=dd[labname][nsubfig]
         except:
            try:
               lbl=dd[labname]
            except:
               return
      txt,optstr=self._getTxtProp(lbl)
#      print nsubfig,axno,labname,lbl,txt,'"%s"'%optstr #debug#
      if not txt: return
      if labname[0]=='y':
         for lbl in ax.get_yticklabels():
            lbl.set_color(txt)
      else:
         for lbl in ax.get_xticklabels():
            lbl.set_color(txt)


   def _setgrid(self,ax,dd,nsubfig=None):
      '''
         Set x/y grids (if so specified)

         2010-09-17  fixed (add ax.get_x/yaxis())
      '''
      for g in ('xgrid','ygrid'):
         if g not in dd: continue
         axname="ax.get_%saxis()"%g[0]
         v=dd[g]
         if v==True:
            optstr="color='gray'"
         elif v==False:
            optstr=False
         else:
            if nsubfig!=None and type(v)==self.list_TYPE:
               optstr=v[nsubfig]
            else:
               optstr=v
         axis=eval(axname)
         evalstr='axis.grid(%s)'%optstr
#         print "evalstr,axname,axis:",evalstr,axname,axis #debug#
         eval(evalstr)

   def _setyminmax(self,ax,dd,nsubfig=None,axno=None):
      '''
         Set yminmax if 'yminmax' and/or 'yminminmaxmax' keys
         found in data dictionary (dd)
      '''
#      print "_setyminmax: nsubfig:%s,axno:%s"%(nsubfig,axno) #debug
      if 'yminmax' in dd:
         ymm=dd['yminmax']
         if nsubfig!=None and type(ymm)==self.list_TYPE:
            ymm=ymm[nsubfig]
            if axno!=None and type(ymm)==self.list_TYPE:
               ymm=ymm[axno]
         if ymm:
            ax.set_ylim(ymm[0],ymm[1])
      else:
         if not 'yminminmaxmax' in dd: return
         ylims=list(ax.get_ylim())
         ymm=dd['yminminmaxmax']
         if nsubfig!=None and type(ymm)==self.list_TYPE:
            ymm=ymm[nsubfig]
            if axno!=None and type(ymm)==self.list_TYPE:
               ymm=ymm[axno]
         if not ymm: return
         if ylims[0]<ymm[0]: ylims[0]=ymm[0]
         if ylims[1]>ymm[1]: ylims[1]=ymm[1]
         ax.set_ylim(ylims[0],ylims[1])

   def __setformatter(self,ax,dd,pars=('yformatter','ymaxnbins'),nsubfig=None,axno=None):
      '''
         Set axis scale formatter, log, nbins
      '''
      logset=False
      if pars[0][0]=='y':
         yax=True
         aa=ax.yaxis
         ll='ylog'
         yy='ymaxnbins'
      else:
         yax=False
         aa=ax.xaxis
         ll='xlog'
         yy='xmaxnbins'
         
      if pars[0] in dd:
         yf=dd[pars[0]]
         if nsubfig!=None and type(yf)==self.list_TYPE:
            yf=yf[nsubfig]
         if yf:
          aa.set_major_formatter(plt.FormatStrFormatter(yf))
      else:
         aa.set_major_formatter(plt.ScalarFormatter(useOffset=False))
         try:
            if dd[ll]==True or dd[ll][nsubfig]==True:
               if yax:
                  ax.set_yscale('log')
               else:
                  ax.set_xscale('log')
               logset=True
         except:
            pass

      if not logset:
         nb=5
         if yy in dd:
            nb=dd[yy]
            if nsubfig!=None and type(nb)==self.list_TYPE:
               nb=dd[yy][nsubfig]
            try:
               if type(nb)!=type(1) or nb<2 or nb>20: nb=5
            except:
               nb=5
         if not yax:
            try:
               if self.xtypes[nsubfig]<2: return
            except:
               pass
            try:
               if self.xtypes[0]>0: return
            except IndexError:
               return
         aa.set_major_locator(plt.MaxNLocator(nbins=nb))
         
   def _setyformatter(self,ax,dd,nsubfig=None,axno=None):
      '''
      '''
      self.__setformatter(ax,dd,nsubfig=nsubfig,axno=axno)
      
   def _setxformatter(self,ax,dd,nsubfig=None,axno=None):
      '''
      '''
      self.__setformatter(ax,dd,nsubfig=nsubfig,axno=axno,pars=('xformatter','xmaxnbins'))

   def _setlineat(self,ax,dd,nsubfig=None):
      '''
         Draw a line at y==dd['lineat'][nsubfig]
      '''
      if not 'lineat' in dd: return
      lineat=dd['lineat']
      if nsubfig!=None and type(lineat)==self.list_TYPE:
         lineat=lineat[nsubfig]
      if not lineat: return
      if type(lineat)==self.tuple_TYPE:
         yval=[lineat[0],lineat[0]]
         try:
            optstr=lineat[1]
         except:
            optstr="'ls':'-','color':'darkgray','lw':1"
      else:
         yval=[lineat,lineat]
         optstr="'ls':'-','color':'darkgray','lw':1"
      ymm=ax.get_ylim()
      xmm=ax.get_xlim()
#      xv=[xmm[0],xmm[1]]
      if not (ymm[0] < yval[0] < ymm[1]): return
      evalstr="ax.plot(xmm,yval,**{%s})"%(optstr,)
      try:
         eval(evalstr)
      except:
         print "evalstr:'%s'"%evalstr
         raise
      ## must re-set limits
      ax.set_xlim(xmm[0],xmm[1])
      ax.set_ylim(ymm[0],ymm[1])
      

   def _setline(self,ax,dd,nsubfig=None):
      '''
         Draw a line at y==dd['line'][nsubfig]
      '''
      if not 'line' in dd: return
      line=dd['line']
      if nsubfig!=None and type(line)==self.list_TYPE:
         line=line[nsubfig]
      if not line: return
      
      xval=line[0]
      yval=line[1]
      if len(line)>2:
         optstr=line[2]
      else:
         optstr="'ls':'-','color':'darkgray','lw':1"
      evalstr="ax.plot(xval,yval,**{%s})"%(optstr,)
      try:
         eval(evalstr)
      except:
         print "evalstr:'%s'"%evalstr
         raise

   def _settext(self,ax,dd,nsubfig=None):
      '''
         Write out dd['text'] (if any)
      '''
      if 'text' not in dd and 'textrel' not in dd: return
      rel=False
      try:
         txt=dd['text']
      except:
         txt=dd['textrel']
         rel=True
      if nsubfig!=None and type(txt)==self.list_TYPE:
         txt=txt[nsubfig]
      if not txt or (type(txt)!=self.tuple_TYPE and type(txt)!=self.list_TYPE) or len(txt)<3: return
      if len(txt)==3:
         kw={'va':'center','ha':'center'}
      else:
         kw=txt[3]
      x=txt[0]
      if rel:
         kw['transform']=ax.transAxes
      else:
         if type(x)==self.DateTime_TYPE:
            x=x.absdays+1
      t=self._checkformathtext(txt[2])
      ax.text(x,txt[1],t,kw)

   def _yfilter(self,dd,x,y,nsubfig):
      '''
         Median filter data based on 'yfilter' parameters
         (winsize,maxdiff). If winsize<3, takes median of
         the whole vector.

         Return x,y,#filtered
      '''
      if 'yfilter' not in dd: return x,y,0
      yf=dd['yfilter']
      isrel=False
      if type(yf)==self.list_TYPE:
         if yf[nsubfig]==None: return x,y,0
         try:
            wins,maxdiff,isrel=yf[nsubfig]
         except:
            try:
               wins,maxdiff=yf[nsubfig]
            except:
               return x,y,0
      else: # one tuple (not very usefull..?)
         try:
            wins,maxdiff,isrel=yf
         except:
            try:
               wins,maxdiff=yf
            except:
               return x,y,0

      leny=len(y)
      if wins<3: wins=leny

      filtered,ind,xx,yy=0,0,[],[]
      i1,i2=0,wins
      while True:
         m=plt.median(y[i1:i2])
         for i in range(i1,i2):
            dff=abs(y[i]-m)
            if dff<maxdiff and (not isrel or m==0) or isrel and m!=0 and dff/m<maxdiff:
               xx.append(x[i])
               yy.append(y[i])
            else:
               filtered+=1
         i1,i2=i2,i2+wins
         if i1>=leny: break
         if i2>leny:
            i2=leny
            i1=leny-wins

      return xx,yy,filtered
      



class cPlotXY(_cPlotBase):
   '''
      Plot (x,y) plots, where x can also be mx.DateTime or python datetime.

      ====  ==================================
      size  fig size in pixels, "width,height"
      dpi   dots-per-inch
      ====  ==================================

      Revision history:
         2009-09-27 first version (Plot())
   '''

   def __init__(self,size="800,600",dpi=100):
      '''
      '''
      _cPlotBase.__init__(self,size=size,dpi=dpi)
      self.datetime0=datetime.datetime(1,1,1)

   def Plot(self,dd={},figfiles="test.png"):
      """
      Do x,y plot based on data/metadata in dict dd (and/or x,y). Write out
      graphic file(s) named in figfiles list. Text (except for 'label') is
      plotted as mathtext if $...$, otherwise use unicode (u"text").

      Only one type of x-axis scale/subfig allowed

      Can handle missing values (== None) (??)

      ========  ================================================
      dd        data dictionary
      figfiles  (list of) result file(s), default: test.png
      x         'x': dd['x'], otherwise list (see dd['x'] below)
      y         'y': dd['y'], otherwise list (see dd['y'] below)
      ========  ================================================

      Data dict (dd) keys:

         'x'
            [[]], [[[]]] each subplots has its own x-vals (axis)
         'y'
            [[]], [[[]]] plots for left y-axis
         'x2'
            [[]], [[[]]] for each right side (y2) plot (however, there is no second "real" x-axis)
         'y2'
            [[]], [[[]]] plots for right (y2) y-axis
         'xerr'
            x error (symmetric) in x units, format: see 'x' above -> ErrorBar plotting
         'yerr'
            y error (symmetric) in y units, format: see 'y' above -> ErrorBar plotting
         'kwargs' 
            dict for 'y2' data formats passed to pylab plot command
            [[{}]]   each subplot's each plot has its own args
            [{}]     each plot has its own, but shared between subplots
            'linestyle','ls'
               '-' '--' '-.' ':' 'steps' 'None'
            'linewidth','lw'
               points
            'marker'
               '+' ',' '.' '1' '2' '3' '4' '<' '>' 'D' 'H' '^' '_' 'd' 'h' 'o' 'p' 's' 'v' 'x' '|' 'None'
            'markeredgecolor','mec'
               color
            'markeredgewidth','mew'
               points
            'markerfacecolor','mfc'
               color
            'markersize','ms'
               points
            'color','c'
               line color
            'label'
         'kwargs2':
            as 'kwargs', except for the 'y2' data
         'text'
            [(xpos,ypos,u'text',kwargs)], one text/subplot, kwargs defaults:
            {'ha':'center','va':'center'}, kwargs can have e.g. 'fontsize':size
         'textrel'
            as 'text', but in 0-1,0-1 (i.e. axis) coordinates
         'xlabel'
            ['x-axis labels'], or also ('label','option string'):
            '' use default, None: don't print
         'xlabeloptions'
            common xlabel size/option string, 'option string'
         'ylabel'
            [['y axis label(s)']], also ('label','option string')
         'ylabeloptions'
            common ylabel size/option string, 'option string'
         'xdtformatter'
            set x tick datetime formatter (do not use default)
         'yticks'
            give discrete yticks values, either (), or ((),())
         'title'
            'title', also ('title','option string')
         'titleoptions'
            default title "option string", i.e. size, font,...
         'xgrid'
            True (for default), or type (e.g. "color='r',linestyle='-',lw=2")
         'ygrid'
            [True] (for default), or type (see above). NOTE: one common grid
            style for both x,ygrid is used for one subfigure (ygrid has priority)
         'xticklabelsize', 'yticklabelsize'
            [[]]
         'xticklabelcolor', 'yticklabelcolor'
            [subfig][axno]
         'lineat'
            [(yval,'linetype')], horisontal line at y==yval, lineprop: color,
            linestyle (ls), linewidth (lw), in format (y,'"color":"gray",...)
         'line'
            [((x0,x1),(y0,y1),'linetype')]  draw a line between (x0,y0)-(x1,y1)
            lineprop: color, linestyle (ls), linewidth (lw), in format (y,'"color":"gray",...)         
         'xminmax'
            [(xmin,xmax)]
         'yminmax'
            [(ymin,ymax)]
         'yminminmaxmax'
            [(yminmin,ymaxmax)]
         'yfilter'
            [(median_window_size,max_diff,is_relative=False)]
         'yformatter'
            ['yformats']
         'ymaxnbins'
            [n], number of MaxNLocator's, default=5
         'xmaxnbins'
            [n], number of MaxNLocator's, default=5 (for != datetime)
         'ylog'
            [True/False]
         'frame'
            {'left':,'bottom':,'right':,'wspace":,'hspace':} (0-1)

      TODO:
         - add 'legend'
         - add default kwargs (?)
         - y2,x2 for EVERYTHING  (what is missing?)
      """
      msg="cPlotXY.Plot(dd,figfiles): %s"
      
      if type(dd)!=self.dict_TYPE:
         raise cPlotError(msg%"dd is not a dictionary!")
         
      if (not dd or 'x' not in dd or 'y' not in dd) or not figfiles:
         raise cPlotError(msg%"no data or fig.files given")
      
      try:
         if not dd['x'][0] or not dd['y'][0]:
            raise cPlotError(msg%"no data!?")
      except:
         raise cPlotError(msg%"no data!??")
         
#      figfls=figfiles if type(figfiles)==self.list_TYPE else list(figfiles)
      xda=dd['x']
      yda=dd['y']
#        print "shape 'x': %s, 'y': %s"%(plt.shape(xda),plt.shape(yda)) #debug#
      try:
         x2da=dd['x2']
         y2da=dd['y2']
 #        print "shape 'x2': %s, 'y2': %s"%(plt.shape(x2da),plt.shape(y2da)) #debug#
      except:
         x2da=None
         y2da=None

      xdim=self.GetListDim(xda)
      ydim=self.GetListDim(yda)
      x2dim=self.GetListDim(x2da)
      y2dim=self.GetListDim(y2da)
      
      if not xdim or not ydim:
         raise cPlotError(msg%"'x','y': not a list?")
      if xdim>3:
         raise cPlotError(msg%"'x','y': must be [], [[]] or [[[]]] (xdim was too large)")
      if xdim==1:
         nsubfigs=1
      else:
         nsubfigs=len(xda)
         if ydim<2 or y2da and y2dim<2:
            raise cPlotError("msg%('y' ('y2') must be [[]] or [[[]]]")

      xtype=[]
      x0=[]
      if xdim==1:
         x0=[xda[0]]
      else:
         for xx in xda:
            try:
               if xdim==2:
                  x0.append(xx[0])
               else:
                  x0.append(xx[0][0])
            except:
               x0.append([])
      self.xtypes=[]
      for xx in x0:
         if not xx:
            xtype.append(None)
         elif type(xx)==self.DateTime_TYPE:         # must convert (.absdays+1)
            xtype.append(2)                       # DateTime
         elif type(xx)==self.datetime_TYPE:
            xtype.append(1)                       # datetime can plotted as is
         else:
            xtype.append(0)                       # number (float)
         self.xtypes.append(xtype[-1])

      axes=self.Fig(nPlots=nsubfigs,useAxesGrid=False)
      ###
      ### plot (sub)fig(s) and find the x min,max ###
      ###
      xminmax=None
      for subfig in range(nsubfigs):
#         print "subfig:",subfig,xdim,ydim #debug
         if 'xminmax' in dd:
            xautoscale=False
            try:
               xminmax=dd['xminmax'][subfig]
               assert type(xminmax) in [self.list_TYPE,self.tuple_TYPE]
            except:
               xminmax=dd['xminmax']
            if type(xminmax[0])==self.DateTime_TYPE:
               xminmax=[xminmax[0].absdays+1,xminmax[1].absdays+1]
            elif type(xminmax[0])==self.datetime_TYPE:
               for idx in range(2):
                  deltat=xminmax[idx]-self.datetime0
                  xminmax[idx]=deltat.days+1+deltat.seconds/86400.
         else:
            xautoscale=True

         if xdim>1:
            xdata=xda[subfig]
         else:
            xdata=xda

         if ydim>1:
            ydata=yda[subfig]
         else:
            ydata=yda
         if x2da:
            try:
               if x2dim>1:
                  x2data=x2da[subfig]
               else:
                  x2data=x2da
            except IndexError:
               x2data=[]
         else:
            x2data=[]
         if y2da and x2data:
            if y2dim>1:
               y2data=y2da[subfig]
            else:
               y2data=y2da
         else:
            y2data=[]

         ax=axes[subfig]
         xax=ax        
         axno=-1
         for xdaa,ydaa in (xdata,ydata),(x2data,y2data): # through two axis
            axno+=1
            if not ydaa: 
               if axno==0:    # should format/write ylabels (no data - empty plot)
                  self._setlabel(ax,dd,'ylabel',subfig,axno)
                  self._setyminmax(ax,dd,subfig,axno)
                  self._setyformatter(ax,dd,subfig,axno)
                  self._setticklabelsize(ax,dd,'yticklabelsize',subfig,axno)
                  self._setticklabelcolor(ax,dd,'yticklabelcolor',subfig,axno)
               continue
            if type(ydaa[0])!=self.list_TYPE: ydaa=[ydaa]
            if type(xdaa[0])!=self.list_TYPE: xdaa=[xdaa]
            plotno=0 
            for xplotdata,yplotdata in zip(xdaa,ydaa):
               if len(xplotdata)!=len(yplotdata):
                  raise cPlotError(msg%("mismatch in data lens, subfig:%s, axno: %s, plotno:%s, xlen:%s, ylen:%s"%(
                     subfig,axno,plotno,len(xplotdata),len(yplotdata))))
               xx=[] ; yy=[]
               for xval,yval in zip(xplotdata,yplotdata):
                  # if yval==None or xval==None: continue --> Segmentation fault (xval comp)
                  if yval is None or xval is None:  continue
                  yy.append(yval)
                  try:
                     if xtype[subfig]==2:
                        xx.append(xval.absdays+1)
                     elif xtype[subfig]==1:
                        deltat=xval-self.datetime0
                        xx.append(deltat.days+1+deltat.seconds/86400.)
                     else:
                        xx.append(xval)
                  except IndexError:
                     xx.append(None)

               if yy:
                  if axno==1:
                     ax=xax.twinx()    # make a twinned 2nd y-axis
                  xx,yy,nfiltered=self._yfilter(dd,xx,yy,subfig)  #TODO properly!
                  if xautoscale:
                     xmm=[min(xx),max(xx)]
                     if xminmax:
                        if xmm[0]<xminmax[0]: xminmax[0]=xmm[0]
                        if xmm[1]>xminmax[1]: xminmax[1]=xmm[1]
                     else:
                        xminmax=xmm[:]
                  
                  # find the kwargs...
                  kwax=dd['kwargs']
                  if axno==1 and 'kwargs2' in dd: kwax=dd['kwargs2']
                  try:
                     kw=kwax[subfig][plotno]
                  except:
                     try:
                        kw=kwax[subfig]
                     except:
                        kw=kwax[plotno]
#                  print "a,s,p...,ax:",axno,subfig,plotno,type(kw),kw,ax  #debug#
### PLOTTING HERE ###                  
                  try:
                     #print "subfig,axno,plotno,**kw:",subfig,axno,plotno,kw  #debug
                     ax.plot(xx,yy,**kw)
                  except:
                     print "\nEXCEPTION in jahplotlib_ng, ax.plot():"
                     print "subfig,axno,plotno:",subfig,axno,plotno
                     print "shape xx,yy:",plt.shape(xx),plt.shape(yy)
                     print
                     print "kw:",kw
                     print
                     raise
               if 'label' in kw:
                  fosize=10
                  if 'xticklabelsize' in dd: fosize=dd['xticklabelsize']
                  ax.legend(loc='best',fontsize=fosize,fancybox=True,labelspacing=0.1,borderpad=0.1,
                     borderaxespad=0.2,columnspacing=0.2,handletextpad=0.03,framealpha=0.5)
               plotno+=1

            self._setlabel(ax,dd,'ylabel',subfig,axno)
            self._setyminmax(ax,dd,subfig,axno)
            self._setyformatter(ax,dd,subfig,axno)
            self._setxformatter(ax,dd,subfig,axno)
            self._setticklabelsize(ax,dd,'yticklabelsize',subfig,axno)
            self._setticklabelcolor(ax,dd,'yticklabelcolor',subfig,axno)

         # for each subfig:
         try:
            if xtype[subfig]:      # datetime or DateTime -> don't let autoscale!
               ddays=xminmax[1]-xminmax[0]
               if ddays<=0:
                  raise cPlotError("axno:%s subfig:%s, xminmax: %s DIFF IS ZERO!?"%(axno,subfig,xminmax))
               if xautoscale:
                  if ddays>=182:
                     xminmax=[int(xminmax[0])-5,int(xminmax[1])+5]
                  elif ddays>=1:
                     xmx=xminmax[1]%1
                     if xmx<.01:
                        xminmax[1]=int(xminmax[1])
                     elif xmx<=.5:
                        xminmax[1]=int(xminmax[1])+0.5
                     else:
                        xminmax[1]=int(xminmax[1])+1
                     xminmax[0]=int(xminmax[0])
                  ddays=xminmax[1]-xminmax[0]
               xdtformatter=None if 'xdtformatter' not in dd or not dd['xdtformatter'] else dd['xdtformatter']
               majLocator,minLocator,dateFormatter=self.dtLocators(ddays,xdtformatter)
            else:
               minLocator=None
               if 'xmaxnbins' not in dd:
                  majLocator=plt.AutoLocator()
               else:
                  majLocator=None
               if 'xformatter' not in dd:
                  dateFormatter=plt.ScalarFormatter()
               else:
                  dateFormatter=None
         except IndexError:
            majLocator,minLocator,dateFormatter=None,None,None
            pass

         if subfig==0:
            self._setlabel(ax,dd,'title')
         self._settext(ax,dd,subfig)

      for subfig in range(nsubfigs):
         ax=axes[subfig]
         if xminmax:
            ax.set_xlim(xminmax[0],xminmax[1])
         if majLocator:
            ax.xaxis.set_major_locator(majLocator)
            if minLocator:
               ax.xaxis.set_minor_locator(minLocator)
         if dateFormatter:
            ax.xaxis.set_major_formatter(dateFormatter)
#         else:
#            ax.xaxis.set_major_formatter(plt.NullFormatter())
         try:
            xl=dd['xlabel'][subfig]
         except:
            xl=None
         if xl!=None:
            self._setlabel(ax,dd,'xlabel',subfig)
         elif subfig==nsubfigs-1:      # scale anyway to the last subplot
            pass
         else:
            ax.xaxis.set_major_formatter(plt.NullFormatter())
         self._setticklabelsize(ax,dd,'xticklabelsize',subfig)
         self._setticklabelcolor(ax,dd,'xticklabelcolor',subfig)
         
         try:
            yt=None
            yt=dd['yticks'][subfig]
            if yt:
               try: 
                  ax.set_yticks(yt[0])
                  ax.set_yticklabels(yt[1])
               except: 
                  ax.set_yticks(yt)
         except:
            if yt:
               print yt
               raise

         self._setlineat(ax,dd,subfig)
         self._setline(ax,dd,subfig)
         self._setgrid(ax,dd,subfig)

      ### adjust plot size (if so indicated)
      if 'frame' in dd:
         plt.subplots_adjust(**dd['frame'])

      ###
      ### save the plotted figure
      ###
      if type(figfiles)==self.list_TYPE:
         for ff in figfiles:
            ff=ff.strip()
            if not ff: continue
            plt.savefig(ff,dpi=self.dpi)
      else:
         plt.savefig(figfiles,dpi=self.dpi)

if __name__=="__main__":
   DTDT=DateTime.DateTime
   x=[DTDT(2009,1,1),DTDT(2009,1,3),DTDT(2009,1,4)]
   y=[[390,385,395],[350,355,359]]
   dd={'title':(u'Nurmijärvi test','fontsize=13'),
      'xlabel':(u'date','fontsize=11'),
      'ylabel':[(r'$tää  o  tekstiä,  \Delta_i^j  \pi$','fontsize=12'),'2nd y-label'],
      'kwargs':[{'c':'red','marker':'x','mec':'darkred'},None],
      'xgrid': True,
      'ygrid': ["color='r',lw=2",True],
      'xticklabelsize':15,
      'yticklabelsize':[10,11],
      'yformatter':['%.2e','%.1f'],
      'lineat': [391,355.5],
      'text':[None,(DTDT(2009,1,2),355,u'HELLO_öäå')],
      'frame':{'left':.2,'right':.95},
      'xminmax':[datetime.datetime(2009,1,2),datetime.datetime(2009,1,3)],
      'yminmax':[None,(352,356)]
      }
   c=cPlotXY()
   c.Plot(dd,x=x,y=y)
