# -*- coding: utf-8 -*-
"""
   Common/general python routines for LGR instruments
   RESP:
      a,b factors (a=1.00*,b=0.*) are average for the whole period
      DD*c+d is time dependent correction for difference between
      C(calc with a,b)-C(eqcorr)

   Revision history:
      2013-10-14  start of devel
      2015-01-21  devel cont
      2015-01-30  WS handling changes
      2015-02-02  bug fixes, update
      2015-04-07  cLGRCalc -> CalcLGRConc.py, +cLGRCalibrations
      2015-04-08  +cCylinders
      2015-05-13  -cCylinders -> cylinders.py/cCylinders
      2015-05-27..28  updates
      2015-06-02  updates
      2015-06-03  backward uncompatible update to cLGR
      2015-10-16  updates
      2016-04-12  manual_lgr_flags_utc -> lgr_manual_flags_utc
"""
import mx.DateTime as DateTime
import os.path
from jahfilelib import ReadIniFile
from jahdblib import cDB
from numpy import average,std

INIPATH="/home/src/eric/LGR/lgrid"
CALINIPATH="/home/src/eric/LGR/lgrcal"

PARSEDT=DateTime.ISO.ParseDateTime
DTDT=DateTime.DateTime
DELTAT=DateTime.TimeDelta
ONEMIN=DELTAT(0,1)

FLOATTYPE=type(1.)
TUPLETYPE=type(())
LISTTYPE=type([])
DTTYPE=type(DateTime.now())

DBNAME="lasers"
FLAGTABLE="lgr_manual_flags_utc"
CYLTABLE="cylinders"
# put these in TIME ORDER
LGR_IDS=[
   '13-0055','14-0134'
]
LGR_TYPES=['CON2OEP']

import warnings
warnings.simplefilter("error") #debug

def GetCalParNames(instrid,onlynames=True):
   '''
      Read and return list of calibration parameter names and versions, i.e.
      [[Cal_NAME,iniparver]] if onlynames==False for INSTRID.cal.ini,
      otherwise only [Cal_NAME].
   '''
   inifile=os.path.join(CALINIPATH,"%s.cal.ini"%(instrid,))
   ini=ReadIniFile(inifile,False)
   parnames=[]
   for key in ini:
      if key[:4]!='Cal_': continue
      if onlynames:
         parnames.append(key[4:])
      else:
         parnames.append([key[4:],ini[key]['parameter_version']])
   return parnames
 


class cLGRCalibrations:
   '''
      Read & access LGR calibration data
   '''
   def __init__(self,lgrid,calparname=None):
      '''
      '''
      self.lgrid=lgrid.lower()
      if calparname:
         self.calparname=calparname
      else:
         self.calparname=GetCalParNames(lgrid)[0]
      
      inifile=os.path.join(CALINIPATH,"%s.cal.ini"%(lgrid,))
      # calibration parameters from [Cal_parname]
      # 'parname':{'compo':[pars]}
      self.calpars={}
      self.cals=[]
      self.defaults={}
      self.calids={} # calid:index to self.cals
      if not os.path.exists(inifile): return
      self.ini=ReadIniFile(inifile,False)
      ## set self.defaults{}, self.cals and self.calpars
      self._parseIni(self.ini)
      
   def GetCalStdVersions(self):
      '''
         return {'compo':int(calstdversion)}
      '''
      dct={}
      for compo in self.calpars:
         if compo in ('parameter_version',): continue
         dct[compo]=int(self.calpars[compo][1])
      return dct
      
   def GetIni(self):
      '''
      '''
      return self.ini
      
   def _parseIni(self,ini):
      '''
         self.cals: [startdt,enddt,'calname',{}]
         self.calpars: {'compo':[pars],'parameter_version':iniver}
#         self.calpars: {'parname':{'compo':[pars],'parameter_version':iniver}}
         self.default: {'name':[data]}
         
         enddt: +ONEMIN
      '''
      genini=ini['General']
      self.defaults={}
      for k in genini:
         v=genini[k]
         self.defaults[k]=self._parsev(v)           
      cals=[]
      calpars={}
      for key in ini:
         if key=='General': continue
         vs=ini[key]
         if key[:4]=='Cal_':
            calparname=key[4:]
            if calparname!=self.calparname: continue
            for k in vs:
               if k=='parameter_version':
                  calpars[k]=int(vs[k])
               else:
                  compo=k.lower()
                  assert compo in ('co','n2o')
                  ss=vs[k].split(',')
                  calpars[compo]=ss
         else:
            ss=vs['start']
            try:
               if ss.find('-')>0:  # has also date
                  dt=PARSEDT(ss)
               else:
                  dt=PARSEDT("%s %s"%(key[:10],ss))
            except:
               print
               print key,vs,ss
               raise
            cals.append([dt,None,key,{}])
            for k in vs:
               if k=='start':continue
               v=vs[k]
               cals[-1][-1][k]=self._parsev(v)
      cals.sort()
      for i,c in enumerate(cals):
         if "cyl_order" not in c[-1] or "minspercyl" not in c[-1] or "avgminutes" not in c[-1]:
            raise ValueError("cLGRCalibrations: MISSING data in %s at %16.16s"%(
               c[-1],c[0]))
         cals[i][1]=c[0]+len(c[-1]['cyl_order'])*DELTAT(0,c[-1]['minspercyl'])
         self.calids[c[2]]=i
      self.cals=cals
      self.calpars=calpars
      
   def _parsev(self,vstr):
      '''
         Return comma separated as list,
         values as int,float or str (in that order)
      '''
      if vstr.find(',')>=0:   # list
         return vstr.split(',')
      try:
         return int(vstr)
      except:
         try:
            return float(vstr)
         except:
            return vstr
            
      
   def GetIniParVersion(self):
      '''
      '''
      return self.calpars['parameter_version']
      
   def GetCalPars(self):
      '''
         return dict {'compo':[pars],'parameter_version':calstdversion}
      '''
      return self.calpars
      

   def GetScales(self,fordbwhere=False):
      '''
         return {'compo':'scale'}, or {'compo':'(component='compo' AND scale='scale')'}
      '''
      dct={}
      for compo in self.calpars:
         if compo in ('parameter_version',): continue
         dct[compo]=self.calpars[compo][0]
      if not fordbwhere: return dct
      for compo in dct:
         dct[compo]="(component='%s' AND scale='%s')"%(compo,dct[compo])
      return dct
      
           
         
   def GetCals(self,startdt,enddt):
      '''
         Return all lgrcals within startdt,enddt as
         [startdt,'calname',{key:value}]
         
         Note: calibration start defined in lgrcal must
         be >=startdt, <enddt to be returned
      '''
      ret=[]
      for c in self.cals:
         if c[0]>=enddt: break
         if c[0]<startdt: continue
         ret.append(c)
      return ret
      
   def GetLGRID(self):
      '''
      '''
      return self.lgrid
      
   def GetACal(self,calid):
      '''
      '''
      try:
         return self.cals[self.calids[calid]]
      except:
         return []
         
   def GetCalIDs(self):
      '''
         Return keys in alphabetical (time) order
      '''
      keys=self.calids.keys()
      keys.sort()
      return keys
      
   def GetDefaults(self):
      '''
         Return self.defaults dict
      '''
      return self.defaults
      
   def HasData(self):
      '''
      '''
      return len(self.cals)>0
   





class cLGR:
   '''
      General LGR utilities
   '''
   def __init__(self,lgrid,iniparver='1'):
      '''
      '''
      self.lgrid=lgrid
      self.lgrtype=self._lgrtype()
      self.inifile=os.path.join(INIPATH,"%s.ini"%lgrid)
      self.iniparver=int(iniparver)
      ## [(sdt,edt),...], {sdt:{"parname":value}}
      self.sedt,self.pars=self._parseIni()
      self.measuredcomponents=self.MeasuredComponents(nounit=True)

   def GetStation(self):
      '''
         Station should not change within the same iniparver
      '''
      return self.pars[self.sedt[0][0]]['station']

   def  GetIniParVer(self):
      return self.iniparver
      
   def  GetLGRType(self):
      '''
      '''
      return self.lgrtype

   def _parseIni(self):
      '''
         Read and parse inifile. Return content of the ini file as
         [(sdt,edt),...], {sdt:{"parname":value}}
      '''
      genpars={}
      sedt=[]
      pars={}
      
      ini=ReadIniFile(self.inifile,printmsg=False)
      for parname in ini['General']:
          genpars[parname]=self._evalval(parname,ini['General'][parname])
      for sect in ini:
         if sect=='General': continue
         if int(sect[0])==self.iniparver:
            sec=ini[sect]
            sdt,edt=sec['data_start_end'].split(',')
            sdt,edt=PARSEDT(sdt),PARSEDT(edt)
            assert sdt<edt
            sedt.append((sdt,edt))
            pars[sdt]={}
            for parname in sec:
               if sec=='data_start_end': continue
               pars[sdt][parname]=self._evalval(parname,sec[parname])
            # add from [General]
            for parname in genpars:
               if parname not in pars[sdt]:
                  pars[sdt][parname]=genpars[parname]
      if len(sedt)>1:
         sedt.sort()    # to time order
         # check that no overlapping periods
         for idx in range(len(sedt)-1):
            sdt,edt=sedt[idx]
            sdtn,edtn=sedt[idx+1]
            if edt>sdtn:   # overlap
               raise ValueError('ini-file, OVERLAPPING periods (%s,%s)'%(sedt[idx],sedt[idx+1]))
      return sedt,pars
      
   def _evalval(self,parname,parval):
      if parname[-3:]=='_dt':
         return PARSEDT(parval)
      elif parname[-7:]=='_to_utc':
         return DELTAT(int(parval))
      elif parname=='h2o':
         if parval: return eval(parval)
         return eval(DEFAULT_H2O)
      elif parname[-6:]=='_stdev' or parname=='calibration':
         return eval(parval)
      elif parname[-5:]=='_mins':     # can be a list
         return [DELTAT(0,int(x)) for x in parval.split(',')]
      elif parname[-6:]=='_ports' or parname=='nvalues_for_hourly_avg':
         return [int(x) for x in parval.split(',')]
      elif parname[-5:]=='_port':
         return int(parval)
      return parval
      


   def _lgrtype(self):
      '''
         Given LGR's id returns its type string
      '''
      if not self.lgrid: return None
      if self.lgrid in ('13-0055','14-0134'):
         return LGR_TYPES[0]
      return None
      
   def GetLGRID(self):
      '''
      '''
      return self.lgrid
      
   def MeasuredComponents(self,nounit=False):
      '''
         Return list of measured components for the LGR in question
         Names are returned as the raw table field names
         WITHOUT avg_ at the start
      '''
      ret=['co','n2o','h2o']
      if nounit: return ret
      return ['%s_ppm'%x for x in ret]


   def GetRawTableName(self):
      '''
         Return name of the db table for one-min raw results
      '''
      return "lgr_%s_1min_raw_utc"%(self.lgrid.lower().replace('-','_'),) 

   def GetResTableName(self):
      '''
         Return name of the db table for one-min results
      '''
      return "lgr_%s_1min_res_utc"%(self.lgrid.lower().replace('-','_'),)

   def GetHMeansTableName(self):
      '''
         Return name of the db table for one-min results
      '''
      return "lgr_%s_hourly_means_utc"%(self.lgrid.lower().replace('-','_'),)
      
   def GetHMeansDataDBDts(self,component,scale):
      '''
         return startdt,enddt of the data in h-means table
      '''
      db=cDB(DBNAME)
      tbl=self.GetHMeansTableName()
      chw="component='%s' AND scale='%s' AND iniparver=%s"%(component,scale,self.iniparver)
      sdt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=False,where=chw)
      edt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,where=chw)
      return sdt,edt
      
   def GetNValues4HMeans(self,dt=None):
      '''
         Return required #values (i.e. minute averages) for hourly average calculation for sample ports. 
         If dt==None:
            [[start,end,{port:nvals}]]
         or if dt is given:
            {port:nvals}
      '''
      ret=[]
      dct={}
      for sdt,edt in self.sedt:
         if dt and dt<sdt: continue
         par=self.pars[sdt]
         for p,m in zip(par['sample_ports'],par['nvalues_for_hourly_avg']):
            dct[p]=m
         if dt: return dct
         ret.append([sdt,edt,dct])
         dct={}
      return ret
      

   def DryConc(self,dt,data):
      '''
         data: [co,std(co),n2o,std(n20),h2o,std(h2o)]
         
         Return [codry,std(co),n2odry,std(n2o),h2o,std(h2o)]
         
         Note: replaces h2o<0 -> 0
         Note2: DD (delta days) not supported currently
      '''
      ret=[]
      h2ocorr=self.GetH2OCorr(dt)
      H=data[-2] if data[-2]>=0 else 0.
      idx=0
      for idx,compo in zip((0,2),self.measuredcomponents[:-1]):
         C=data[idx]
         if H is not None and C is not None:
            ret.append(eval(h2ocorr[compo]))
            ret.append(data[idx+1])
         else:
            ret.extend([None,None])
      ret.extend([H,data[-1]])
      return ret
      
   def GetH2OCorr(self,dt):
      '''
         return dict (h2o={} in ini)
      '''
      for sdt,edt in self.sedt:
         if dt<sdt: continue
         return self.pars[sdt]['h2o']
      return None
            
      
   
   def GetCalibration(self):
      '''
         return [[sdt,edt,{'compo':eq}]]
      '''
      ret=[]
      for sdt,edt in self.sedt:
         ret.append([sdt,edt,{}])
         for compo in ('co','n2o'):
            try:
               ret[-1][-1][compo]=self.pars[sdt]['calibration'][compo]
            except:
               pass
      return ret

      
   def GetSamplePorts(self,dt=None):
      '''
         Return list of sample ports in use. If dt==None:
            [[start,end,[ports]]
         or if dt is given:
            [ports]
      '''
      ret=[]
      for sdt,edt in self.sedt:
         if dt and dt<sdt: continue
         if dt: 
            return self.pars[sdt]['sample_ports']
         ret.append([sdt,edt,self.pars[sdt]['sample_ports']])
      return ret
      
   def GetWSPort(self,dt=None):
      '''
         Return WS port number
      '''
      for sdt,edt in self.sedt:
         if not dt or dt>=sdt and (edt is None or dt<edt):
            return self.pars[sdt]['ws_port']
      return None
      

   def GetRawDataDBDts(self,startdt=None,enddt=None,todataend=False):
      '''
         return (start_of_first_WS,end_of_last_WS)
         
         or if startdt and/or enddt given
         start_of_first WS before startdt,
         and/or end_of_last WS before enddt
         
         if todataend (and no enddt), return last found 
         data (not checking for WS)
      '''
      db=cDB(DBNAME)
      rtbl=self.GetRawTableName()
      
      if startdt:
         wsp=self.GetWSPort(startdt)
         if not wsp: wsp=self.GetWSPort()
         sdt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True,
            where="port=%s AND datetime_utc<='%s'"%(wsp,startdt))
         if sdt:
            sdtx=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True,
               where="port<>%s AND datetime_utc<'%s'"%(wsp,sdt))
            if sdtx: sdt=sdtx+ONEMIN
         else:    # not found - find first later
            sdt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=False,
               where="port=%s AND datetime_utc>='%s'"%(wsp,startdt))
      else:
         wsp=self.GetWSPort()
         sdt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=False,where="port=%s"%(wsp,))
         
      if todataend: 
         edt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True)
         return sdt,edt
      if enddt:
         edt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True,
            where="port=%s AND datetime_utc<='%s'"%(wsp,enddt))
         if edt:  # check that not in middle of WS
            edtx=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=False,
               where="port<>%s AND datetime_utc>='%s'"%(wsp,edt))
            if edtx: edt=edtx-ONEMIN
      else:
         edt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True,where="port=%s"%(wsp,))
      return sdt,edt
      
   
   def GetResDataDBDts(self,component,scale,startdt=None,enddt=None):
      '''
         return (start_of_first_WS,end_of_last_WS) in restbl
         where start_of_first_WS<=startdt, and end_of_last_WS<=enddt
      '''
      db=cDB(DBNAME)
      tbl=self.GetResTableName()
      wsp=self.GetWSPort(startdt)
      cwh="component='%s' AND scale='%s' AND iniparver=%s"%(
         component,scale,self.iniparver)
      
      if startdt:
         sdt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,
            where="%s AND port=%s AND datetime_utc<='%s'"%(
            cwh,wsp,startdt))
         if sdt:
            sdtx=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,
               where="%s AND port<>%s AND datetime_utc<='%s'"%(
               cwh,wsp,sdt))
            if sdtx: sdt=sdtx+ONEMIN
      else:
         sdt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=False,
            where="%s AND port=%s"%(cwh,wsp))

      if enddt:
         edt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,
            where="%s AND port=%s AND datetime_utc<='%s'"%(
            cwh,wsp,enddt))
         if edt:  # check that not in middle of WS
            edtx=db.GetDataDt(tbl,dtcolname='datetime_utc',last=False,
               where="%s AND port<>%s AND datetime_utc>='%s'"%(cwh,wsp,edt))
            if edtx: edt=edtx-ONEMIN
      else:
         edt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,
            where="%s AND port=%s"%(cwh,wsp))
      return sdt,edt
   

   def GetStartEndDates(self):
      '''
      '''
      return self.sedt
