# -*- coding: utf-8 -*-
'''
nmealib
=============

   :Release: |__version__|

   A simple (primitive) NMEA 183 parsing library. 
   
   Messages (currently) with "enhanced" processing (see cParseNMEA._processMessage)
      GGA
         Global Positioning System Fix Data. Time, Position and fix related data
         for a GPS receiver. 15 comma separated values.
         $--GGA,hhmmss.ss,llll.ll,a,yyyyy.yy,a,x,xx,x.x,x.x,M,x.x,M,x.x,xxxx*hh
            
      RMC
         Recommended Minimum Navigation Information
         $--RMC,hhmmss.ss,A,llll.ll,a,yyyyy.yy,a,x.x,x.x,xxxx,x.x,a*hh

   Revision history:
      * 2014-07-10 (0.1) first version
      
   .. |__version__| replace:: v 0.1 Jul 2014 byjah
'''
__version__="v 0.1 Jul 2014 byjah"

import operator
import mx.DateTime as DateTime

NOW=DateTime.now

RAWHEADERS={
   'GGA':
   ['Time_UTC','Latitude_ddmm.mmmm','N/S','Longitude_ddmm.mmmm','E/W',
   'GPSQuality_0/1(/2)','nSatellites','HorizPrecDil','AltitudeMSL','AltUnits',
   'GeoidalSeparation','GSUnits','AgeOfDiffData','DiffRefStation'],
   'RMC':
   ['Time_UTC','Status','Latitude_ddmm.mmmm','N/S','Longitude_ddmm.mmmm','E/W',
   'GroundSpeed_knots','TrackMade_deg_true','Date_ddmmyy','MagnVariation_deg',
   'E/W']
}
RESHEADERS={
   'GGA':
   ['Time_UTC','Latitude_DD','Longitude_DD',
   'GPSQuality_0/1(/2)','nSatellites','HorizPrecDil','AltitudeMSL_m',
   'GeoidalSeparation_m'],
   'RMC':
   ['Time_UTC','Latitude_DD','Longitude_DD',
   'GroundSpeed_knots','TrackMade_deg_true','Date_ddmmyy']
}

class cParseNMEA:
   '''
   '''
   def __init__(self,usechcksum=True):
      '''
      '''
      self.usechcksum=usechcksum
      self.Reset()
      
   def Reset(self):
      '''
      '''
      self.talker=None
      self.sentenceid=None
      self.vals=[]
      self.reslst=[]
      self.extra={}
      self.warndt=None
      
   def Parse(self,nmeastr):
      '''
      Parse NMEA string, return values as a list of strings.
      
      In case of self.usechcksum and checksum error, or otherwise malfored
      message, return None
      
      If supported sentence identifier, class instance can also
      be queried for additional data
      
      Each sentence starts with a "$" sign and ends with <CR><LF>
         $ttsss,d1,d2,....<CR><LF>
      The first two letters following the „$” are the talker identifier (GP = GPS).
      The next three characters (sss) are the sentence identifier, 
      followed by a number of data fields separated by commas, followed by 
      an optional checksum, and terminated by carriage return/line feed. 
      The data fields are uniquely defined for each sentence type.
      
      A sentence may contain up to 80 characters plus "$" and CR/LF.
      If data for a field is not available, the field is omitted, but the 
      delimiting commas are still sent, with no space between them. The checksum
      field consists of a "*" and two hex digits representing the exclusive OR 
      of all characters between, but not including, the "$" and "*". 
      '''
      ss=nmeastr.strip()
      if not ss or ss[0]!='$': return None
      s=ss.split(',')
      if s[-1][-3]=='*':
         chksum=s[-1][-2:]
         s[-1]=s[-1][:-3]
         if self.usechcksum:
            if not self._checksumOK(ss[1:-3],chksum): 
               print "Checksum error in message (%s) - skipped"%(s[0],)
               self.Reset()
               return None
      self._processMessage(s)         
      return self.vals
      
   def GetId(self):
      '''
         Return message ID
      '''
      return self.sentenceid
      
   def GetRes(self):
      '''
         Return (headerstr,resstr)
      '''
      try:
         return self.extra['res_hdr'],self.extra['res_str']
      except:
         return '',''
         
   def GetResTime(self):
      '''
      '''
      try:
         if self.sentenceid in ('GGA','RMC'):
            return self.vals[1]
         else:
            return None
      except:
         return None
      
   def _processMessage(self,msglst):
      '''
      Messages with "enhanced" processing
         GGA
            Global Positioning System Fix Data. Time, Position and fix related data
            for a GPS receiver. 15 comma separated values.
            $--GGA,hhmmss.ss,llll.ll,a,yyyyy.yy,a,x,xx,x.x,x.x,M,x.x,M,x.x,xxxx*hh
               1) Time (UTC)
               2) Latitude
               3) N or S (North or South)
               4) Longitude
               5) E or W (East or West)
               6) GPS Quality Indicator,
                  0 - fix not available,
                  1 - GPS fix,
                  2 - Differential GPS fix
               7) Number of satellites in view, 00 - 12
               8) Horizontal Dilution of precision
               9) Antenna Altitude above/below mean-sea-level (geoid)
               10) Units of antenna altitude, meters
               11) Geoidal separation, the difference between the WGS-84 earth
                  ellipsoid and mean-sea-level (geoid), "-" means mean-sea-level below ellipsoid
               12) Units of geoidal separation, meters
               13) Age of differential GPS data, time in seconds since last SC104
                  type 1 or 9 update, null field when DGPS is not used
               14) Differential reference station ID, 0000-1023
               15) Checksum
               
         RMC
            Recommended Minimum Navigation Information
            $--RMC,hhmmss.ss,A,llll.ll,a,yyyyy.yy,a,x.x,x.x,xxxx,x.x,a*hh
               1) Time (UTC)
               2) Status, V = Navigation receiver warning
               3) Latitude
               4) N or S
               5) Longitude
               6) E or W
               7) Speed over ground, knots
               8) Track made good, degrees true
               9) Date, ddmmyy
               10) Magnetic Variation, degrees
               11) E or W
               12) Checksum  
      '''
      self.vals=msglst[:]
      self.talker=msglst[0][1:3]
      self.sentenceid=msglst[0][3:]
      self.reslst=[]
      if self.sentenceid in ('GGA','RMC'):
         self.extra={'res_hdr':','.join(RESHEADERS[self.sentenceid])}
         self.reslst=[self.vals[1]]
         lalo=self._getLaLo(self.vals)
         self.reslst.extend(lalo)
         if self.sentenceid=='GGA':
            for idx in (6,7,8,9,11):
               self.reslst.append(self.vals[idx])
               if idx in (9,11) and self.vals[idx+1]!='M':
                  if not self.warndt or (NOW()-self.warndt).minutes>5:
                     print "Altitude not in meters - fix the program (hdr or conversion)!"
                     self.warndt=NOW()
         else:
            for idx in (7,8,9):
               self.reslst.append(self.vals[idx])
         self.extra['res_str']=','.join(self.reslst)
      else:
         try:
            self.extra={'res_hdr':','.join(RAWHEADERS[self.sentenceid])}
         except:
            self.extra={'res_hdr':''}
         self.extra['res_str']=','.join(self.vals[1:])

         
         
   def _getLaLo(self,vals):
      '''
         Change ddmm.mmmm -> DD (decimal degrees) with 5 decimals.
         S/W -> negative number.
         Return as a list of strings.
      '''
      la=lo=0
      if self.sentenceid=='GGA':
         idxs=(2,4)
      else:
         idxs=(3,5)
      for idx in idxs:
         v=vals[idx]
         i=v.find('.')
         if not v or i<0:
            l=''
         else:
            m=float(v[i-2:])
            l=float(v[:i-2])+m/60.
         if idx in (2,3):
            la=l
            if vals[idx+1]=='S': la=-la
         else:
            lo=l
            if vals[idx+1]=='W': lo=-lo
      if la and lo:
         return ['%.5f'%(la,),'%.5f'%(lo,)]
      else:
         return ['','']
         
         
   def _checksumOK(self,msg,chcksum):
      '''
         Return True if chcksum OK, otherwise False
      '''
      return int(chcksum,16)==reduce(operator.xor,map(ord,msg),0)
   
   def CombineWithMessage(self,cPN):
      '''
         Combine (i.e. same Time_UTC) with message instance cPN.
         Return (hdr,msgstr) - if same time not found return None
         
         Only works if sentenceid is in RESHEADERS, otherwise return None
      '''
      if self.sentenceid not in RESHEADERS or cPN.sentenceid not in RESHEADERS: 
         return None,None
      if not(self.GetResTime() and self.GetResTime()==cPN.GetResTime()):
         return None,None
      h,r=self.GetRes()
      hhlst,rrlst=[],[]
      for i,rh in enumerate(RESHEADERS[cPN.sentenceid]):
         if rh not in RESHEADERS[self.sentenceid]:
            hhlst.append(rh)
            rrlst.append(cPN.reslst[i])
      if not hhlst:     # no change -> identical instances
         return h,r
      return "%s,%s"%(h,','.join(hhlst)),"%s,%s"%(r,','.join(rrlst))
         
      

def test():
   '''
   '''
   print "\n\nTesting cParseNMEA...\n"
   m1="$GPGGA,101217.000,6012.2373,N,02457.6688,E,1,08,0.9,65.1,M,19.5,M,,0000*62\r\n"
   m1e="$GPGGA,101217.000,6012.2373,N,02457.6688,E,1,08,0.9,65.1,M,19.5,M,,0000*61\r\n"
   m2="$GPRMC,101217.000,A,6012.2373,N,02457.6688,E,0.12,13.69,080714,,,A*5A\r\n"
   m3="$GPGSA,A,3,04,17,25,24,12,01,06,02,,,,,1.6,0.9,1.3*3B\r\n"

   
   c1=cParseNMEA()
   c2=cParseNMEA()
   for m in (m1,m1e,m2,m3):
      print "Msg:'%s'"%(m.strip(),)
      print "--> %s"%(c1.Parse(m),)
      print c1.GetResTime(),c1.GetRes()
      print
   print "Testing combine..."
   c1.Parse(m1)
   c2.Parse(m2)
   print "'%s'+'%s':"%(m2,m1)
   print c2.CombineWithMessage(c1)

   
if __name__=='__main__':
   test()
      
      
      
