# -*- coding: utf-8 -*-
"""
   Common/general python routines for picarro WS-CRDS instruments
      
   All picarro values in + out in ppm's!!!!!!

   Revision history:
      2015-08-2x  Devel start based on previous picarro.py
      2015-08-31  Some modifications and additions
      2015-09-18  cPicarroCalibrations(calparname by default None) if not needed (i.e. for flagging)
      2015-09-22  if not h2o= in ini, use default cal. factors
      2015-09-23  +cPicarroWS
      2015-09-24  update & bug fixes
      2015-09-25  +GetPicarroPorts()
      2015-09-30  bug fixes
      2015-10-05  WS handling bug fixe, enhancements
      2015-10-28  minor bug fix
      2016-02-24  cPicarro.GetWSPort(dt=None) -> use last WS port 
      2016-04-14  manual_picarro_flags_utc -> picarro_manual_flags_utc, GetWS() fix
      2016-05-02  +cfkads2135, +IsCalCyl
      2016-05-16  cal_cyls__COMPO update
      2016-05-17  +cPicarro._checkSETimes()
      2016-05-24..25  fixes
      2016-05-31  wsreldifflimits 
      2016-06-28  -cPicarro._checkSETimes()
      2016-08-16  bug fix in cPicarro.GetFlusFlowFlaggingConditions() (did not work correctly)
      2016-12-23  +cfads2135, "convert_ports", cPicarro.GetConvertPorts()
      2017-01-23  numpy.average -> numpy.nanmean, nanstd for some
      2017-01-26  + some __repr__(self), bug fix in _readWSData(), WS extrapolation: max 6 h
      2017-01-30  +GetPicarrosAtStation()
      2017-02-07  +compo= to some methods
      2017-02-09  -GetWS*() (-> cCylinders), wsreldifflimits->wsdifflimits
      2017-02-20  -GetPicarrosAtStation() (-> instruments.GetInstrumentsAtStation())
      2017-09-19  +cfkads2164
      2018-01-10  cPicarro: GetPars(),GetZeroDt()
      2018-01-22  cPicarroCalibrations, minspercyl can be also a list (instead of single value)
"""
import mx.DateTime as DateTime
import os,glob
from collections import OrderedDict
from jahfilelib import ReadIniFile
import cylinders
from jahdblib import cDB
import numpy as np
#from collections import OrderedDict

DEBUG=0

INIPATH="/home/src/eric/picarro/picarroid"
CALINIPATH="/home/src/eric/picarro/picarrocal"

PARSEDT=DateTime.ISO.ParseDateTime
DTDT=DateTime.DateTime
DELTAT=DateTime.TimeDelta
FROMABSDAYS=DateTime.DateTimeFromAbsDays
NOW=DateTime.now
GMT=DateTime.gmt
ONEMIN=DELTAT(0,1)
EXTRAPOLATIONLIMIT=0.25    # no of days

DBNAME="lasers"
FLAGTABLE="picarro_manual_flags_utc"

PICARRO_IDS=[
   # sam
   'cfads039','cfads058','cfads098','cfads100',
   # sod 1st pic
   'cfads2237',
   # krova grad
   'cfads2135',
   # kum,uto,sod,sam,uto,mob
   'cfkads2018','cfkads2035','cfkads2058','cfkads2066','cfkads2083','cfkads2095',
   'cfkads2135',
   # sam icos
   'cfkads2164'
]

PICARRO_TYPES=['G2401','G2301','G1301']

DEFAULT_H2O="{'co2':'C/(1-0.012*H-2.674e-4*H*H)','ch4':'C/(1-0.00982*H-2.393e-4*H*H)','co':'C'}"


############################

def GetCalParNames(picid,onlynames=True):
   '''
      Read and return list of calibration parameter names and versions, i.e.
      [[Cal_NAME,iniparver]] if onlynames==False for PICID.cal.ini,
      otherwise only [Cal_NAME].
   '''
   inifile=os.path.join(CALINIPATH,"%s.cal.ini"%(picid,))
   ini=ReadIniFile(inifile,False)
   parnames=[]
   for key in ini:
      if key[:4]!='Cal_': continue
      if onlynames:
         parnames.append(key[4:])
      else:
         parnames.append([key[4:],ini[key]['parameter_version']])
   return parnames

def GetPicarroPorts(picid=None,addall=False,hmeanstbl=False,rawtable=False):
   '''
      Return {'picid':[ports]} for the [picid], or for all in PICARRO_IDS
      if picid==None
      
      Revision history:
         2015-08-17  +db.rollback()
         2015-09-25  +"where flag",  copied from picarro.py
   '''
   if picid:
      if type(picid)!=type([]):
         picids=[picid]
      else:
         picids=picid
   else:
      picids=PICARRO_IDS
   db=cDB(DBNAME)
   cu=db.cursor()
   ret={}
   for pid in picids:
      cpic=cPicarro(pid)
      if not hmeanstbl:
         if not rawtable:
            t=cpic.GetResTableName()
         else:
            t=cpic.GetRawTableName()
      else:
         t=cpic.GetHMeansTableName()
      try:
         sq="""SELECT DISTINCT port FROM "%s" WHERE flag IN (' ','.','C') ORDER BY port;"""%(t,)
         cu.execute(sq)
         da=cu.fetchall()
         d=[x[0] for x in da]
         if addall:
            ret[pid]=['all']+d
         else:
            ret[pid]=d
      except:
         ret[pid]=['']
         db.rollback()
   cu.close()
   db.close()
   return ret
   

class cPicarro:
   '''
   '''
   def __init__(self,picid,iniparver='1'):
      '''
      '''
      self.picid=picid.lower()
      self.pictype=self._pictype()
      self.inifile=os.path.join(INIPATH,"%s.ini"%(self.picid,))
      self.iniparver=int(iniparver)
      if not 0 <= self.iniparver <=9: 
         raise ValueError("\nini-file: only version numbers 0..9 supported")
      ## [(sdt,edt),...], {sdt:{"parname":value}}
      self.sedt,self.pars=self._parseIni()
      self.measuredcomponents=self.MeasuredComponents()     # includes units, with h2o
      self.compos=[x[:x.find('_')] for x in self.measuredcomponents[:-1]]
      
   def __repr__(self):
      '''
      '''
      return "cPicarro(%s,%s)"%(self.picid,self.iniparver)
      
   def GetIniParVer(self):
      '''
      '''
      return self.iniparver
      
   def GetPars(self):
      '''
      '''
      return self.pars

   def GetStation(self):
      '''
         Station should not change within the same iniparver
      '''
      return self.pars[self.sedt[0][0]]['station']
      

   def GetStartEndDates(self):
      '''
      '''
      return self.sedt
   
   def GetMatchingSEDTs(self,dasedt,adddays=None):
      '''
         Return [[startdt,enddt]] for data periods within dasedt (==(sdt,edt))
         
         If adddays is not None:
            subract/add adddays from/to startdt/enddt but not beyond data limits
      '''
      if DEBUG: print "GetMatchingSEDTs((%16.16s,%16.16s),%s)"%(dasedt[0],dasedt[1],adddays)
      ret=[]
      psdt,pedt=dasedt
      dds=None if adddays is None else int(adddays)
      for s,e in self.sedt:
         if e<=psdt: continue
         if s>=pedt: break
         ret.append([max(s,psdt),min(e,pedt)])
         if dds:
            if ret[-1][0]!=s: ret[-1][0]=max(s,ret[-1][0]-dds)
            if ret[-1][1]!=e: ret[-1][1]=min(e,ret[-1][1]+dds)
      return ret
   
   def GetPicarroID(self):
      '''
      '''
      return self.picid
   def GetPicarroType(self):
      '''
      '''
      return self.pictype
      
   def GetRawTableName(self):
      '''
         Return name of the db table for one-min raw avgs
      '''
      return "picarro_%s_1min_raw_utc"%(self.picid.lower().replace('-','_'),)

   def GetResTableName(self):
      '''
         Return name of the db table for one-min results
      '''
      return "picarro_%s_1min_res_utc"%(self.picid.lower().replace('-','_'),)
      
   def GetHMeansTableName(self):
      '''
         Return name of the db table for one-min results
      '''
      return "picarro_%s_hourly_means_utc"%(self.picid.lower().replace('-','_'),)

   def GetAuxTableName(self,dt=None):
      '''
         Return name of the db table for aux data, None if not defined
      '''
      pars=self.pars[self._getsdt(dt)]
      if 'aux_data_files' in pars and pars['aux_data_files']:
         return "picarro_%s_aux_utc"%(self.picid,)
      return None
   
   def GetConvertPorts(self,dt=None):
      '''
         Return conversion dict (if any)
      '''
      pars=self.pars[self._getsdt(dt)]
      if pars and 'convert_ports' in pars: return pars['convert_ports']
      return None
      
   def GetAuxFlaggingConditions(self,startdt=None,enddt=None):
      '''
         Return list of flagging conditions based on aux data table:
         [[startdt,enddt,auxcond,rawcond,minsafter]]
         
         2017-02-09..10: +minsafter, bug fix
      '''
      startt=self.sedt[0][0] if not startdt else startdt
      if type(startt)==type(''): startt=PARSEDT(startt)
      endt=self.sedt[-1][1] if not enddt else enddt
      if type(endt)==type(''): endt=PARSEDT(endt)
      res=[]
      for sdt,edt in self.sedt:
         if edt<=startt: continue
         if sdt>=endt: break
         par=self.pars[sdt]
         for k in par:
            if k.find('aux_flagging_condition_')==0:
               cond=par[k].split(',')
               if not cond or len(cond)==0: continue
               if len(cond)==1: cond.append('')
               if len(cond)<3: cond.append('1')
               r=[sdt,edt,cond[0],cond[1],cond[2]]
               if r not in res: res.append(r)
      return res
      
   def GetFlushFlowFlaggingConditions(self,startdt=None,enddt=None):
      '''
         Return list of flagging conditions based on flushing flow ('flush_flow_flagging_condition_#':
         [[startdt,enddt,colname,limit,minsafter,port]]
      '''
      startt=self.sedt[0][0] if not startdt else startdt
      if type(startt)==type(''): startt=PARSEDT(startt)
      endt=self.sedt[-1][1] if not enddt else enddt
      if type(endt)==type(''): endt=PARSEDT(endt)
      res=[]
      for sdt,edt in self.sedt:
         lateststart=max(sdt,startt)
         earliestend=min(edt,endt)
         if earliestend<=lateststart: continue
         par=self.pars[sdt]
         for k in par:
            if k.find('flush_flow_flagging_condition_')==0:
               cond=par[k].split(',')
               if not cond or len(cond)<6: continue
               s,e=PARSEDT(cond[1]),PARSEDT(cond[2])
               r=[max(s,sdt,startt),min(e,edt,endt),cond[0],float(cond[3]),int(cond[4]),cond[5]]
               if r not in res: 
                  res.append(r)
      return res
      
      
      
   def _getsdt(self,dt):
      '''
         Get sdt to index self.pars
      '''
      if not dt:     # return last
         return self.sedt[-1][0]
      for sdt,edt in self.sedt:
         if sdt <= dt < edt: return sdt
      return None
      

   def _evalval(self,parname,parval):
      if parname[-3:]=='_dt':
         return PARSEDT(parval)
      elif parname[-7:]=='_to_utc':
         return DELTAT(int(parval))
      elif parname=='h2o':
         if parval: return eval(parval)
         return eval(DEFAULT_H2O)
      elif parname[-6:]=='_stdev' or parname=='calibration':
         return eval(parval)
      elif parname[-5:]=='_mins':     # can be a list
         return [DELTAT(0,int(x)) for x in parval.split(',')]
      elif parname=='convert_ports':
         return eval(parval)
      elif parname[-6:]=='_ports' or parname=='nvalues_for_hourly_avg':
         return [int(x) for x in parval.split(',')]
      return parval

   def _parseIni(self):
      '''
         Return content of the ini-file as
         [(sdt,edt),...], {sdt:{"parname":value}}
      '''
      genpars,sedt,pars={},[],{}
      
      ini=ReadIniFile(self.inifile,False)
      for parname in ini['General']:
          genpars[parname]=self._evalval(parname,ini['General'][parname])
      for sect in ini:
         if sect=='General': continue
         if int(sect[0])==self.iniparver:
            sec=ini[sect]
            sdt,edt=sec['data_start_end'].split(',')
            sdt,edt=PARSEDT(sdt),PARSEDT(edt)
            assert sdt<edt
            sedt.append((sdt,edt))
            pars[sdt]={}
            for parname in sec:
               if sec=='data_start_end': continue
               pars[sdt][parname]=self._evalval(parname,sec[parname])
            # add from [General]
            for parname in genpars:
               if parname not in pars[sdt]:
                  pars[sdt][parname]=genpars[parname]
      if len(sedt)>1:
         sedt.sort()    # to time order
         # check that no overlapping periods
         for idx in range(len(sedt)-1):
            sdt,edt=sedt[idx]
            sdtn,edtn=sedt[idx+1]
            if edt>sdtn:   # overlap
               raise ValueError('ini-file, OVERLAPPING periods (%s,%s)'%(sedt[idx],sedt[idx+1]))
      return sedt,pars
      
   def MeasuredComponents(self,onlynameswithouth2o=False):
      '''
         Return list of measured (GHG) components for the 
         picarro picid in order of co2,ch4[,co],h2o
         Names are returned as the raw table field names
      '''
      ret=['co2_ppm','ch4_ppm']
      if self.pictype in ('G2401',):
         ret.append('co_ppm')
      if onlynameswithouth2o:
         return [x[:-4] for x in ret]
      ret.append('h2o_rep_percent')
      return ret
      
   def _pictype(self):
      '''
      '''
      if self.picid[:7]=='cfkads2':
         return 'G2401'
      elif self.picid[:6]=='cfads2':
         return 'G2301'
      elif self.picid[:5]=='cfads':
         return 'G1301'
      raise ValueError("cPicarro._pictype, undetermined picarro type for id: %s"%(self.picid,))
      
   def GetRawDataDBDts(self):
      '''
         Return rawdatastartdt,rawdataenddt so that they are within
         ini-file determined limits
      '''
      db=cDB(DBNAME)
      rtbl=self.GetRawTableName()
      sdt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=False,where="datetime_utc>='%s'"%(self.sedt[0][0],))
      edt=db.GetDataDt(rtbl,dtcolname='datetime_utc',last=True,where="datetime_utc<='%s'"%(self.sedt[-1][1],))
      return self._withinSEDT(sdt),self._withinSEDT(edt)
      
   def _withinSEDT(self,dt):
      '''
         Check that dt is withing self.sedt, if not adjust upwards (if possible)
         and return possibly adjusted time
      '''
      if not dt: return dt
      i=-1
      for sdt,edt in self.sedt:
         i+=1
         if dt>edt: continue
         if dt<sdt: break
         if sdt<=dt<=edt: return dt
      if dt>self.sedt[-1][1]: return self.sedt[-1][1] # last possible
      if dt<self.sedt[0][0]: return self.sedt[0][0]   # first possible
      return self.sedt[i-1][1]
      
      
   def GetResDataDBDts(self,component,scale):
      '''
         Return resdatastart,resdataenddt
      '''
      db=cDB(DBNAME)
      tbl=self.GetResTableName()
      cwh="component='%s' AND scale='%s' AND iniparver=%s"%(
         component,scale,self.iniparver)
      sdt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=False,where="datetime_utc>='%s' AND %s"%(
         self.sedt[0][0],cwh))
      edt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,where="datetime_utc<='%s' AND %s"%(
         self.sedt[-1][1],cwh))
      return self._withinSEDT(sdt),self._withinSEDT(edt)
      
   def GetHMeansDataDBDts(self,component,scale):
      '''
         return startdt,enddt of the data in h-means table
      '''
      db=cDB(DBNAME)
      tbl=self.GetHMeansTableName()
      chw="component='%s' AND scale='%s' AND iniparver=%s"%(component,scale,self.iniparver)
      sdt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=False,where=chw)
      edt=db.GetDataDt(tbl,dtcolname='datetime_utc',last=True,where=chw)
      return sdt,edt

   def GetValveFiles(self,sdt,edt):
      '''
         Return list of files,toutc [filename],[DELTAT]
      '''
      return self.GetRawDataFiles(sdt,edt,'valve')
      
   def GetAuxDataFiles(self,sdt,edt):
      '''
         Return list of files,toutc [filename],[DELTAT]
      '''
      return self.GetRawDataFiles(sdt,edt,'aux')

   def GetRawDataFiles(self,sdt,edt,rawvalve='raw'):
      '''
         Return list of files,toutc [filename],[DELTAT]
      '''
      if sdt>edt:
         raise ValueError("cPicarro.GetRawDataFiles(%s,%s,%s), start>end!?"%(sdt,edt,rawvalve))
      if rawvalve=='raw':
         dt=DTDT(sdt.year,sdt.month,sdt.day)-1
         enddt=edt
      else:
         dt=sdt-1
         enddt=edt+1
      p=0
      fls,toutc=[],[]
      pfile,fmask,tou=None,None,None
      while dt<enddt:
         if self.sedt[p][1] and dt>self.sedt[p][1]:
            p+=1
            if p>=len(self.sedt): break
            fmask=None
         if not fmask:
            try:
               fmask=self.pars[self.sedt[p][0]]['%s_data_files'%(rawvalve,)]
               tou=self.pars[self.sedt[p][0]]['%s_data_to_utc'%(rawvalve,)]
            except:
               print dt
               print self.sedt
               print p
               print self.sedt[p]
               print rawvalve
               raise
         fl=dt.strftime(fmask)
         if fl!=pfile:
            files=glob.glob(fl)
            files.sort()
            for file in files:
               fls.append(file)
               toutc.append(tou)
            pfile=fl
         dt+=1
      return fls,toutc
      
   def GetSamplePorts(self,dt=None):
      '''
         Return list of sample ports in use. If dt==None:
            [[start,end,[ports]]
         or if dt is given:
            [ports]
      '''
      ret=[]
      for sdt,edt in self.sedt:
         if dt:
            if dt<sdt: continue
            return self.pars[sdt]['sample_ports']
         ret.append([sdt,edt,self.pars[sdt]['sample_ports']])
      return ret
      
   def GetZeroDt(self,dt=None):
      '''
         Return of zero_dt=s
         If dt==None:
            [[start,end,zerodt]]
         else:
            zerodt
      '''
      ret=[]
      for sdt,edt in self.sedt:
         if dt:
            if dt<sdt: continue
            return self.pars[sdt]['zero_dt']
         ret.append([sdt,edt,self.pars[sdt]['zero_dt']])
      return ret
         
      
   def GetNValues4HMeans(self,dt=None):
      '''
         Return required #values (i.e. minute averages) for hourly average calculation for sample ports. 
         [[start,end,{port:nvals}]]
      '''
      ret=[]
      dct={}
      for sdt,edt in self.sedt:
         if dt and dt<sdt: continue
         par=self.pars[sdt]
         for p,m in zip(par['sample_ports'],par['nvalues_for_hourly_avg']):
            dct[p]=m
         if dt: 
            return [[sdt,edt,dct]]
         ret.append([sdt,edt,dct])
         dct={}
      return ret
            
      
   def GetH2OCorr(self,dt):
      '''
         return dict (h2o={} in ini)
      '''
      for sdt,edt in self.sedt:
         if dt<sdt: continue
         return self.pars[sdt]['h2o']
      return None
      
      
   def GetFlushTimes(self,dt=None):
      '''   
         Return sampleflush,cylinderflush as
         [[sdt,edt,[DELTAT(mins)]]],[[sdt,edt,[DELTAT(mins)]]]
      '''
      reta=[] ; retb=[]
      for sdt,edt in self.sedt:
         if dt and dt<sdt: continue
         if dt:
            return self.pars[sdt]['sample_flush_mins'],self.pars[sdt]['cylinder_flush_mins']
         reta.append([sdt,edt,self.pars[sdt]['sample_flush_mins']])
         retb.append([sdt,edt,self.pars[sdt]['cylinder_flush_mins']])
      return reta,retb

   def GetStdevLimits(self,dt=None):
      '''
         Return flag_cylinder_stdev,flag_sample_stdev dictionaries if dt given.
         Otherwise return [{startdt,enddt,{}]
      '''
      reta,retb=[],[]
      for sdt,edt in self.sedt:
         if dt and dt<sdt: continue
         if dt:
            return self.pars[sdt]['flag_cylinder_stdev'],self.pars[sdt]['flag_sample_stdev']
         reta.append([sdt,edt,self.pars[sdt]['flag_cylinder_stdev']])
         retb.append([sdt,edt,self.pars[sdt]['flag_sample_stdev']])
      return reta,retb
            

   def DryConc(self,dt,data,stdev=True):
      '''
         data:
            in order of MeasuredComponents (last two H2O): avg,stdev pairs
            i.e. co2,co2stdev,ch4,ch4stdev,[co,costdev],h2o,h2ostdev
         
         Return as [c1,stdev(c1),...,h2o,stdev(h2o)]
         
         Note DD (delta days) not supported (currently)
         
         2017-02-03  +stdev= 
      '''
      ret=[]
      h2ocorr=self.GetH2OCorr(dt)
      H=data[-2] if stdev else data[-1]
      idx=0
      for compo in self.compos:
         C=data[idx]
         if H is not None and C is not None:
            ret.append(eval(h2ocorr[compo]))
            if stdev: ret.append(data[idx+1])
         else:
            if stdev:
               ret.extend([None,None])
            else:
               ret.append(None)
         idx+=1+(stdev==True)
      if stdev:
         ret.extend([H,data[-1]])
      else:
         ret.append(H)
      return ret
      
   def GetCalibration(self):
      '''
         return [[sdt,edt,co2eq,ch4eq[,coeq]]]
      '''
      ret=[]
      for sdt,edt in self.sedt:
         ret.append([sdt,edt])
         for compo in ('co2','ch4','co'):
            try:
               ret[-1].append(self.pars[sdt]['calibration'][compo])
            except:
               pass
      return ret
      
      
class cPicarroCalibrations:
   '''
      Read & access picarro "manual" calibration data
      (i.e. in a file under CALINPATH)
   '''
   def __init__(self,picid,calparname=None):
      '''
      '''
      self.picid=picid.lower()
      inifile=os.path.join(CALINIPATH,"%s.cal.ini"%(picid,))
      #   self.calpars: {'compo':[pars],'parameter_version':iniver}
      #   self.cals: [startdt,enddt,'calname',{}]
      #   self.default: {'name':[data]}
      if calparname:
         self.calparname=calparname
      else:
         self.calparname=GetCalParNames(picid)[0]
      self.calpars={}
      self.cals=[]
      self.calids={} # calid:index to self.cals
      if not os.path.exists(inifile): return
      self.ini=ReadIniFile(inifile,False)
      ## set self.cals and self.calpars
      self._parseIni(self.ini)
      self.iniparver=self.GetIniParVersion()
      ## cCylinders(station,scales)
      self.cpic=cPicarro(self.picid,self.iniparver)
      self.componames=self.cpic.MeasuredComponents(onlynameswithouth2o=True)
      self.ccyls=cylinders.cCylinders(self.cpic.GetStation(),self.GetScales(),self.GetCalStdVersions(),self.picid)

   def __repr__(self):
      '''
      '''
      return "cPicarroCalibrations(%s,%s(?))"%(self.picid,self.calparname)
      
   def GetPicarroID(self):
      '''
      '''
      return self.picid
      
   def _parseIni(self,ini):
      '''
         self.cals: [startdt,enddt,'calname',{}]
         self.calpars: {'compo':[pars],'parameter_version':iniver}
         self.default: {'name':[data]}

         enddt: +ONEMIN
      '''
      genini=ini['General']
      defaults={}
      for k in genini:
         v=genini[k]
         defaults[k]=self._parsev(v)           
      cals=[]
      calpars={}
      
      for key in ini:
         if key=='General': continue
         vs=ini[key]
         if key[:4]=='Cal_':     # calibration parameters
            calparname=key[4:]
            if calparname!=self.calparname: continue
            for k in vs:
               if k=='parameter_version':
                  calpars[k]=int(vs[k])
               else:
                  compo=k.lower()
                  assert compo in ('co','co2','ch4')
                  ss=vs[k].split(',')
                  calpars[compo]=ss
         else:                # individual calibrations
            ss=vs['start']
            try:
               if ss.find('-')>0:  # has also date
                  dt=PARSEDT(ss)
               else:
                  dt=PARSEDT("%s %s"%(key[:10],ss))
            except:
               print
               print key,vs,ss
               raise
            cals.append([dt,None,key,dict(defaults)]) # copy defaults dict (not deepcopy)
            for k in vs:
               if k=='start':continue
               v=vs[k]
               cals[-1][-1][k]=self._parsev(v)
      cals.sort()
      
      for i,c in enumerate(cals):
         if "cyl_order" not in c[-1] or "minspercyl" not in c[-1] or "avgminutes" not in c[-1]:
            print c
            raise ValueError("cPicarroCalibrations: MISSING data in %s at %16.16s"%(
               c[-1],c[0]))
         try:
            cals[i][1]=c[0]+len(c[-1]['cyl_order'])*DELTAT(0,c[-1]['minspercyl'])
         except:
            cals[i][1]=c[0]+DELTAT(0,sum([int(x) for x in c[-1]['minspercyl']]))
         self.calids[c[2]]=i
      self.cals=cals
      self.calpars=calpars
      
   def _parsev(self,vstr):
      '''
         Return comma separated as list,
         values as int,float or str (in that order)
      '''
      if vstr.find(',')>=0:   # list
         return vstr.split(',')
      try:
         return int(vstr)
      except:
         try:
            return float(vstr)
         except:
            return vstr
   
   def GetCalParName(self):
      '''
      '''
      return self.calparname
      
   def GetIniParVersion(self):
      '''
      '''
      return self.calpars['parameter_version']
      
   def GetCalPars(self):
      '''
         return dict {'compo':[pars],'parameter_version':calstdversion}
      '''
      return self.calpars
      
   def GetScales(self,fordbwhere=False):
      '''
         return {'compo':'scale'}, or {'compo':'(component='compo' AND scale='scale')'}
      '''
      dct={}
      for compo in self.calpars:
         if compo in ('parameter_version',): continue
         dct[compo]=self.calpars[compo][0]
      if not fordbwhere: return dct
      for compo in dct:
         dct[compo]="(component='%s' AND scale='%s')"%(compo,dct[compo])
      return dct
      
   def GetCalStdVersions(self):
      '''
         return {'compo':int(calstdversion)}
      '''
      dct={}
      for compo in self.calpars:
         if compo in ('parameter_version',): continue
         dct[compo]=int(self.calpars[compo][1])
      return dct
   
   
   def GetCals(self,startdt,enddt):
      '''
         Return all piccals within startdt,enddt as
         [startdt,enddt,'calname',{key:value}]
         
         Note: calibration start defined in piccal must
         be >=startdt, <enddt to be returned
      '''
      ret=[]
      for c in self.cals:
         if c[0]>=enddt: break
         if c[0]<startdt: continue
         ret.append(c)
      return ret
      
   def GetACal(self,calid):
      '''
      '''
      try:
         return self.cals[self.calids[calid]]
      except:
         return []
         
   def GetCalIDs(self):
      '''
         Return keys in alphabetical (time) order
      '''
      keys=self.calids.keys()
      keys.sort()
      return keys
           
   def HasData(self):
      '''
      '''
      return len(self.cals)>0
      
   def IsCalCyl(self,cal,cylid,compos=[]):
      '''
         Return True if cylid is calibration cylinder for components named
         in compos. Or if compos is left empty, if cal cyl for some component.
      '''
      dct=cal[-1]
      componames=compos if compos else self.componames
      for co in componames:
         k='cal_cyls__%s'%(co,)
         if k not in dct:
            k='cal_cyls__%s'%(co.upper(),)
         if k in dct:
            if cylid in dct[k]: return True
         else:
            if cylid in dct['cal_cyls']: return True
      return False
      
   def GetCalData(self,calname,iswscal,showdebug=False):
      '''
         Return data for calname as 
         [[dt_utc,cylid,port,isws,avg_x,stdev_x...]] 
         as DRY values. Water vapor (with stdev) is the last pair of values.
         dt_utc is at the end of the period (last measurement)
         
         Add "skip" handling?
      '''
      ret=[]
      cal=self.GetACal(calname)
      if not cal: return []   # no data
      # cal==[sdt,edt,calid,{}]
      cmins=cal[-1]['avgminutes']
      try:
         mmins=DELTAT(0,cal[-1]['minspercyl'])
      except:
         mmins=[DELTAT(0,int(x)) for x in cal[-1]['minspercyl']]
      rawtbl=self.cpic.GetRawTableName()
      compos=self.cpic.MeasuredComponents()
      
      db=cDB(DBNAME)
      cu=db.cursor()
      compolist=[]
      for co in compos:
         compolist.append('avg_%s,stdev_%s'%(co,co))
      dt=cal[0]
      nowsfound=True
         
      for cylid in cal[-1]['cyl_order']:
         try:
            dte=dt+mmins-ONEMIN
         except:
            dte=dt+mmins[cal[-1]['cyl_order'].index(cylid)]-ONEMIN
         isws=False
         if cal[-1]['ws_port']:
            try:
               if self.ccyls.GetCylPortFromID(cylid,dt)==cal[-1]['ws_port']: isws=True
            except:
               pass
         else:
            isws=self.ccyls.IsWS(dt,cylid)
         if isws: 
            nowsfound=False
         if iswscal:
            if not self.IsCalCyl(cal,cylid) and not isws:
               dt=dte+ONEMIN
               continue
         sqls='''SELECT port,%s FROM "%s" WHERE datetime_utc>='%s' AND datetime_utc<='%s'
            AND flag IN ('C','F') ORDER BY datetime_utc;'''%(','.join(compolist),rawtbl,dt,dte)
            #AND flag='C' ORDER BY datetime_utc;'''%(','.join(compolist),rawtbl,dt,dte)
         if showdebug: print "\n%s"%(sqls,)
         cu.execute(sqls)
         if cu.rowcount==0:
            print "no data? cylid %s: %16.16s - %16.16s,  check flagging"%(cylid,dt,dte)
            print sqls
            dt=dte+ONEMIN
            continue
         dd=cu.fetchall()
         if showdebug: print "--> len:%s, data:%s"%(len(dd),dd)
         if len(dd)>cmins:    # if more data returned than required, shorten it
            dd=dd[-cmins:]
         rs=[]
         pport=None
         for d in dd:  # do not include h2o
            if pport is None:
               pport=d[0]
            elif d[0]!=pport:
               print "\nWarning: cPicarroCalibrations.GetCalData(): port changed in middle of calibration:\n%s"%(sqls,)
            rs.append(self.cpic.DryConc(dte,d[1:]))
         if rs:
            avgs=np.average(rs,0)
            stds=np.std(rs,0,ddof=1)
            ret.append([dte,cylid,pport,isws])
            for idx in range(0,len(avgs),2):
               ret[-1].extend([avgs[idx],stds[idx]])
         dt=dte+ONEMIN
      if nowsfound:
         print "\n!!! NOTE !!!\nGetCalData(%s,%s) no WS measurement!"%(calname,iswscal)
         #raise ValueError("\ncPicarroCalibrations.GetCalData(%s,%s) - NO WS found, check cylinders table & WS port!",
         #   calname,iswscal)
      return ret



class cPicarroWS:
   '''
      WS measurements/results
   '''
   def __init__(self,cpic,ccyls,sedt,wsdifflimits=None):
      '''
         Get WS data for the period
         
         2017-02-07:
            reldifflimits a dict (see self.wsreldifflimits below)
            -> wsdifflimits (absolute, not relative)
      '''
      if DEBUG: print "cPicarroWS.__init__(%s,%s,(%16.16s,%16.16s),%s)"%(
         cpic,ccyls,sedt[0],sedt[1],wsdifflimits)
      self.cpic=cpic
      self.ccyls=ccyls
      self.sedt=sedt
      self.datasedt=cpic.GetMatchingSEDTs(sedt,1)
      self.wseqsidx=0
      self.ncompos=len(self.cpic.MeasuredComponents(True))
      self.compoidxdct=OrderedDict({'co2':0,'ch4':1})
      if len(cpic.MeasuredComponents(True))==3: self.compoidxdct['co']=2 
      self.wsdifflimits={'co2':.3,'ch4':.01,'co':.03}
      if wsdifflimits:
         self.wsdifflimits.update(wsdifflimits)
      self.wseqs=self._readWSData()

         
   def __repr__(self):
      '''
      '''
      return "cPicarroWS(cPicarro,cCylinders,(%16.16s,%16.16s),%s)"%(
         self.sedt[0],self.sedt[1],self.wsdifflimits)
         
      
   def _readWSData(self):
      '''
      '''
      MINLIMS={'co2_ppm':330,'ch4_ppm':1.4,'co_ppm':0.05}
      wscyldata=self.ccyls.GetWSCylinders(self.cpic.GetPicarroID())
      if not wscyldata:
         raise ValueError("\ncPicarroWS._readWSData() - NO WS data from cCylinders.GetWSCylinders()!?")
      # [[cylid,installdt,enddt,station,instrid,port,role]]
      p=[]
      for dasdt,daedt in self.datasedt:
         for wsc in wscyldata:
            if wsc[2] and wsc[2]<=dasdt: continue
            if wsc[1]>=daedt: break
            if not wsc[2]:
               edt=daedt
            else:
               edt=min(daedt,wsc[2])
            sdt=max(dasdt,wsc[1])
            p.append("(datetime_utc>='%s' AND datetime_utc<='%s' AND port=%s)"%(sdt,edt,wsc[5]))
      flds,pp=['datetime_utc'],[]
      for i,co in enumerate(self.cpic.MeasuredComponents()):
         flds.append('avg_%s'%(co,))
         if co[:4]=='h2o_': continue
         pp.append('avg_%s>%s'%(co,MINLIMS[co]))
      sq="SELECT %s FROM %s WHERE flag IN (' ','.','C') AND (%s) AND (%s) ORDER BY datetime_utc;"%(
         ','.join(flds),self.cpic.GetRawTableName(),' OR '.join(p),' AND '.join(pp))

      db=cDB(DBNAME)
      cu=db.cursor()
      try:
         cu.execute(sq)
      except:
         print "\ncPicarroWS._readWSData():"
         print sq
         raise
      rawdata=cu.fetchall()
      cu.close() ; db.close()
      if len(rawdata)<1:
         raise ValueError("\ncPicarroWS._readWSData() - NO WS data in %s\n%s\n"%(
            self.cpic.GetRawTableName(),sq))
      
      ## calculate dry averages,stds for all the components for each meas. period
      ## ws==[absdays (at middle),co2avg,co2std,...,h2oavg,h2ostd]
      ## rda==rawdata, except [0]==absdays
      idx,ws,rda,wschangedts=0,[],[],[]
      lenrawdata=len(rawdata)
      while idx<lenrawdata:
         rda.append([rawdata[idx][0].absdays]+list(rawdata[idx][1:]))
         if rda and (idx==lenrawdata-1 or (rawdata[idx+1][0]-rawdata[idx][0]).minutes>9): # calc
            rdaar=np.array(rda,dtype=np.float)     # convert None->nan
            try:  
               avgs=np.nanmean(rdaar,0)
            except:
               print "\n_readWSData: datalen:%s, idx:%s, len(da):%s, da:%s"%(
                  lenrawdata,idx,len(rda),rda)
               raise
            w=[avgs[0]]
            if len(rda)>1:
               try:
                  stds=np.nanstd(rdaar,0,ddof=1)
               except:
                  print 
                  print rda, DateTime.DateTimeFromAbsDays(rda[0][0])
                  print
                  raise
            else:
               stds=[0]*len(rda[0])
            for a,s in zip(avgs[1:],stds[1:]): 
               w.extend([a,s])
            dryc=self.cpic.DryConc(rawdata[idx][0],w[1:])
            
            if len(ws)>0:  # check for too large change
               for commppo,iii in zip(('co2','ch4','co'),range(1,2*self.ncompos,2)):
                  if abs(ws[-1][iii]-dryc[iii-1])>self.wsdifflimits[commppo]: 
                     if self.ccyls.GetWSCylIDs(FROMABSDAYS(ws[-1][0]))==self.ccyls.GetWSCylIDs(FROMABSDAYS(w[0])):
                        ca,cb=ws[-1][iii],dryc[iii-1]
                        print "  _readWSData(): %19.19s,%s - too large WS diff: %.5f->%.5f (%.5f) - SKIPPED"%(
                           FROMABSDAYS(w[0]),commppo,ca,cb,abs(ca-cb))
                        dryc=None
                        break
                     else:    # WSCHANGE
                        wschdt=self.ccyls.GetWSCylIDs(FROMABSDAYS(w[0]))
                        if wschdt is None: continue
                        wschdt=wschdt[0][0]
                        if wschdt not in wschangedts: 
                           wschangedts.append(wschdt)
                           print "Note: WS change at %19.19s"%(wschdt,)
            if dryc:
               ws.append([w[0]]+dryc)
            rda=[]
         idx+=1

         
      wscyls=self.ccyls.GetWSCylinders()  #[[cylid,installdt,enddt,port]]
      nwscyls=len(wscyls)
      wscylidx=0
      if wscyls[wscylidx][2] and nwscyls>1:
         while wscyls[wscylidx][2].absdays<ws[0][0]: 
            wscylidx+=1
            if not wscyls[wscylidx][2]: break
      wscyl=wscyls[wscylidx]
      
      # interpolation...
      hasco=len(ws[0])>7
      ## ws[[absdays,co2,ch4,[co]]]
      # extend for half a day before (and after)
      wseqs=[[ws[0][0]-.5,ws[0][0]]] #[[absdays_start,absdays_end,eqstr]]
      if hasco:
         wseqs[-1].append("[%s,%s,%s]"%(ws[0][1],ws[0][3],ws[0][5]))
      else:
         wseqs[-1].append("[%s,%s]"%(ws[0][1],ws[0][3]))
         
      # TODO: WS change???
      for idx in range(len(ws)-1):
         if wscyl[2] and ws[idx+1][0]>wscyl[2].absdays:
            # time to extend current and upcoming (if any)
            wseqs.append([ws[idx][0],wscyl[2].absdays])
            if hasco:
               wseqs[-1].append("[%s,%s,%s]"%(ws[idx][1],ws[idx][3],ws[idx][5]))
            else:
               wseqs[-1].append("[%s,%s]"%(ws[idx][1],ws[idx][3]))
            wscylidx+=1
            wscyl=wscyls[wscylidx]
            if wscyl[1].absdays<ws[idx+1][0]:  # extend
               wseqs.append([wscyl[1].absdays,ws[idx+1][0]])
               if hasco:
                  wseqs[-1].append("[%s,%s,%s]"%(ws[idx+1][1],ws[idx+1][3],ws[idx+1][5]))
               else:
                  wseqs[-1].append("[%s,%s]"%(ws[idx+1][1],ws[idx+1][3]))
            continue
               
         ddt=ws[idx+1][0]-ws[idx][0]
         if hasco:
            dy=[ws[idx+1][1]-ws[idx][1],ws[idx+1][3]-ws[idx][3],ws[idx+1][5]-ws[idx][5]]
            dydt=[dy[0]/ddt,dy[1]/ddt,dy[2]/ddt]
            eq="[(absdays-%s)*%s+%s,(absdays-%s)*%s+%s,(absdays-%s)*%s+%s]"%(
               ws[idx][0],dydt[0],ws[idx][1],
               ws[idx][0],dydt[1],ws[idx][3],
               ws[idx][0],dydt[2],ws[idx][5])
         else:
            dy=[ws[idx+1][1]-ws[idx][1],ws[idx+1][3]-ws[idx][3]]
            dydt=[dy[0]/ddt,dy[1]/ddt]
            eq="[(absdays-%s)*%s+%s,(absdays-%s)*%s+%s]"%(ws[idx][0],dydt[0],ws[idx][1],
               ws[idx][0],dydt[1],ws[idx][3])
         wseqs.append([ws[idx][0],ws[idx+1][0],eq])
      wseqs.append([ws[-1][0],ws[-1][0]+.5])
      if hasco:
         wseqs[-1].append("[%s,%s,%s]"%(ws[-1][1],ws[-1][3],ws[-1][5]))
      else:
         wseqs[-1].append("[%s,%s]"%(ws[-1][1],ws[-1][3]))
      return wseqs
      
      
   def GetCWSpWS(self,dt,compo=None):
      '''
         Return C(WS)/WS at time dt as [co2ratio,ch4ratio,[coratio]] if compo is None,
         otherwise return only the ratio
         
         2017-02-07 +compo
      '''
      rws=self.GetWS(dt,compo)
      if not rws: return None
      cws=self.GetCWS(dt,compo)
      if not cws: return None
      if compo: return cws/rws
      return [c/r for c,r in zip(cws,rws)]
      
   
   def GetWS(self,dt,compo=None):
      '''
         Return (interpolated) dry WS raw reading at time dt
         No extrapolation: return None if outside [first,last]
         
         2017-02-07 +compo
      '''
      absdays=dt.absdays
      if absdays<self.wseqs[0][0]-EXTRAPOLATIONLIMIT or absdays>self.wseqs[-1][0]+EXTRAPOLATIONLIMIT: 
         return None
      
      if self.wseqsidx>0 and absdays<self.wseqs[self.wseqsidx][1]: 
         self.wseqsidx=0            # not needed if always called with increasing dt
      try:
         while absdays>=self.wseqs[self.wseqsidx][1]: 
            self.wseqsidx+=1
            if self.wseqsidx==len(self.wseqs):
               self.wseqsidx-=1
               break
      except IndexError:
         pass
      try:
         dd=absdays-self.wseqs[self.wseqsidx][0]
         rws=eval(self.wseqs[self.wseqsidx][2])
      except:
         print "\nPicarroWS.GetWS(%s):"%(dt,)
         print "absdays:%s,wseqidx:%s,len(wseqs):%s"%(absdays,self.wseqsidx,len(self.wseqs))
         raise
      if compo is not None:
         return rws[self.compoidxdct[compo]]
      return rws
      
   def GetCWS(self,dt,compo=None):
      '''
         Return WS concentration
         If compo is None, return as a list, otherwise only for the component in question
      '''
      if compo:
         return self.ccyls.GetWSCylConc(dt,compo)
      ret=[]
      for co in self.compoidxdct:
         ret.append(self.ccyls.GetWSCylConc(dt,co))
      return ret
      
      
   def GetWSCWS(self,dt,compo=None):
      '''
         Return interpolated [WS],  [WS conc]
      '''
      rws=self.GetWS(dt,compo)
      cws=self.GetCWS(dt,compo)
      return rws,cws
      


def ParseDateTime(dtstr,tstr=''):
   '''
      Parse datetime
   '''
   ds=dtstr
   if tstr:
      ds="%s %s"%(dtstr.strip(),tstr.strip())
   try:
      dt=PARSEDT(ds)
   except:
      try:
         dt=DateTime.strptime(ds,"%m/%d/%y %H:%M:%S")
      except:
         ds="%s %8.8s"%(dtstr.strip(),tstr.strip())
         try:
            dt=DateTime.strptime(ds,"%m/%d/%y %H:%M:%S")
         except:
            print "'%s','%s' -> '%s'"%(dtstr,tstr,ds)
            raise
   return dt
