import math

def polfit(dx, dy, dsd, nfp, mode):
    """
    Least squares fit to data with a polynomial curve

    np: number of points
    dx[]: independent vars
    dy[]: dependent vars
    dsd[] std dev for y data points
    nfp: no of coefficients (pol deg + 1)
    mode:   1: (instrumental) weight(i) = 1.0/dsd(i)**2
            0: no weighting
           -1: (statistical) weight(i) = 1.0/y(i)
    returns tuple ([coeff], chisq, stdev, r2)

    http://ast.star.rl.ac.uk/dev/starman/lib_source.html#polfit
    (correlation coefficient added byjah)

    Revision history:
      2003-03-17 R2 calc. fixed (<=1.0)
    """

    # Initialize ...
    np = len(dx)

    da = []; dsumx = []; dsumy = []
    darray = []
    da.append(dy[0])
    retv = []
    for k in range(1, nfp):
        da.append(0.)
        retv.append(0.)
    dchisq = drms = 0.
    retval = (retv, 0., 0., 0.)
    for k in range(0, 19):
        if k < 10:
            dsumy.append(0.)
        dsumx.append(0.)
    for k in range(0, 10):
        darray.append(dsumy[:])


    # Check for all values same...
    same = 1
    for k in range(0, np):
        if dy[k] != dy[0]: same = 0
    if same: return retval

    # Accumulate weighted sums
    nmax = 2 * nfp - 1
    dweight = 1.
    for i in range(0, np):
        dxi = dx[i]; dyi = dy[i]
        if mode == 1:
            dweight = 1./(dsd[i]*dsd[i])
        elif mode == -1:
            if dyi != 0: dweight = 1./abs(dyi)
        dxterm = dweight

        for n in range(0, nmax):
            dsumx[n] += dxterm
            dxterm *= dxi
        dyterm = dweight * dyi
        for n in range(0, nfp):
            dsumy[n] += dyterm
            dyterm *= dxi
        dchisq += dweight * dyi * dyi

    # Construct matrices and calc. coefficients
    for j in range(0, nfp):
        for k in range(0, nfp):
            n = j + k
            darray[j][k] = dsumx[n]
    ddelta = determ(darray, nfp)

    # If determ is zero, exit
    if ddelta == 0:
        dchisq = 0.
        for j in (0, nfp): da[j] = 0.
        return (da, dchisq, drms, 0.)

    for l in range(0, nfp):
        for j in range(0, nfp):
            for k in range(0, nfp):
                n = j + k
                darray[j][k] = dsumx[n]
            darray[j][l] = dsumy[j]
        da[l] = determ(darray, nfp)/ddelta

    # Calc chi square
    nfree = np - nfp
    if nfree <= 0: return (da, dchisq, drms, 1.)
    for j in range(0, nfp):
        dchisq -= (2. * da[j] * dsumy[j])
        for k in range(0, nfp):
            n = j + k
            dchisq += (da[j] * da[k] * dsumx[n])
    dchisq /= nfree

    # Calc std dev
    drms = 0.; dysum = 0.
    for k in range(0, np):
        dcalc = da[0]
        dysum += dy[k]
        if nfp != 1:
            for j in range(1, nfp):

                dcalc += (da[j] * (dx[k] ** j))
            drms += (dy[k] - dcalc)*(dy[k] - dcalc)
    drms = math.sqrt(drms/nfree)

    # Calc R2
    dysum /= np; s1 = 0.; s2 = 0.
    for k in range(0, np):
        yi = 0.
        for i in range(nfp): yi += (dx[k]**i)*da[i]
        s1 += (yi - dysum) ** 2
        s2 += (dy[k] - dysum) ** 2
    r2=s1/s2
    if r2>1: r2=1-(r2-1)    # is this right?
    return (da, dchisq, drms, r2)



def determ(array, norder):
    """
    Calc the determinant of a square matrix
    Destroys the input matrix array. Dim valid up to 10

    http://ast.star.rl.ac.uk/dev/starman/lib_source.html#determ
    """
    deter = 1.
    for k in range(0, norder):
        # interchange columns if diagonal element is zero
        if array[k][k] == 0:
            j = k
            while array[k][j] == 0 and j < norder - 1: j += 1
            if j == norder - 1: return 0.
            for i in range(k, norder):
                save = array[i][j]
                array[i][j] = array[i][k]
                array[i][k] = save
            deter *= -1.

        # subtract row k from lower rows to get diagonal matrix
        deter *= array[k][k]
#        print "k, deter, arr: ", k, deter,array[k][k]
        if k < norder - 1:
            k1 = k + 1
            i = k1
            while 1:
                j = k1
                while 1:
                    array[i][j] -= array[i][k] * array[k][j]/array[k][k]
                    j += 1
                    if j >= norder: break
                i += 1
                if i >= norder: break

    return deter
