"""TELNET client class.

Based on telnetlib.py standard python library.  Modified by Alastair Basden
to allow a timeout to be set when making the connection.  By default this
is 2 seconds (which may not be long enough in some situations).
Also implements a read_for(timeout) function, which will either return
immediately if data is ready, or wait for timeout seconds, and then return,
either any data, or nothing.
Also implements a test() function, which tests the connection.  If okay,
returns 0.  Otherwise, returns 1, and any unread data can be received by
calling read_very_lazy().
This should then be checked immediately if the data is wanted, before a
new connection is made.

This version copyrighted by the author, Alastair Basden, please feel free to use it
free of charge, but this paragraph must be kept intact.

Based on RFC 854: TELNET Protocol Specification, by J. Postel and
J. Reynolds

Example:

>>> import tellib
>>> a=tellib.Telnet()
>>> a.open("www.python.org",79) #connect to finger port
>>> a.write('guido\r\n')
>>> print tn.read_all()
Login       Name               TTY         Idle    When    Where
guido    Guido van Rossum      pts/2        <Dec  2 11:10> snag.cnri.reston..
 
>>>

Note that read_all() won't read until eof -- it just reads some data
-- but it guarantees to read at least one byte unless EOF is hit.

It is possible to pass a Telnet object to select.select() in order to
wait until more data is available.  Note that in this case,
read_eager() may return '' even if there was data on the socket,
because the protocol negotiation may have eaten the data.  This is why
EOFError is needed in some cases to distinguish between "no data" and
"connection closed" (since the socket also appears ready for reading
when it is closed).

Bugs:
- may hang when connection is slow in the middle of an IAC sequence

To do:
- option negotiation
- timeout should be intrinsic to the connection object instead of an
  option on one of the read calls only

"""
## 2010-09-02 byjah. Fix by adding timeout to call __init__

# Imported modules
import sys
import socket
import select
import telnetlib


# Tunable parameters
DEBUGLEVEL = telnetlib.DEBUGLEVEL

# Telnet protocol defaults
TELNET_PORT = telnetlib.TELNET_PORT
# Telnet protocol characters (don't change)
IAC  = telnetlib.IAC # "Interpret As Command"
DONT = telnetlib.DONT
DO   = telnetlib.DO
WONT = telnetlib.WONT
WILL = telnetlib.WILL
theNULL = telnetlib.theNULL


class Telnet(telnetlib.Telnet):

   """
   test() - tests the connection returning 1 if an error.  Unread data can
   then be received by read_very_lazy()
   
   read_for(timeout) - gets data from the connection, or returns with nothing
   if timeout.
   
   timeout(val) - changes the timeout value.
   open() - tries to open the connection within timeout.
   
   """

   def __init__(self, host=None, port=0, timeout=2):
      """Constructor.
      
      When called without arguments, create an unconnected instance.
      With a hostname argument, it connects the instance; a port
      number is optional.
      
      """
      self._timeout=timeout
      telnetlib.Telnet.__init__(self,host,port,timeout)
   
   def open(self, host, port=0, timeout=None):
      """Connect to a host.
      
      The optional second argument is the port number, which
      defaults to the standard telnet port (23).
      
      Don't try to reopen an already connected instance.
      
      """
#      from time import time
#      print "%s tellib open, host='%s',port='%s',timeout='%s'"%(
#         time(),host,port,str(timeout)) #debug
      if(timeout!=None): self._timeout=timeout
      timeout=self._timeout
      
      self.eof = 0
      self.cookedq=""
      if not port:
         port = TELNET_PORT
      self.host = host
      self.port = port
         
      msg = "getaddrinfo returns an empty list"
      for res in socket.getaddrinfo(host,port,0,socket.SOCK_STREAM):
         af,socktype,proto,canonname,sa = res
         try:
            self.sock=socket.socket(af, socktype, proto)
            self.sock.settimeout(self._timeout)
            self.sock.connect(sa)
         except socket.error, msg:
            if self.sock:
               self.sock.close()
            self.sock = None
            continue
         break
      if not self.sock:
         raise socket.error, msg
         
